/* $Id: objtoc.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright 
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/* 
 *  Generate calls that translate objects to their C equivalents.  In
 *  the case of object that do not correspond directly to C data types,
 *  return the value pointer, or at least try to do something 
 *  intelligent.
 *
 *  In this version, C int and char types are synonymous, Float class
 *  translates to double, as in C, and a long double class isn't 
 *  implemented.
 *
 *  Long int and int are synonymous, as in GNU C for most, if not all, 
 *  platforms.
 *
 *  NOTE - Future versions might allow the class library to define
 *  what instance variable to return in a C context.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "lex.h"
#include "list.h"

#define IS_SIGN(c) ((c == '-' || c == '+'))

/*
 *  Static in case a function needs to do a conversion independent
 *  of its value.
 */
static void *valptr;

/*
 *  This buffer should be large enough for char * return values.
 */
#define MAX_RETURN_VALPTRS MAXARGS
static LIST *return_valptrs = NULL;
static LIST *return_valptrs_head = NULL;
static int n_return_vals = 0;

static char *buf_from_return_list (int length) {
  LIST *l, *t;
  l = new_list ();
  l -> data = __xalloc (length + 1);
  if (return_valptrs == NULL) {
    return_valptrs = return_valptrs_head = l;
  } else {
    if (n_return_vals >= MAX_RETURN_VALPTRS) {
      t = list_unshift (&return_valptrs);
      --n_return_vals;
      __xfree (t -> data);
      __xfree (t);
    }
    list_add (return_valptrs_head, l);
    return_valptrs_head = l;
  }
  ++n_return_vals;
  return (char *)l -> data;
}

void delete_return_valptrs (void) {
  delete_list (&return_valptrs);
}

char __ctalk_to_c_char (OBJECT *o) {
  char c;
  OBJECT *o_value;
  char buf[MAXLABEL];

  if (!o) {
#ifdef WARN_UNDEFINED_OBJECT_RETURNS
    _warning ("__ctalk_to_c_char: Undefined value.\n");
#endif
    return '\0';
  }

  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE)) != NULL) {
    strcpy (buf, o_value -> __o_value);
    while ((*buf == '\'') && (buf[1] != '\0'))
      TRIM_CHAR (buf);
    if (strlen (buf) > 1) { /* Convert from Integer if possible. */
      c = (char)atoi (buf);
    } else {
      sscanf (buf, "%c", &c);
    }
  } else {

    if (o -> __o_value)
      strcpy (buf, o -> __o_value);
    else
      strcpy (buf, NULLSTR);

    while ((*buf == '\'') && (buf[1] != '\0'))
    while (*buf == '\'')
      TRIM_CHAR (buf);
    sscanf (buf, "%c", &c);
  }
  return c;
}

double __ctalk_to_c_double (OBJECT *o) {
  double d;
  OBJECT *o_value;

  if (!o) {
#ifdef WARN_UNDEFINED_OBJECT_RETURNS
    _warning ("__ctalk_to_c_double: Undefined value.\n");
#endif
    return 0.0;
  }

  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE)) != NULL) 
    sscanf (o_value -> __o_value, "%lf", &d);
  else
    sscanf (o -> __o_value, "%lf", &d);
  return d;
}

/* The double conversions don't generally need a, "keep" parameter,
   because they're not used in conditional predicates, unlike the
   int conversion functions.  That could change, of course. */
double __ctalkToCDouble (OBJECT *o) {
  return __ctalk_to_c_double (o);
}

int __ctalk_to_c_int (OBJECT *o) {
  int i, idx, bin_exp;
  char c;
  OBJECT *o_value;
  char buf[MAXLABEL];

  if (!o) {
#ifdef WARN_UNDEFINED_OBJECT_RETURNS
    _warning ("__ctalk_to_c_int: Undefined value.\n");
#endif
    return 0;
  }

  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE)) != NULL) {
    strcpy (buf, o_value -> __o_value);
  } else {
    strcpy (buf, o -> __o_value);
    o_value = o;
  }

  while (*buf == '\'')
    TRIM_CHAR (buf);

  if (!strcmp (o_value -> __o_classname, INTEGER_CLASSNAME) ||
      __ctalkIsSubClassOf (o_value -> __o_classname, INTEGER_CLASSNAME)) {
#if defined(__GNUC__) && defined(__sparc__) && defined(__svr4__)
    if (isdigit ((int)buf[0]) || IS_SIGN(buf[0])) {
#else
    if (isdigit (buf[0]) || IS_SIGN(buf[0])) {
#endif
      switch (radix_of (buf))
	{
	case hexadecimal:
	  if (!sscanf (buf, "%x", &i))
	    sscanf (buf, "%X", &i);
	  break;
	case octal:
	  sscanf (buf, "%o", &i);
	  break;
	case binary:
	  for (idx = strlen(buf) - 1, bin_exp = 1, i = 0; idx >= 0; idx--) {
	    if ((buf[idx] == 'b') || (buf[idx] == 'B'))
	      continue;
	    i += (buf[idx] == '1') ? bin_exp: 0;
	    bin_exp *= 2;
	  }
	  break;
	case decimal:
	default:
	  sscanf (buf, "%d", &i);
	  break;
	}
    } else {
      if (!strcmp (buf, NULLSTR)) {
	i = 0;
      } else {
	if (*buf == '\0') {
	  i = 0;
	} else {
	  _warning ("__ctalk_to_c_int: Incompatible return type.\n");
	}
      }
    }
  } else {
    if (!strcmp (o_value -> __o_classname, CHARACTER_CLASSNAME)) {
      if (!strcmp (buf, NULLSTR)) {
	i = 0;
      } else {
	sscanf (buf, "%c", &c);
	i = (int)c;
      }
    } else {
      /*
       *  Any other class.
       */
      if ((*buf == '\0') || !strlen (buf) || !strcmp (buf, NULLSTR)) {
	i = 0;
      } else {
	if (strstr (buf, "ll") || strstr (buf, "L")) {
	  long long int l;
	  sscanf (buf, "%lld", &l);
	  i = (int)l;
	} else {
	  if (strstr (buf, "l")) {
	    long int l;
	    sscanf (buf, "%ld", &l);
	    i = (int)l;
	  } else {
	    sscanf (buf, "%d", &i);
	  }
	}
      }
    }
  } 

  return i;
}

int __ctalkToCInteger (OBJECT *o, int keep) {
  int i, idx, bin_exp;
  char c;
  OBJECT *o_value;
  char buf[MAXLABEL];

  if (!IS_OBJECT(o)) {
#ifdef WARN_UNDEFINED_OBJECT_RETURNS
    _warning ("__ctalk_to_c_int: Undefined value.\n");
#endif
    return 0;
  }

  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE)) != NULL) {
    strcpy (buf, o_value -> __o_value);
  } else {
    if (o -> __o_value)
      strcpy (buf, o -> __o_value);
    else 
      strcpy (buf, NULLSTR);
    o_value = o;
  }

  while (*buf == '\'')
    TRIM_CHAR (buf);

  if (!strcmp (o_value -> __o_classname, INTEGER_CLASSNAME) ||
      __ctalkIsSubClassOf (o_value -> __o_classname, INTEGER_CLASSNAME)) {
#if defined(__GNUC__) && defined(__sparc__) && defined(__svr4__)
    if (isdigit ((int)buf[0]) || IS_SIGN(buf[0])) {
#else
    if (isdigit (buf[0]) || IS_SIGN(buf[0])) {
#endif
      switch (radix_of (buf))
	{
	case hexadecimal:
	  if (!sscanf (buf, "%x", &i))
	    sscanf (buf, "%X", &i);
	  break;
	case octal:
	  sscanf (buf, "%o", &i);
	  break;
	case binary:
	  for (idx = strlen(buf) - 1, bin_exp = 1, i = 0; idx >= 0; idx--) {
	    if ((buf[idx] == 'b') || (buf[idx] == 'B'))
	      continue;
	    i += (buf[idx] == '1') ? bin_exp: 0;
	    bin_exp *= 2;
	  }
	  break;
	case decimal:
	default:
	  sscanf (buf, "%d", &i);
	  break;
	}
    } else {
      if (!strcmp (buf, NULLSTR)) {
	i = 0;
      } else {
	if (*buf == '\0') {
	  i = 0;
	} else {
	  _warning ("__ctalkToCInteger: Incompatible return type.\n");
	}
      }
    }
  } else {
    if (!strcmp (o_value -> __o_classname, CHARACTER_CLASSNAME)) {
      if (!strcmp (buf, NULLSTR)) {
	i = 0;
      } else {
	sscanf (buf, "%c", &c);
	i = (int)c;
      }
    } else {
      /*
       *  Any other class.
       */
      if ((*buf == '\0') || !strlen (buf) || !strcmp (buf, NULLSTR)) {
	i = 0;
      } else {
	if (strstr (buf, "ll") || strstr (buf, "L")) {
	  long long int l;
	  sscanf (buf, "%lld", &l);
	  i = (int)l;
	} else {
	  if (strstr (buf, "l")) {
	    long int l;
	    sscanf (buf, "%ld", &l);
	    i = (int)l;
	  } else {
	    void *__r;
	    if ((__r = generic_ptr_str (buf)) != NULL) {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
	      _warning ("__ctalkToCInteger: Cast from pointer to int may be truncated.\n");
#endif
	      /*
	       *  A compiler warning here is okay.
	       */
	      i = (int)__r;
	    } else {
	      if (isalpha((int)buf[0])) {
		i = (int)buf[0];
	      } else {
		sscanf (buf, "%d", &i);
	      }
	    }
	  }
	}
      }
    }
  } 

    if (!keep) {
      OBJECT *__t;
      if (((__t = __ctalk_get_object(o->__o_name, o->__o_classname)) != o) &&
	  !__ctalkFindClassVariable (o->__o_name, FALSE) &&
	  !is_receiver (o) &&
	  !is_arg (o)) {
	__ctalkDeleteObjectInternal (o -> __o_p_obj ? o -> __o_p_obj : o);
      }
    }
  return i;
}

long long int __ctalk_to_c_longlong (OBJECT *o) {
  long long int l;
  OBJECT *o_value;
  char buf[MAXLABEL];
  if (!o) {
#ifdef WARN_UNDEFINED_OBJECT_RETURNS
    _warning ("__ctalk_to_c_longlong: Undefined value.\n");
#endif
    return 0LL;
  }
  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE)) != NULL) {
    strcpy (buf, o_value -> __o_value);
  } else {
    strcpy (buf, o -> __o_value);
  }

#if defined(__GNUC__) && defined(__sparc__) && defined(__svr4__)
  if (isdigit ((int)buf[0]) || IS_SIGN(buf[0])) {
#else
  if (isdigit (buf[0]) || IS_SIGN(buf[0])) {
#endif
    sscanf (buf, "%lld", &l);
  } else {
    if (!strcmp (buf, NULLSTR)) {
      l = 0ll;
    } else {
      if (*buf == '\0') {
	l = 0ll;
      } else {
	_warning ("__ctalk_to_c_longlong: Incompatible return type %s.\n", buf);
      }
    }
  }

  return l;
}

char *__ctalk_to_c_char_ptr (OBJECT *o) {
  OBJECT *o_value, *o_return;
  if (!IS_OBJECT (o)) return NULL;
  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE)) != NULL) {
    if (o_value -> __o_value) {

      if ((o_return = obj_ref_str(o_value -> __o_value)) != NULL) {
	return (char *)o_return;
      } else {
	return o_value -> __o_value;
      }

    } else {

      return NULL;

    }
  } else { 

    if (o -> __o_value) {
      if (!strncmp (o -> __o_value, "0x", 2)) {
	sscanf (o -> __o_value, "0x%p", &valptr);
	return (char *)valptr;
      } else {
	return o -> __o_value;
      }
    } else {

      return NULL;

    }
  }
}

static void __delete_char_ptr_arg (OBJECT *__o) {
  OBJECT *__t;
  if (((__t = __ctalk_get_object(__o->__o_name, __o->__o_classname)) != __o) &&
      !__ctalkFindClassVariable (__o->__o_name, FALSE) &&
      !is_receiver (__o) &&
      !is_arg (__o)) {
    if (__o -> __o_p_obj) {
      if (((__t = __ctalk_get_object
	    (__o->__o_p_obj->__o_name, __o->__o_p_obj->__o_classname)) 
	   != __o -> __o_p_obj) &&
	  !__ctalkFindClassVariable (__o->__o_p_obj->__o_name, FALSE) &&
	  !is_receiver (__o->__o_p_obj) &&
	  !is_arg (__o->__o_p_obj)) {
	__ctalkDeleteObjectInternal (__o -> __o_p_obj ? __o -> __o_p_obj : __o);
      }
    }
  }
}

char *__ctalkToCCharPtr (OBJECT *o, int keep) {
  OBJECT *o_value, *o_return;
  char *v;
  static char v_buf[MAXLABEL];
  if (!IS_OBJECT (o)) return NULL;
  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE)) != NULL) {
    if (o_value -> __o_value != NULL) {

      if (((o_return = obj_ref_str(o_value -> __o_value)) != NULL)
	  && strcmp (o_value -> __o_classname, STRING_CLASSNAME)) {
	if (!keep) 
	  __delete_char_ptr_arg (o);
	return (char *)o_return;
      } else {
	if (IS_OBJECT (o_value) && (o_value -> __o_value != NULL)) {
	  if (!keep) 
	    __delete_char_ptr_arg (o);
	  v = buf_from_return_list (strlen (o_value -> __o_value));
	  strcpy (v, o_value -> __o_value);
	  return v;
	} else {
	  return NULL;
	}
      }
    } else {
      return NULL;
    }
  } else { 

    if (((o_return = obj_ref_str(o -> __o_value)) != NULL)
	&& strcmp (o_value -> __o_classname, STRING_CLASSNAME)) {
      if (!keep) 
	__delete_char_ptr_arg (o);
      return (char *)o_return;
    } else {
      if (!keep) 
	__delete_char_ptr_arg (o);
      if (o -> __o_value) {
	v = buf_from_return_list (strlen (o -> __o_value));
	strcpy (v, o -> __o_value);
	return v;
      } else {
	strcpy (v_buf, NULLSTR);
	return v_buf;
      }
    }
  }
}

/* 
 * Remember, we're returning a pointer here. 
 * TODO - there might be a warning here if the
 * function returns a scalar as a pointer.
 *
 ***
 * TODO - Do this for all subclasses of Magnitude,
 * works for String class, too.
 */

void *__ctalk_to_c_ptr (OBJECT *o) {
  OBJECT *o_value;
  RADIX radix;
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
  long unsigned int retval;
#else
  unsigned int retval;
#endif
  unsigned long long int retval_ll;
  char ptr_buf[32];

  if (!IS_OBJECT (o)) return NULL;
  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE)) == NULL)
    o_value = o;
  if (!strcmp (o_value -> __o_classname, INTEGER_CLASSNAME) ||
      __ctalkIsSubClassOf (o_value -> __o_classname, INTEGER_CLASSNAME)) {

    if (o_value -> __o_value) {

      if ((radix = radix_of (o_value -> __o_value)) == hexadecimal) {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
 	sscanf (o_value -> __o_value, "%lx", &retval);
#else
 	sscanf (o_value -> __o_value, "%x", &retval);
#endif
      } else {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
 	sscanf (o_value -> __o_value, "%ld", &retval);
#else
 	/*** FIX! Needs work... */
	sscanf (o_value -> __o_value, "%d", &retval);
#endif
      }
    } else {

      retval = 0;

    } 
    sprintf (ptr_buf, "%#x", retval);
    if (!chkptrstr (ptr_buf))
      _warning ("__ctalk_to_c_ptr (): Warning: value of object, \"%s,\" is not a pointer.\n",
		o -> __o_name);
    return (void *)retval;

  } else {
    if (!strcmp (o_value -> __o_classname, LONGINTEGER_CLASSNAME) ||
	__ctalkIsSubClassOf (o_value -> __o_classname, LONGINTEGER_CLASSNAME)) {

      if (o_value -> __o_value) {

 	if ((radix = radix_of (o_value -> __o_value)) == hexadecimal) {
 	  sscanf (o_value -> __o_value, "%llx", &retval_ll);
 	} else {
 	  sscanf (o_value -> __o_value, "%lld", &retval_ll);
 	}

#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
 	if (retval_ll > LONG_MAX) 
 	  _warning ("Warning: value of %s is truncated.\n", o -> __o_name);
 	return (void *) ((long unsigned int)(retval_ll & LONG_MAX));
#else
	sprintf (ptr_buf, "%#x", retval);
	if (!chkptrstr (ptr_buf))
	  _warning ("__ctalk_to_c_ptr (): Warning: value of object, \"%s,\" is not a pointer.\n",
		    o -> __o_name);
 	if (retval_ll > INT_MAX) 
 	  _warning ("Warning: value of %s is truncated.\n", o -> __o_name);
 	return (void *) ((unsigned int)(retval_ll & INT_MAX));
#endif
      } else {

	return (void *)0;

      }
    } else {
      if (!strcmp (o_value -> __o_classname, SYMBOL_CLASSNAME) ||
	  __ctalkIsSubClassOf (o_value -> __o_classname, SYMBOL_CLASSNAME)) {

	if (o_value -> __o_value) {

	  if (!strcmp (o_value -> __o_value, NULLSTR)) {
	    retval = 0;
	  } else {
	    if ((radix = radix_of (o_value -> __o_value)) == hexadecimal) {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
	      sscanf (o_value -> __o_value, "%lx", &retval);
#else
	      sscanf (o_value -> __o_value, "%x", &retval);
#endif
	    } else {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
	      sscanf (o_value -> __o_value, "%ld", &retval);
#else
	      sscanf (o_value -> __o_value, "%d", &retval);
#endif
	    }
	  }
	} else {

	  retval = 0;

	}
	if (retval > INT_MAX) 
	  _warning ("Warning: value of %s, \"%s,\" is truncated.\n", 
		    o -> __o_name, o -> __o_value);
	return (void *) retval;
      } else {

	if (o_value -> __o_value) {

	  if (obj_ref_str(o_value -> __o_value) ||
	      generic_ptr_str (o_value -> __o_value)) {
	    return __ctalkStrToPtr (o_value -> __o_value);
	  } else {
	    if (!strcmp (o_value -> __o_superclassname, "FileStream")) {
	      return __ctalkStrToPtr (o_value -> __o_value);
	    } else {
	      return o_value -> __o_value;
	    }
	  }

	} else {

	  return NULL;

	}
      }
    }
  }
}

void *__ctalkToCArrayElement (OBJECT *o) {
  OBJECT *value_var;
  if ((value_var = __ctalkGetInstanceVariable (o, "value", FALSE)) == NULL)
    value_var = o;

  if ((!strcmp (o -> __o_classname, CHARACTER_CLASSNAME) ||
      __ctalkIsSubClassOf (o -> __o_classname, CHARACTER_CLASSNAME)) &&
      (strcmp (o -> __o_classname, STRING_CLASSNAME) &&
       !__ctalkIsSubClassOf (o -> __o_classname, STRING_CLASSNAME))) {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
    long int c;
#else
    int c;
#endif
    void *char_value;
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
    c = (long int)__ctalk_to_c_char (value_var);
#else
    c = (int)__ctalk_to_c_char (value_var);
#endif
    char_value = (void *)c;
    return char_value;
  }
  /*
   *  Doubles are not compatible here.
   */

  if (!strcmp (o -> __o_classname, FLOAT_CLASSNAME) ||
      __ctalkIsSubClassOf (o -> __o_classname, FLOAT_CLASSNAME)) {
    _warning ("warning: Float array element returned as char pointer.\n");
    return value_var -> __o_value;
  }

  if (!strcmp (o -> __o_classname, INTEGER_CLASSNAME) ||
      __ctalkIsSubClassOf (o -> __o_classname, INTEGER_CLASSNAME)) {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
    long int i;
#else
    int i;
#endif
    void *int_val;
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
    i = (long int)__ctalk_to_c_int (value_var);
#else
    i = __ctalk_to_c_int (value_var);
#endif
    int_val = (void *)i;
    return int_val;
  }

  if (!strcmp (o -> __o_classname, LONGINTEGER_CLASSNAME) ||
      __ctalkIsSubClassOf (o -> __o_classname, LONGINTEGER_CLASSNAME)) {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
    long int l;
#else
    int l;
#endif
    void *ll_val;
    l = __ctalk_to_c_longlong (value_var);
    _warning ("warning: LongInteger array element truncated to int.\n");
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
    ll_val = (void *)(l & LONG_MAX);
#else
    ll_val = (void *)(l & INT_MAX);
#endif
    return (void *)ll_val;
  }

  if (!strcmp (o -> __o_classname, STRING_CLASSNAME) ||
      __ctalkIsSubClassOf (o -> __o_classname, STRING_CLASSNAME))
    return __ctalk_to_c_char_ptr (value_var);

  return value_var -> __o_value;
}

char *__ctalkArrayElementToCCharPtr (OBJECT *o) {
  static char *s;
  s = (char *)__ctalkToCArrayElement (o);
  return s;
}

char __ctalkArrayElementToCChar (OBJECT *o) {
  static char c;
  int i;
  i = (int)__ctalkToCArrayElement (o);
  c = (char)i;
  return c;
}

void *__ctalkArrayElementToCPtr (OBJECT *o) {
  return __ctalkToCArrayElement (o);
}

double __ctalkArrayElementToCDouble (OBJECT *o) {
  static double d = 0.0f;
  OBJECT *o_value;
  int r = 0;

  if (__ctalkIsSubClassOf (o -> __o_classname, FLOAT_CLASSNAME)) {

    if ((o_value = __ctalkGetInstanceVariableByName 
	 (o -> __o_name, "value", FALSE))
	!= NULL) {

      if (o_value -> __o_value) {
	r = sscanf (o_value -> __o_value, "%lf", &d);
      } else {
	r = sscanf (o -> __o_value, "%lf", &d);
      }
    } else {

      d = 0.0f;

    }
  }
  
  if (r == 0) {
    _warning("__ctalkArrayElementToCDouble: Object, \"%s,\""
	     "has invalid value.\n");
  }

  return d;
}

int __ctalkArrayElementToCInt (OBJECT *o) {
  static int i;
  i = (int)__ctalkToCArrayElement (o);
  return i;
}

long long int __ctalkArrayElementToCLongLongInt (OBJECT *o) {
  long long int l;
  int i;
  i = (int)__ctalkToCArrayElement (o);
  l = (long long int)i;
  return l;
}

int __ctalkToCIntArrayElement (OBJECT *o) {
  OBJECT *value_var;
  if ((value_var = __ctalkGetInstanceVariable (o, "value", FALSE)) == NULL)
    value_var = o;

  if (!strcmp (o -> __o_classname, INTEGER_CLASSNAME) ||
      __ctalkIsSubClassOf (o -> __o_classname, INTEGER_CLASSNAME)) {
    return __ctalk_to_c_int (value_var);
  }

  if (!strcmp (o -> __o_classname, LONGINTEGER_CLASSNAME) ||
      __ctalkIsSubClassOf (o -> __o_classname, LONGINTEGER_CLASSNAME)) {
    _warning ("warning: LongInteger array element truncated to int.\n");
    return (int) __ctalk_to_c_longlong (value_var);
  }

  return 0;
}
