/* $Id: pattypes.c,v 1.4 2014/05/08 10:42:56 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "lex.h"
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "typeof.h"
#include "rtinfo.h"

/*
 *  Complete expression is within arrayopen.. arrayclose pair.
 */
int is_simple_subscript (MESSAGE_STACK messages, int expr_start_idx,
			 int stack_start_idx, int stack_end_idx) {
  int prev_tok_idx, next_tok_idx, expr_end_idx;
  char *exprbuf;
  
  if ((prev_tok_idx = __ctalkPrevLangMsg (messages, expr_start_idx, 
				      stack_start_idx)) == ERROR)
    return FALSE;

  if (M_TOK(messages[prev_tok_idx]) != ARRAYOPEN) return FALSE;

  exprbuf = collect_expression (messages, expr_start_idx,
				&expr_end_idx, stack_start_idx,
				stack_end_idx);
  free (exprbuf);
  
  if ((next_tok_idx = __ctalkNextLangMsg (messages, expr_end_idx, 
				      stack_end_idx)) == ERROR)
    return FALSE;

  if (M_TOK(messages[next_tok_idx]) != ARRAYCLOSE) return FALSE;

  return TRUE;
}

int is_fmt_arg (MESSAGE_STACK messages, int expr_start_idx,
		int stack_start_idx, int stack_end_idx) {
  return (is_char_ptr_fmt_arg (messages,expr_start_idx,
			       stack_start_idx, stack_end_idx) ||
	  is_ptr_fmt_arg (messages,expr_start_idx,
			   stack_start_idx, stack_end_idx) ||
	  is_char_fmt_arg (messages,expr_start_idx,
			   stack_start_idx, stack_end_idx) ||
	  is_double_fmt_arg (messages,expr_start_idx,
			   stack_start_idx, stack_end_idx) ||
	  is_int_fmt_arg (messages,expr_start_idx,
			  stack_start_idx, stack_end_idx) ||
	  is_long_long_int_fmt_arg (messages,expr_start_idx,
				    stack_start_idx, stack_end_idx));
}

FMT_ARG_CTYPE is_fmt_arg_ctype (MESSAGE_STACK messages, int expr_start_idx,
				int stack_start_idx, int stack_end_idx) {
  if (is_char_ptr_fmt_arg (messages,expr_start_idx,
			   stack_start_idx, stack_end_idx))
    return fmt_arg_char_ptr;
  
  if (is_ptr_fmt_arg (messages,expr_start_idx,
		      stack_start_idx, stack_end_idx))
    return fmt_arg_ptr;

  if (is_char_fmt_arg (messages,expr_start_idx,
		       stack_start_idx, stack_end_idx))
    return fmt_arg_char;

  if (is_double_fmt_arg (messages,expr_start_idx,
			 stack_start_idx, stack_end_idx))
    return fmt_arg_double;
  if (is_int_fmt_arg (messages,expr_start_idx,
		      stack_start_idx, stack_end_idx))
    return fmt_arg_int;
  if (is_long_long_int_fmt_arg (messages,expr_start_idx,
				stack_start_idx, stack_end_idx))
    return fmt_arg_long_long_int;

  return null_collection_context;
}

int is_char_ptr_fmt_arg (MESSAGE_STACK messages, int expr_start_idx,
			 int stack_start_idx, int stack_end_idx) {
  int i, n_th_arg;
  int paren_lvl, block_lvl, array_lvl, start_paren_lvl, start_array_lvl;
  MESSAGE *m;
  int fmt_str_length;
  char *fmt_char_ptr;
  enum {
    ptr_fmt_arg_null,
    ptr_fmt_arg_expr,
    ptr_fmt_arg_separator,
    ptr_fmt_arg_literal
  } state;

  paren_lvl = block_lvl = array_lvl = n_th_arg = 0;
  start_paren_lvl = start_array_lvl = -1;
  for (i = expr_start_idx, state = ptr_fmt_arg_null; i <= stack_start_idx; i++) {

    m = messages[i];
    if (M_ISSPACE(m)) continue;

    switch (M_TOK(m))
      {
      case LABEL:
	if ((state == ptr_fmt_arg_null) || (state == ptr_fmt_arg_separator))
	  state = ptr_fmt_arg_expr;
	break; 
      case ARGSEPARATOR:
	if (start_paren_lvl != -1) {
	  if ((paren_lvl == start_paren_lvl) && !block_lvl && !array_lvl) {
	    state = ptr_fmt_arg_separator;
	    --n_th_arg;
	  }
	} else {
	  if (!paren_lvl && !block_lvl && !array_lvl) {
	    state = ptr_fmt_arg_separator;
	    --n_th_arg;
	  }
	}
	break;
      case LITERAL:
 	if (state == ptr_fmt_arg_separator) {
 	  if ((fmt_str_length = is_printf_fmt (M_NAME(m))) != 0) {
	    int __n;
	    /*
	     *  n_th_arg is -1 for the first argument, -2 for the 
	     *  second, -3 for the third, ....
	     */
  	    if ((n_th_arg == -1) && is_char_ptr_fmt(M_NAME(m))) {
  	      return TRUE;
 	    } else {
	      fmt_char_ptr = index (M_NAME(m), '%');


 	      for (__n = -2; ; __n--) {
		if ((fmt_char_ptr = index (fmt_char_ptr + 1, '%')) == NULL)
		  return FALSE;

		/* Check for literal "%%" */
		if (fmt_char_ptr[1] == '%')
		  continue;

		if (__n == n_th_arg)
		  break;
	      }

	      return is_char_ptr_fmt (fmt_char_ptr) && (__n == n_th_arg);

	    }
	  }
 	} else {
	  state = ptr_fmt_arg_expr;
	}
	break;
      case OPENPAREN:
 	++paren_lvl;
 	if (state == ptr_fmt_arg_null) {
 	  start_paren_lvl = paren_lvl;
	  state = ptr_fmt_arg_expr;
	} else {
	  if (start_paren_lvl != -1) {
	    if ((paren_lvl < start_paren_lvl) && (state == ptr_fmt_arg_expr)) {
	      return FALSE;
	    }
	  } else {
	    int __prev_idx;
	    if ((__prev_idx = __ctalkPrevLangMsg (messages, i, stack_start_idx)) 
		!= ERROR) {
	      /*
	       *  Function call syntax.
	       */
	      if (M_TOK(messages[__prev_idx]) == LABEL)
		return FALSE;
	    }
	  }
	}
	break;
      case CLOSEPAREN:
 	--paren_lvl;
	if (state == ptr_fmt_arg_separator)
	  state = ptr_fmt_arg_expr;
	break;
      case OPENBLOCK:
 	++block_lvl;
	break;
      case CLOSEBLOCK:
 	--block_lvl;
	break;
      case ARRAYOPEN:
 	++array_lvl;
 	if (state == ptr_fmt_arg_null)
 	  start_array_lvl = array_lvl;
	break;
      case ARRAYCLOSE:
 	--array_lvl;
	break;
      default:
	break;
      } /* switch (M_TOK(m)) */
  }
  return TRUE;
}

int is_ptr_fmt_arg (MESSAGE_STACK messages, int expr_start_idx,
		    int stack_start_idx, int stack_end_idx) {
  int i, n_th_arg;
  int paren_lvl, block_lvl, array_lvl, start_paren_lvl, start_array_lvl;
  MESSAGE *m;
  int fmt_str_length;
  char *fmt_char_ptr;
  enum {
    ptr_fmt_arg_null,
    ptr_fmt_arg_expr,
    ptr_fmt_arg_separator,
    ptr_fmt_arg_literal
  } state;

  paren_lvl = block_lvl = array_lvl = n_th_arg = 0;
  start_paren_lvl = start_array_lvl = -1;
  for (i = expr_start_idx, state = ptr_fmt_arg_null; i <= stack_start_idx; i++) {

    m = messages[i];
    if (M_ISSPACE(m)) continue;

    switch (M_TOK(m))
      {
      case LABEL:
	if ((state == ptr_fmt_arg_null) || (state == ptr_fmt_arg_separator))
	  state = ptr_fmt_arg_expr;
	break; 
      case ARGSEPARATOR:
	if (start_paren_lvl != -1) {
	  if ((paren_lvl == start_paren_lvl) && !block_lvl && !array_lvl) {
	    state = ptr_fmt_arg_separator;
	    --n_th_arg;
	  }
	} else {
	  if (!paren_lvl && !block_lvl && !array_lvl) {
	    state = ptr_fmt_arg_separator;
	    --n_th_arg;
	  }
	}
	break;
      case LITERAL:
 	if (state == ptr_fmt_arg_separator) {
 	  if ((fmt_str_length = is_printf_fmt (M_NAME(m))) != 0) {
	    int __n;
	    /*
	     *  n_th_arg is -1 for the first argument, -2 for the 
	     *  second, -3 for the third, ....
	     */
  	    if ((n_th_arg == -1) && is_ptr_fmt(M_NAME(m))) {
  	      return TRUE;
 	    } else {
	      fmt_char_ptr = index (M_NAME(m), '%');

 	      for (__n = -2; ; __n--) {
		if ((fmt_char_ptr = index (fmt_char_ptr + 1, '%')) == NULL)
		  return FALSE;

		/* Check for literal "%%" */
		if (fmt_char_ptr[1] == '%')
		  continue;

		if (__n == n_th_arg)
		  break;
	      }

	      return is_ptr_fmt (fmt_char_ptr) && (__n == n_th_arg);

	    }
	  }
 	} else {
	  state = ptr_fmt_arg_expr;
	}
	break;
      case OPENPAREN:
 	++paren_lvl;
 	if (state == ptr_fmt_arg_null) {
 	  start_paren_lvl = paren_lvl;
	  state = ptr_fmt_arg_expr;
	} else {
	  if (start_paren_lvl != -1) {
	    if ((paren_lvl < start_paren_lvl) && (state == ptr_fmt_arg_expr)) {
	      return FALSE;
	    }
	  } else {
	    int __prev_idx;
	    if ((__prev_idx = __ctalkPrevLangMsg (messages, i, stack_start_idx)) 
		!= ERROR) {
	      /*
	       *  Function call syntax.
	       */
	      if (M_TOK(messages[__prev_idx]) == LABEL)
		return FALSE;
	    }
	  }
	}
	break;
      case CLOSEPAREN:
 	--paren_lvl;
	if (state == ptr_fmt_arg_separator)
	  state = ptr_fmt_arg_expr;
	break;
      case OPENBLOCK:
 	++block_lvl;
	break;
      case CLOSEBLOCK:
 	--block_lvl;
	break;
      case ARRAYOPEN:
 	++array_lvl;
 	if (state == ptr_fmt_arg_null)
 	  start_array_lvl = array_lvl;
	break;
      case ARRAYCLOSE:
 	--array_lvl;
	break;
      default:
	break;
      } /* switch (M_TOK(m)) */
  }
  return TRUE;
}

int is_double_fmt_arg (MESSAGE_STACK messages, int expr_start_idx,
			 int stack_start_idx, int stack_end_idx) {
  int i, n_th_arg;
  int paren_lvl, block_lvl, array_lvl, start_paren_lvl, start_array_lvl;
  MESSAGE *m;
  int fmt_str_length;
  char *fmt_char_ptr;
  enum {
    ptr_fmt_arg_null,
    ptr_fmt_arg_expr,
    ptr_fmt_arg_separator,
    ptr_fmt_arg_literal
  } state;

  paren_lvl = block_lvl = array_lvl = n_th_arg = 0;
  start_paren_lvl = start_array_lvl = -1;
  for (i = expr_start_idx, state = ptr_fmt_arg_null; i <= stack_start_idx; i++) {

    m = messages[i];
    if (M_ISSPACE(m)) continue;

    switch (M_TOK(m))
      {
      case LABEL:
	if ((state == ptr_fmt_arg_null) || (state == ptr_fmt_arg_separator))
	  state = ptr_fmt_arg_expr;
	break; 
      case ARGSEPARATOR:
	if (start_paren_lvl != -1) {
	  if ((paren_lvl == start_paren_lvl) && !block_lvl && !array_lvl) {
	    state = ptr_fmt_arg_separator;
	    --n_th_arg;
	  }
	} else {
	  if (!paren_lvl && !block_lvl && !array_lvl) {
	    state = ptr_fmt_arg_separator;
	    --n_th_arg;
	  }
	}
	break;
      case LITERAL:
 	if (state == ptr_fmt_arg_separator) {
 	  if ((fmt_str_length = is_printf_fmt (M_NAME(m))) != 0) {
	    int __n;
	    /*
	     *  n_th_arg is -1 for the first argument, -2 for the 
	     *  second, -3 for the third, ....
	     */
  	    if ((n_th_arg == -1) && is_double_fmt(M_NAME(m))) {
  	      return TRUE;
 	    } else {
	      fmt_char_ptr = index (M_NAME(m), '%');

 	      for (__n = -2; ; __n--) {
		if ((fmt_char_ptr = index (fmt_char_ptr + 1, '%')) == NULL)
		  return FALSE;

		/* Check for literal "%%" */
		if (fmt_char_ptr[1] == '%')
		  continue;

		if (__n == n_th_arg)
		  break;
	      }

	      return is_double_fmt (fmt_char_ptr) && (__n == n_th_arg);


	    }
	  }
 	} else {
	  state = ptr_fmt_arg_expr;
	}
	break;
      case OPENPAREN:
 	++paren_lvl;
 	if (state == ptr_fmt_arg_null) {
 	  start_paren_lvl = paren_lvl;
	  state = ptr_fmt_arg_expr;
	} else {
	  if (start_paren_lvl != -1) {
	    if ((paren_lvl < start_paren_lvl) && (state == ptr_fmt_arg_expr)) {
	      return FALSE;
	    }
	  } else {
	    int __prev_idx;
	    if ((__prev_idx = __ctalkPrevLangMsg (messages, i, stack_start_idx)) 
		!= ERROR) {
	      /*
	       *  Function call syntax.
	       */
	      if (M_TOK(messages[__prev_idx]) == LABEL)
		return FALSE;
	    }
	  }
	}
	break;
      case CLOSEPAREN:
 	--paren_lvl;
	if (state == ptr_fmt_arg_separator)
	  state = ptr_fmt_arg_expr;
	break;
      case OPENBLOCK:
 	++block_lvl;
	break;
      case CLOSEBLOCK:
 	--block_lvl;
	break;
      case ARRAYOPEN:
 	++array_lvl;
 	if (state == ptr_fmt_arg_null)
 	  start_array_lvl = array_lvl;
	break;
      case ARRAYCLOSE:
 	--array_lvl;
	break;
      default:
	break;
      } /* switch (M_TOK(m)) */
  }
  return TRUE;
}

int is_char_fmt_arg (MESSAGE_STACK messages, int expr_start_idx,
		     int stack_start_idx, int stack_end_idx) {
  int i, n_th_arg;
  int paren_lvl, block_lvl, array_lvl, start_paren_lvl, start_array_lvl;
  MESSAGE *m;
  int fmt_str_length;
  char *fmt_char_ptr;
  enum {
    ptr_fmt_arg_null,
    ptr_fmt_arg_expr,
    ptr_fmt_arg_separator,
    ptr_fmt_arg_literal
  } state;

  paren_lvl = block_lvl = array_lvl = n_th_arg = 0;
  start_paren_lvl = start_array_lvl = -1;
  for (i = expr_start_idx, state = ptr_fmt_arg_null; i <= stack_start_idx; i++) {

    m = messages[i];
    if (M_ISSPACE(m)) continue;

    switch (M_TOK(m))
      {
      case LABEL:
	if ((state == ptr_fmt_arg_null) || (state == ptr_fmt_arg_separator))
	  state = ptr_fmt_arg_expr;
	break; 
      case ARGSEPARATOR:
	if (start_paren_lvl != -1) {
	  if ((paren_lvl == start_paren_lvl) && !block_lvl && !array_lvl) {
	    state = ptr_fmt_arg_separator;
	    --n_th_arg;
	  }
	} else {
	  if (!paren_lvl && !block_lvl && !array_lvl) {
	    state = ptr_fmt_arg_separator;
	    --n_th_arg;
	  }
	}
	break;
      case LITERAL:
 	if (state == ptr_fmt_arg_separator) {
 	  if ((fmt_str_length = is_printf_fmt (M_NAME(m))) != 0) {
	    int __n;
	    /*
	     *  n_th_arg is -1 for the first argument, -2 for the 
	     *  second, -3 for the third, ....
	     */
  	    if ((n_th_arg == -1) && is_char_fmt(M_NAME(m))) {
  	      return TRUE;
 	    } else {
	      fmt_char_ptr = index (M_NAME(m), '%');

 	      for (__n = -2; ; __n--) {
		if ((fmt_char_ptr = index (fmt_char_ptr + 1, '%')) == NULL)
		  return FALSE;

		/* Check for literal "%%" */
		if (fmt_char_ptr[1] == '%')
		  continue;

		if (__n == n_th_arg)
		  break;
	      }

	      return is_char_fmt (fmt_char_ptr) && (__n == n_th_arg);

	    }
	  }
 	} else {
	  state = ptr_fmt_arg_expr;
	}
	break;
      case OPENPAREN:
 	++paren_lvl;
 	if (state == ptr_fmt_arg_null) {
 	  start_paren_lvl = paren_lvl;
	  state = ptr_fmt_arg_expr;
	} else {
	  if (start_paren_lvl != -1) {
	    if ((paren_lvl < start_paren_lvl) && (state == ptr_fmt_arg_expr)) {
	      return FALSE;
	    }
	  } else {
	    int __prev_idx;
	    if ((__prev_idx = __ctalkPrevLangMsg (messages, i, stack_start_idx)) 
		!= ERROR) {
	      /*
	       *  Function call syntax.
	       */
	      if (M_TOK(messages[__prev_idx]) == LABEL)
		return FALSE;
	    }
	  }
	}
	break;
      case CLOSEPAREN:
 	--paren_lvl;
	if (state == ptr_fmt_arg_separator)
	  state = ptr_fmt_arg_expr;
	break;
      case OPENBLOCK:
 	++block_lvl;
	break;
      case CLOSEBLOCK:
 	--block_lvl;
	break;
      case ARRAYOPEN:
 	++array_lvl;
 	if (state == ptr_fmt_arg_null)
 	  start_array_lvl = array_lvl;
	break;
      case ARRAYCLOSE:
 	--array_lvl;
	break;
      default:
	break;
      } /* switch (M_TOK(m)) */
  }
  return TRUE;
}

int is_int_fmt_arg (MESSAGE_STACK messages, int expr_start_idx,
		     int stack_start_idx, int stack_end_idx) {
  int i, n_th_arg;
  int paren_lvl, block_lvl, array_lvl, start_paren_lvl, start_array_lvl;
  MESSAGE *m;
  int fmt_str_length;
  char *fmt_char_ptr;
  enum {
    int_fmt_arg_null,
    int_fmt_arg_expr,
    int_fmt_arg_separator,
    int_fmt_arg_literal
  } state;

  paren_lvl = block_lvl = array_lvl = n_th_arg = 0;
  start_paren_lvl = start_array_lvl = -1;
  for (i = expr_start_idx, state = int_fmt_arg_null; i <= stack_start_idx; i++) {

    m = messages[i];
    if (M_ISSPACE(m)) continue;

    switch (M_TOK(m))
      {
      case LABEL:
	if ((state == int_fmt_arg_null) || (state == int_fmt_arg_separator))
	  state = int_fmt_arg_expr;
	break; 
      case ARGSEPARATOR:
	if (start_paren_lvl != -1) {
	  if ((paren_lvl == start_paren_lvl) && !block_lvl && !array_lvl) {
	    state = int_fmt_arg_separator;
	    --n_th_arg;
	  }
	} else {
	  if (!paren_lvl && !block_lvl && !array_lvl) {
	    state = int_fmt_arg_separator;
	    --n_th_arg;
	  }
	}
	break;
      case LITERAL:
 	if (state == int_fmt_arg_separator) {
 	  if ((fmt_str_length = is_printf_fmt (M_NAME(m))) != 0) {
	    int __n;
	    /*
	     *  n_th_arg is -1 for the first argument, -2 for the 
	     *  second, -3 for the third, ....
	     */
  	    if ((n_th_arg == -1) && is_int_fmt(M_NAME(m))) {
  	      return TRUE;
 	    } else {
	      fmt_char_ptr = index (M_NAME(m), '%');

 	      for (__n = -2; ; __n--) {
		if ((fmt_char_ptr = index (fmt_char_ptr + 1, '%')) == NULL)
		  return FALSE;

		/* Check for literal "%%" */
		if (fmt_char_ptr[1] == '%')
		  continue;

		if (__n == n_th_arg)
		  break;
	      }

	      return is_int_fmt (fmt_char_ptr) && (__n == n_th_arg);
	    }
	  }
 	} else {
	  state = int_fmt_arg_expr;
	}
	break;
      case OPENPAREN:
 	++paren_lvl;
 	if (state == int_fmt_arg_null) {
 	  start_paren_lvl = paren_lvl;
	  state = int_fmt_arg_expr;
	} else {
	  if (start_paren_lvl != -1) {
	    if ((paren_lvl < start_paren_lvl) && (state == int_fmt_arg_expr)) {
	      return FALSE;
	    }
	  } else {
	    int __prev_idx;
	    if ((__prev_idx = __ctalkPrevLangMsg (messages, i, stack_start_idx)) 
		!= ERROR) {
	      /*
	       *  Function call syntax.
	       */
	      if (M_TOK(messages[__prev_idx]) == LABEL)
		return FALSE;
	    }
	  }
	}
	break;
      case CLOSEPAREN:
 	--paren_lvl;
	if (state == int_fmt_arg_separator)
	  state = int_fmt_arg_expr;
	break;
      case OPENBLOCK:
 	++block_lvl;
	break;
      case CLOSEBLOCK:
 	--block_lvl;
	break;
      case ARRAYOPEN:
 	++array_lvl;
 	if (state == int_fmt_arg_null)
 	  start_array_lvl = array_lvl;
	break;
      case ARRAYCLOSE:
 	--array_lvl;
	break;
      default:
	break;
      } /* switch (M_TOK(m)) */
  }
  return TRUE;
}

int is_long_long_int_fmt_arg (MESSAGE_STACK messages, int expr_start_idx,
		     int stack_start_idx, int stack_end_idx) {
  int i, n_th_arg;
  int paren_lvl, block_lvl, array_lvl, start_paren_lvl, start_array_lvl;
  MESSAGE *m;
  int fmt_str_length;
  char *fmt_char_ptr;
  enum {
    llint_fmt_arg_null,
    llint_fmt_arg_expr,
    llint_fmt_arg_separator,
    llint_fmt_arg_literal
  } state;

  paren_lvl = block_lvl = array_lvl = n_th_arg = 0;
  start_paren_lvl = start_array_lvl = -1;
  for (i = expr_start_idx, state = llint_fmt_arg_null; i <= stack_start_idx; i++) {

    m = messages[i];
    if (M_ISSPACE(m)) continue;

    switch (M_TOK(m))
      {
      case LABEL:
	if ((state == llint_fmt_arg_null) || (state == llint_fmt_arg_separator))
	  state = llint_fmt_arg_expr;
	break; 
      case ARGSEPARATOR:
	if (start_paren_lvl != -1) {
	  if ((paren_lvl == start_paren_lvl) && !block_lvl && !array_lvl) {
	    state = llint_fmt_arg_separator;
	    --n_th_arg;
	  }
	} else {
	  if (!paren_lvl && !block_lvl && !array_lvl) {
	    state = llint_fmt_arg_separator;
	    --n_th_arg;
	  }
	}
	break;
      case LITERAL:
 	if (state == llint_fmt_arg_separator) {
 	  if ((fmt_str_length = is_printf_fmt (M_NAME(m))) != 0) {
	    int __n;
	    /*
	     *  n_th_arg is -1 for the first argument, -2 for the 
	     *  second, -3 for the third, ....
	     */
  	    if ((n_th_arg == -1) && is_long_long_int_fmt(M_NAME(m))) {
  	      return TRUE;
 	    } else {
	      fmt_char_ptr = index (M_NAME(m), '%');

 	      for (__n = -2; ; __n--) {
		if ((fmt_char_ptr = index (fmt_char_ptr + 1, '%')) == NULL)
		  return FALSE;

		/* Check for literal "%%" */
		if (fmt_char_ptr[1] == '%')
		  continue;

		if (__n == n_th_arg)
		  break;
	      }

	      return is_long_long_int_fmt (fmt_char_ptr) && (__n == n_th_arg);

	    }
	  }
 	} else {
	  state = llint_fmt_arg_expr;
	}
	break;
      case OPENPAREN:
 	++paren_lvl;
 	if (state == llint_fmt_arg_null) {
 	  start_paren_lvl = paren_lvl;
	  state = llint_fmt_arg_expr;
	} else {
	  if (start_paren_lvl != -1) {
	    if ((paren_lvl < start_paren_lvl) && (state == llint_fmt_arg_expr)) {
	      return FALSE;
	    }
	  } else {
	    int __prev_idx;
	    if ((__prev_idx = __ctalkPrevLangMsg (messages, i, stack_start_idx)) 
		!= ERROR) {
	      /*
	       *  Function call syntax.
	       */
	      if (M_TOK(messages[__prev_idx]) == LABEL)
		return FALSE;
	    }
	  }
	}
	break;
      case CLOSEPAREN:
 	--paren_lvl;
	if (state == llint_fmt_arg_separator)
	  state = llint_fmt_arg_expr;
	break;
      case OPENBLOCK:
 	++block_lvl;
	break;
      case CLOSEBLOCK:
 	--block_lvl;
	break;
      case ARRAYOPEN:
 	++array_lvl;
 	if (state == llint_fmt_arg_null)
 	  start_array_lvl = array_lvl;
	break;
      case ARRAYCLOSE:
 	--array_lvl;
	break;
      default:
	break;
      } /* switch (M_TOK(m)) */
  }
  return TRUE;
}

int is_char_ptr_fmt (char *s) {
  int i;
  char prev_char = 0;
  char buf[MAXMSG];
  enum {
    char_ptr_fmt_arg_null,
    char_ptr_fmt_arg_percent,
    char_ptr_fmt_arg_precision,
    char_ptr_fmt_arg_length_mod,
    char_ptr_fmt_fmt_char
  } state;

  strcpy (buf, s);
  if (*buf == '"') TRIM_LITERAL(buf);
  if (*buf == '\'') TRIM_CHAR_BUF(buf);

  for (i = 0, state = char_ptr_fmt_arg_null; buf[i]; i++) {

    if (state == char_ptr_fmt_arg_null) {
      if (!IS_STDARG_FMT_CHAR (buf[i])) continue;
      if (iscntrl ((int) buf[i])) continue;
    }

    switch (buf[i])
      {
      case '%':
	switch (state)
	  {
	  case char_ptr_fmt_arg_null:
	    state = char_ptr_fmt_arg_percent;
	    break;
	  case char_ptr_fmt_arg_percent:
	    state = char_ptr_fmt_arg_null;
	    break;
	    /* Any other state means we've reached a following '%' format. */
	  default:
	    return FALSE;
	  }

	break;
      case '.':
	if (state == char_ptr_fmt_arg_percent)
	  state = char_ptr_fmt_arg_precision;
	break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	if (state == char_ptr_fmt_arg_percent)
	  state = char_ptr_fmt_arg_precision;
	break;
      case 'l':
	if ((state == char_ptr_fmt_arg_percent) ||
	    (state == char_ptr_fmt_arg_precision))
	  state = char_ptr_fmt_arg_length_mod;
	break;
      case 's':
      case 'S':
	if (state == char_ptr_fmt_arg_percent) {
	  if (prev_char == '%')
	    return TRUE;
	}
	if ((state == char_ptr_fmt_arg_precision) ||
	    (state == char_ptr_fmt_arg_length_mod)) {
	  return TRUE;
	} else {
	  if (buf[i+1] == '\0') return FALSE;
	}
	break;
      }
    if (!isspace ((int)buf[i]))
      prev_char = buf[i];
  }
  return FALSE;
}

int is_ptr_fmt (char *s) {
  int i;
  char buf[MAXMSG];
  enum {
    ptr_fmt_arg_null,
    ptr_fmt_arg_percent,
    ptr_fmt_arg_pound,
  } state;

  strcpy (buf, s);
  if (*buf == '"') TRIM_LITERAL(buf);
  if (*buf == '\'') TRIM_CHAR_BUF(buf);

  for (i = 0, state = ptr_fmt_arg_null; buf[i]; i++) {

    if (state == ptr_fmt_arg_null) {
      if (!IS_STDARG_FMT_CHAR (buf[i])) continue;
      if (iscntrl ((int) buf[i])) continue;
    }

    switch (buf[i])
      {
      case '%':
	switch (state)
	  {
	  case ptr_fmt_arg_null:
	    state = ptr_fmt_arg_percent;
	    break;
	  case ptr_fmt_arg_percent:
	    state = ptr_fmt_arg_null;
	    break;
	    /* Any other state means we've reached a following '%' format. */
	  default:
	    return FALSE;
	    break;
	  }
	break;
      case 'p':
	if (state == ptr_fmt_arg_percent)
	  return TRUE;
	break;
      case '#':
	if (state == ptr_fmt_arg_percent)
	  state = ptr_fmt_arg_pound;
	break;
      case 'x':
	if (state == ptr_fmt_arg_pound)
	  return TRUE;
	break;
      }
  }
  return FALSE;
}

int is_char_fmt (char *s) {
  int i;
  char buf[MAXMSG];
  char prev_char = 0;
  enum {
    char_fmt_arg_null,
    char_fmt_arg_percent,
    char_fmt_arg_precision,
    char_fmt_arg_length_mod,
    char_fmt_fmt_char
  } state;

  strcpy (buf, s);
  if (*buf == '"') TRIM_LITERAL(buf);
  if (*buf == '\'') TRIM_CHAR_BUF(buf);

  for (i = 0, state = char_fmt_arg_null; buf[i]; i++) {

    if (state == char_fmt_arg_null) {
      if (!IS_STDARG_FMT_CHAR (buf[i])) continue;
      if (iscntrl ((int) buf[i])) continue;
    }

    switch (buf[i])
      {
      case '%':
	switch (state)
	  {
	  case char_fmt_arg_null:
	    state = char_fmt_arg_percent;
	    break;
	  case char_fmt_arg_percent:
	    state = char_fmt_arg_null;
	    break;
	    /* Any other state means we've reached a following '%' format. */
	  default:
	    return FALSE;
	    break;
	  }
	break;
      case '.':
	if (state == char_fmt_arg_precision)
	  continue;
	if (state == char_fmt_arg_percent)
	  state = char_fmt_arg_precision;
	break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	if (state == char_fmt_arg_percent)
	  state = char_fmt_arg_precision;
	break;
      case 'l':
	if ((state == char_fmt_arg_percent) ||
	    (state == char_fmt_arg_precision))
	  state = char_fmt_arg_length_mod;
	break;
      case 'c':
	if (state == char_fmt_arg_percent) {
	  if (prev_char == '%') {
	    return TRUE;
	  }
	}
	break;
      }
    if (!isspace((int)buf[i]))
      prev_char = buf[i];
  }
  return FALSE;
}

int is_int_fmt (char *s) {
  int i;
  char buf[MAXMSG];
  enum {
    int_fmt_arg_null,
    int_fmt_arg_percent,
    int_fmt_arg_precision,
    int_fmt_arg_length_mod,
    int_fmt_fmt_char
  } state;

  strcpy (buf, s);
  if (*buf == '"') TRIM_LITERAL(buf);
  if (*buf == '\'') TRIM_CHAR_BUF(buf);

  for (i = 0, state = int_fmt_arg_null; buf[i]; i++) {

    if (state == int_fmt_arg_null) {
      if (!IS_STDARG_FMT_CHAR (buf[i])) continue;
      if (iscntrl ((int) buf[i])) continue;
    }
    switch (buf[i])
      {
      case '%':
	switch (state)
	  {
	  case int_fmt_arg_null:
	    state = int_fmt_arg_percent;
	    break;
	  case int_fmt_arg_percent:
	    state = int_fmt_arg_null;
	    break;
	    /* Any other state means we've reached a following '%' format. */
	  default:
	    return FALSE;
	    break;
	  }
	break;
      case '.':
	if (state == int_fmt_arg_percent)
	  state = int_fmt_arg_precision;
	break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	if (state == int_fmt_arg_percent)
	  state = int_fmt_arg_precision;
	break;
      case 'l':
	if ((state == int_fmt_arg_percent) ||
	    (state == int_fmt_arg_precision))
	  state = int_fmt_arg_length_mod;
	break;
      case 'i':
      case 'd':
      case 'o':
      case 'u':
      case 'x':
      case 'X':
	if ((state == int_fmt_arg_percent) ||
	    (state == int_fmt_arg_precision))
	  return TRUE;
	break;
      }
  }
  return FALSE;
}

int is_long_long_int_fmt (char *s) {
  int i;
  char buf[MAXMSG];
  enum {
    llint_fmt_arg_null,
    llint_fmt_arg_percent,
    llint_fmt_arg_precision,
    llint_fmt_arg_length_mod,
    llint_fmt_fmt_char
  } state;

  strcpy (buf, s);
  if (*buf == '"') TRIM_LITERAL(buf);
  if (*buf == '\'') TRIM_CHAR_BUF(buf);

  for (i = 0, state = llint_fmt_arg_null; buf[i]; i++) {

    if (state == llint_fmt_arg_null) {
      if (!IS_STDARG_FMT_CHAR (buf[i])) continue;
      if (iscntrl ((int) buf[i])) continue;
    }
    switch (buf[i])
      {
      case '%':
	switch (state)
	  {
	  case llint_fmt_arg_null:
	    state = llint_fmt_arg_percent;
	    break;
	  case llint_fmt_arg_percent:
	    state = llint_fmt_arg_null;
	    break;
	    /* Any other state means we've reached a following '%' format. */
	  default:
	    return FALSE;
	    break;
	  }
	break;
      case '.':
	if (state == llint_fmt_arg_percent)
	  state = llint_fmt_arg_precision;
	break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	if (state == llint_fmt_arg_percent)
	  state = llint_fmt_arg_precision;
	break;
      case 'L':
	if ((state == llint_fmt_arg_percent) ||
	    (state == llint_fmt_arg_precision))
	  state = llint_fmt_arg_length_mod;
	break;
      case 'l':
	if ((state == llint_fmt_arg_percent) ||
	    (state == llint_fmt_arg_precision)) {
	  if (buf[i+1] == 'l') {
	    ++i;
	    state = llint_fmt_arg_length_mod;
	  }
	}
	break;
      case 'i':
      case 'd':
      case 'o':
      case 'u':
      case 'x':
      case 'X':
	if (state == llint_fmt_arg_length_mod)
	  return TRUE;
	break;
      }
  }
  return FALSE;
}

int is_printf_fmt_msg (MESSAGE_STACK messages, int fmt_tok_idx,
		       int stack_start_idx, int stack_end_idx) {
  return is_printf_fmt (M_NAME(messages[fmt_tok_idx]));
}

int is_function_call (MESSAGE_STACK messages, int fn_label_idx,
		      int stack_start_idx, int stack_end_idx) {
  int open_paren_idx, close_paren_idx;
  if ((open_paren_idx = __ctalkNextLangMsg (messages, fn_label_idx,
					    stack_end_idx)) == ERROR)
    return FALSE;
  if (M_TOK(messages[open_paren_idx]) != OPENPAREN)
    return FALSE;
  if ((close_paren_idx = __ctalkMatchParen (messages, 
					    open_paren_idx,
					    stack_end_idx)) == ERROR)
    return FALSE;
  return TRUE;
}

int is_struct_or_union_expr (MESSAGE_STACK messages, 
			     int expr_start_label_idx,
			     int stack_start_idx,
			     int stack_end_idx) {
  int next_tok_idx, close_paren_idx;
  if ((next_tok_idx = __ctalkNextLangMsg(messages,
					 expr_start_label_idx,
					 stack_end_idx)) != ERROR) {
    if ((M_TOK(messages[next_tok_idx]) == PERIOD) ||
	(M_TOK(messages[next_tok_idx]) == DEREF)) {
      return TRUE;
    }
    /*
     *  Also match <fn> () -> or <fn> () .
     */
    if (M_TOK(messages[next_tok_idx]) == OPENPAREN) {
      if ((close_paren_idx = __ctalkMatchParen (messages,
						next_tok_idx,
						stack_end_idx)) != ERROR) {
	if ((next_tok_idx = __ctalkNextLangMsg(messages,
					       close_paren_idx,
					       stack_end_idx)) != ERROR) {
	  if ((M_TOK(messages[next_tok_idx]) == PERIOD) ||
	      (M_TOK(messages[next_tok_idx]) == DEREF)) {
	    return TRUE;
	  }
	}
      }
    }
  }
  return FALSE;
}

/*
 *  Returns stack index of closing bracket, 0 if no subscript,
 *  -1 if a bracket mismatch.
 */
int is_subscript_expr (MESSAGE_STACK messages, 
		       int expr_start_label_idx,
		       int stack_end_idx) {
  int next_tok_idx, i, n_brackets;
  if ((next_tok_idx = __ctalkNextLangMsg(messages,
					 expr_start_label_idx,
					 stack_end_idx)) != ERROR) {
    if (M_TOK(messages[next_tok_idx]) == ARRAYOPEN) {
      n_brackets = 1;
      for (i = next_tok_idx; i > stack_end_idx; i--) {
	if (M_TOK(messages[i]) == ARRAYCLOSE) {
	  if (--n_brackets == 0) {
	    return i;
	  }
	}
      }
      return ERROR;
    }
  }
  return FALSE;
}

/*
 *  Returns stack index of terminal label, 0 if no struct or deref 
 *  operator, -1 on error.
 */
int is_struct_expr_b (MESSAGE_STACK messages, 
		       int expr_start_label_idx,
		       int stack_end_idx) {
  int next_tok_idx, next_tok_idx_2, i, n_derefs;
  n_derefs = 0;
  i = expr_start_label_idx;
  while ((next_tok_idx = __ctalkNextLangMsg(messages,
					    i,
					    stack_end_idx)) != ERROR) {
    if ((M_TOK(messages[next_tok_idx]) == PERIOD) ||
	(M_TOK(messages[next_tok_idx]) == DEREF)) {
      ++n_derefs;
      i = next_tok_idx;
      continue;
    }
    if (M_TOK(messages[next_tok_idx]) == LABEL) {
      if ((next_tok_idx_2 = 
	   __ctalkNextLangMsg(messages,
			      next_tok_idx,
			      stack_end_idx)) != ERROR) {
	if ((M_TOK(messages[next_tok_idx_2]) == PERIOD) ||
	    (M_TOK(messages[next_tok_idx_2]) == DEREF)) {
	  i = next_tok_idx;
	  continue;
	} else {
	  if (n_derefs) {
	    return next_tok_idx;
	  } else {
	    return 0;
	  }
	}
      } else {
	if (n_derefs) {
	  return next_tok_idx;
	} else {
	  return 0;
	}
      }
    }
  }
  return FALSE;
}

/*
 *  Returns stack index of terminal label, 0 if no struct or deref 
 *  operator, -1 on error.
 */
int is_struct_expr_b_rev (MESSAGE_STACK messages, 
		       int expr_end_label_idx,
		       int stack_start_idx) {
  int prev_tok_idx, prev_tok_idx_2, i, n_derefs;
  n_derefs = 0;
  i = expr_end_label_idx;
  while ((prev_tok_idx = __ctalkPrevLangMsg(messages,
					    i,
					    stack_start_idx)) != ERROR) {
    if ((M_TOK(messages[prev_tok_idx]) == PERIOD) ||
	(M_TOK(messages[prev_tok_idx]) == DEREF)) {
      ++n_derefs;
      i = prev_tok_idx;
      continue;
    }
    if (M_TOK(messages[prev_tok_idx]) == LABEL) {
      if ((prev_tok_idx_2 = 
	   __ctalkPrevLangMsg(messages,
			      prev_tok_idx,
			      stack_start_idx)) != ERROR) {
	if ((M_TOK(messages[prev_tok_idx_2]) == PERIOD) ||
	    (M_TOK(messages[prev_tok_idx_2]) == DEREF)) {
	  i = prev_tok_idx;
	  continue;
	} else {
	  if (n_derefs) {
	    return prev_tok_idx;
	  } else {
	    return 0;
	  }
	}
      } else {
	if (n_derefs) {
	  return prev_tok_idx;
	} else {
	  return 0;
	}
      }
    }
  }
  return FALSE;
}

/*
 *  Returns the index of the argument, or -1;
 *
 *  Args except for printf format args.
 */
int obj_expr_is_arg (MESSAGE_STACK messages, int expr_start_idx,
		     int stack_start_idx, int stack_end_idx,
		     int *fn_label_idx) {
  int arg_idx = -1;
  int i;
  int n_parens = 0;

  for (i = expr_start_idx; i <= stack_start_idx; i++) {
    switch (M_TOK(messages[i]))
      {
      case ARGSEPARATOR:
	++arg_idx;
	break;
      case OPENPAREN:
	--n_parens;
	if (n_parens < 0) {
	  ++arg_idx;
	  if ((*fn_label_idx = 
 	       __ctalkPrevLangMsg (messages, i, stack_start_idx)) != ERROR) {
	    if (M_TOK(messages[*fn_label_idx]) == LABEL)
	      return arg_idx;

	    /* Just an open paren starting an argument, not the arglist -
	       reset. */
	    if (M_TOK(messages[*fn_label_idx]) == ARGSEPARATOR) {
	      --arg_idx;
	      n_parens = 0;
	    }
	  }
	}
	break;
      case CLOSEPAREN:
	++n_parens;
	break;
      case SEMICOLON:
	return ERROR;
	break;
      case OPENBLOCK:
	return ERROR;
	break;
      case LITERAL:
	if (is_printf_fmt (M_NAME(messages[i])))
	  return ERROR;
	break;
      }
  }
  return ERROR;
}

int is_double_fmt (char *s) {
  int i;
  char buf[MAXMSG];
  enum {
    double_fmt_arg_null,
    double_fmt_arg_percent,
    double_fmt_arg_precision,
    double_fmt_arg_length_mod,
    double_fmt_fmt_char
  } state;

  strcpy (buf, s);
  if (*buf == '"') TRIM_LITERAL(buf);
  if (*buf == '\'') TRIM_CHAR_BUF(buf);

  for (i = 0, state = double_fmt_arg_null; buf[i]; i++) {

    if (state == double_fmt_arg_null) {
      if (!IS_STDARG_FMT_CHAR (buf[i])) continue;
      if (iscntrl ((int) buf[i])) continue;
    }

    switch (buf[i])
      {
      case '%':
	switch (state)
	  {
	  case double_fmt_arg_null:
	    state = double_fmt_arg_percent;
	    break;
	  case double_fmt_arg_percent:
	    state = double_fmt_arg_null;
	    break;
	    /* Any other state means we've reached a following '%' format. */
	  default:
	    return FALSE;
	    break;
	  }
	break;
      case '.':
	if (state == double_fmt_arg_precision)
	  continue;
	if (state == double_fmt_arg_percent)
	  state = double_fmt_arg_precision;
	break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	if (state == double_fmt_arg_precision)
	  continue;
	if (state == double_fmt_arg_percent)
	    state = double_fmt_arg_precision;
	break;
      case 'l':
	if ((state == double_fmt_arg_percent) ||
	    (state == double_fmt_arg_precision)) {
	  state = double_fmt_arg_length_mod;
	} 
	break;
      case 'e':
      case 'E':
      case 'f':
      case 'F':
      case 'g':
      case 'G':
	if ((state == double_fmt_arg_percent) ||
	    (state == double_fmt_arg_precision) ||
	    (state == double_fmt_arg_length_mod))
	  return TRUE;
	break;
      }
  }
  return FALSE;
}

int rcvr_is_start_of_expr (MESSAGE_STACK messages, int idx, int stack_start) {
  int i;
  for (i = idx; i <= stack_start; i++) {
    if (M_ISSPACE(messages[i]))
      continue;
    if ((M_TOK(messages[i]) == SEMICOLON) ||
	(M_TOK(messages[i]) == COLON) ||
	(M_TOK(messages[i]) == OPENBLOCK) ||
	(M_TOK(messages[i]) == ARGSEPARATOR) ||
	(M_TOK(messages[i]) == ARRAYOPEN))
      return TRUE;
    if ((M_TOK(messages[i]) != LABEL) &&
	(M_TOK(messages[i]) != METHODMSGLABEL))
      return FALSE;
  }
  return TRUE;
}

int have_label_terminator (MESSAGE_STACK messages, int start, int end) {
  int i;
  for (i = start; i <= end; i++) {
    if (M_TOK(messages[i]) == COLON)
      return i;
  }
  return -1;
}


/*
 *  Only recognzes a "(...) <label>" construct from within the parentheses,
 *  checks for C keywords also; i.e., control structure keywords like 
 *  "break," "return," and, "goto," "if," "while," and so on and on, are 
 *  valid also.
 *
 *  Also check for an opening parenthesis preceded by a control 
 *  keyword, which is a C construct; e.g., 
 *
 *    <ctrl keyword> (...) <label>
 *
 *  Returns the stack index of the closing paren on the subexpression if
 *  true, 0 if false.
 *
 */
int is_in_rcvr_subexpr (MESSAGE_STACK messages, int idx, 
			int stack_start, int stack_end) {
  int i, lasttok = -1, lasttok_idx = -1, /* Avoid warnings */
    n_parens;
  MESSAGE *m;
  n_parens = 0;
  for (i = idx; i > stack_end; i--) {
    m = messages[i];
    if (M_ISSPACE(m))
      continue;
    switch (M_TOK(m))
      {
      case OPENPAREN:
	if ((lasttok == CLOSEPAREN) && 
	    (n_parens < 0)) {
	  return FALSE;
	} else {
	  n_parens += 1;
	}
	break;
      case CLOSEPAREN:
	n_parens -= 1;
	break;
      case SEMICOLON:
	if ((lasttok == CLOSEPAREN) && 
	    (n_parens < 0)) {
	  return FALSE;
	}
	if (IS_CONSTANT_TOK(lasttok))
	  return FALSE;
	if ((lasttok == LABEL) &&
	    (n_parens == 0))
	  return FALSE;
	break;
      case CLOSEBLOCK:
      case OPENBLOCK:
      case ARGSEPARATOR:
      case ARRAYOPEN:
      case ARRAYCLOSE:
	if ((lasttok == CLOSEPAREN) && 
	    (n_parens < 0)) {
	  return FALSE;
	}
	break;
      case LABEL:
	if ((lasttok == CLOSEPAREN)  &&
	    (n_parens < 0)) {
	  if (!is_c_keyword (M_NAME(m))) {
	    /*
	     *  Check if the opening paren is preceded by
	     *  a control keyword, which is a C construct; e.g.
	     *
	     *   <ctrl_keyword> (....) <message m>
	     */
	    int i_2, n_parens_2;
	    n_parens_2 = 0;
	    for (i_2 = lasttok_idx; i_2 <= stack_start; i_2++) {
	      if (M_ISSPACE(messages[i_2]))
		continue;
	      if (M_TOK(messages[i_2]) == CLOSEPAREN)
		--n_parens_2;
	      if (M_TOK(messages[i_2]) == OPENPAREN)
		++n_parens_2;
	      /*
	       *  Sanity check so we don't need to 
	       *  look at the entire input.
	       */
	      if ((M_TOK(messages[i_2]) == OPENBLOCK) ||
		  (M_TOK(messages[i_2]) == CLOSEBLOCK))
		  return FALSE;
	      if (n_parens_2 == 0) {
		int prev_tok_idx_2;
		if ((prev_tok_idx_2 = __ctalkPrevLangMsg 
		     (messages, i_2, stack_start)) != ERROR) {
		  if (is_ctrl_keyword (M_NAME(messages[prev_tok_idx_2]))) {
		    return FALSE;
		  } else {
		    MESSAGE *m_2;
		    int i_3;
		    /*
		     *  Check for a struct expression alone within the
		     *  parentheses.
		     */
		    for (i_3 = i_2; i_3 >= lasttok_idx; i_3--) {
		      m_2 = messages[i_3];
		      if (!M_ISSPACE(m_2)) {
			if ((m_2 -> attrs & RCVR_TOK_IS_C_STRUCT_EXPR) ||
			    (M_TOK(m_2) == INCREMENT) ||
			    (M_TOK(m_2) == DECREMENT) ||
			    (M_TOK(m_2) == OPENPAREN) ||
			    (M_TOK(m_2) == CLOSEPAREN)) {
			  continue;
			} else {
			  return lasttok_idx;
			}
		      }
		    }
		    return FALSE;
		  }
		}
	      }
	    }
	    if (lasttok == SEMICOLON) {
	      if (is_c_keyword (M_NAME(m))) {
		return FALSE;
	      }
	    }
	  } else {
	    return FALSE;
	  }
	}
	break;
      }
    lasttok = M_TOK(m);
    lasttok_idx = i;
  }
  return FALSE;
}

int is_fn_style_label (MESSAGE_STACK messages, int idx, int stack_top) {
  int next_tok;
  if (M_TOK(messages[idx]) == LABEL) {
    if ((next_tok = __ctalkNextLangMsg (messages, idx, stack_top)) 
	!= ERROR) {
      if (M_TOK(messages[next_tok]) == OPENPAREN) {
	return TRUE;
      }
    }
  }
  return FALSE;
}

int is_first_ctrlblk_pred_tok (MESSAGE_STACK messages, int idx, int stack_start) {
  int i;
  for (i = idx+1; i <= stack_start; i++) {
    if (M_ISSPACE(messages[i]))
      continue;
    if (M_TOK(messages[i]) != OPENPAREN)
      break;
  }
  if (is_ctrl_keyword (M_NAME(messages[i])))
    return TRUE;
  return FALSE;
}

#define IS_LEADING_CTRLBLK_UNARY_MATH_OP(tok) \
   ((tok == INCREMENT) || \
    (tok == DECREMENT) || \
    (tok == BIT_COMP) ||  \
    (tok == LOG_NEG) ||	  \
    (tok == AMPERSAND) || \
    (tok == ASTERISK))


/*
 *  Recognizes <tok><label> or <tok><(+><label>
 *  Skips leading OPENPAREN tokens.  Used in method_call *only*.
 *  This is basically more convenient than scanback_other ().
 *  Returns the stack index of the non-OPENPAREN token before
 *  the label, -1 on error.
 */
int is_leading_prefix_op (MESSAGE_STACK messages, int label_idx, 
			  int stack_start) {
  int j;
  for (j = label_idx + 1; j <= stack_start; j++) {
    if (M_ISSPACE(messages[j]))
	continue;
    if (M_TOK(messages[j]) == OPENPAREN) {
      continue;
    } else {
      return j;
    }
  }
  return -1;
}

/*
 *  Recognizes <var_end_tok><postfix_tok> or 
 *  <var_end_tok><)+><postfix_tok>
 *  Skips trailing CLOSEPAREN tokens.  Used in method_call *only*.
 *  This is basically more convenient than scanforward_other ().
 *  Returns the stack index of the non-CLOSEPAREN token after the var
 *  end, -1 on error.
 */
int is_trailing_postfix_op (MESSAGE_STACK messages, int last_var_tok_idx, 
			    int stack_end) {
  int j;
  for (j = last_var_tok_idx - 1; j >= stack_end; j--) {
    if (M_ISSPACE(messages[j]))
	continue;
    if (M_TOK(messages[j]) == CLOSEPAREN) {
      continue;
    } else {
      return j;
    }
  }
  return -1;
}

int is_leading_ctrlblk_prefix_op_tok (MESSAGE_STACK messages, int idx, 
				      int stack_start) {
  int i;
  for (i = idx+1; i <= stack_start; i++) {
    if (M_ISSPACE(messages[i]))
      continue;
    if ((M_TOK(messages[i]) != OPENPAREN) && 
	(!IS_LEADING_CTRLBLK_UNARY_MATH_OP(M_TOK(messages[i]))))
      break;
  }
  if (is_ctrl_keyword (M_NAME(messages[i])))
    return TRUE;
  return FALSE;
}

int prefix_op_is_sizeof (MESSAGE_STACK messages, int idx, int stack_start_idx) {
  int i;
  int have_prefix_attr;
  MESSAGE *m;
  enum {
    prefix_op_is_sizeof_null,
    prefix_op_is_sizeof_openparen/*  , */
  } state;

  state = prefix_op_is_sizeof_null;
  have_prefix_attr = FALSE;
  for (i = idx; i <= stack_start_idx; i++) {
    if (M_ISSPACE(messages[i]))
      continue;
    m = messages[i];
    if ((M_TOK(m) == SEMICOLON) ||
	(M_TOK(m) == OPENBLOCK))
      break;
    if (m -> attrs & TOK_IS_PREFIX_OPERATOR) {
      if (!have_prefix_attr)
	have_prefix_attr = TRUE;
      switch (state)
	{
	case prefix_op_is_sizeof_null:
	  if (M_TOK(m) == OPENPAREN)
	    state = prefix_op_is_sizeof_openparen;
	  break;
	case prefix_op_is_sizeof_openparen:
	  if ((M_TOK(m) == LABEL) || 
	      (M_TOK(m) == SIZEOF)) {
	    if (str_eq (M_NAME(m), "sizeof"))
	      return i;
	  }
	  break;
	}
    } else {
      if (have_prefix_attr)
	return 0;
    }
  }
  return 0;
}

int method_contains_argblk (MESSAGE_STACK messages, int method_open_block,
			    int method_close_block) {
  int i, lookahead;
  int prev_idx = -1, prev_idx_2 = -1, prev_idx_3 = -1;

  for (i = method_open_block; i >= method_close_block; i--) {

      if (M_ISSPACE(messages[i]))
	continue;

    for (lookahead = i - 1; lookahead >= method_close_block; lookahead--) {

      if (M_ISSPACE(messages[lookahead]))
	continue;

      if ((M_TOK(messages[i]) == LABEL) && 
	  (M_TOK(messages[lookahead]) == OPENBLOCK)) {

	/* expressions like, "else {" */
	if (is_c_keyword (messages[i] -> name))
	  break;
	/* expressions like "struct {" or "struct <name> {" */
	if (prev_idx != -1) {
	  if (is_c_keyword (messages[prev_idx] -> name))
	    break;
	}
	if (prev_idx_3 != -1) {
	  if (is_c_keyword (messages[prev_idx_3] -> name))
	    break;
	}
	if (prev_idx_2 != -1) {
	  if (is_c_keyword (messages[prev_idx_2] -> name))
	    break;
	}
	if (prev_idx != -1) {
	  if (is_c_keyword (messages[prev_idx] -> name))
	    break;
	}

	return TRUE;
      } else {
	break;
      }

    }

    prev_idx_3 = prev_idx_2;
    prev_idx_2 = prev_idx;
    prev_idx = i;
  }

  return FALSE;
}

