/* $Id: rtnewobj.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2014  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/* 
 * Run time object constructors, destructors, and basic 
 * garbage collection.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "lex.h"
#include "typeof.h"

void __call_delete_object_internal (OBJECT *);

extern OBJECT *__ctalk_classes;

static inline OBJECT *__class_obj_by_name (char *name) {
  OBJECT *o;
  for (o = __ctalk_classes; o; o = o -> next) 
    if (str_eq (name, o -> __o_name))
      return o;
  return NULL;
}

OBJECT *__ctalkCreateObjectInit (char *name, char *class, 
				    char *superclass, int scope,
				    char *value) {
  static OBJECT *o, *o_value, *o_superclass;
  static char *o_value_ptr, *o_value_value_ptr;
  Boolean is_class_object = False;

  if (!strcmp (class, "Class"))
    is_class_object = True;

  if (superclass) {
    o_superclass = __ctalkGetClass (superclass);
  } else {
    o_superclass = NULL;
  }

  o = (OBJECT *)__xalloc (sizeof (struct _object));
  o_value = (OBJECT *)__xalloc (sizeof (struct _object));
  o_value_ptr = (char *)__xalloc (strlen (NULLSTR) + 1);
  o_value_value_ptr = (char *)__xalloc (strlen (NULLSTR) + 1);

  o -> __o_value = o_value_ptr;
  o_value -> __o_value = o_value_value_ptr;
  o -> instancevars = o_value;
  o_value -> __o_p_obj = o;

#ifndef __sparc
  o -> sig = 0xd3d3d3;
#else
  strcpy (o -> sig, "OBJECT");
#endif
  if (name)
    strcpy (o -> __o_name, name);
  else
    strcpy (o -> __o_name, NULLSTR);
  strcpy (o -> __o_classname, class);
  o -> __o_class = __ctalkClassObject (o);
  if (superclass)
    strcpy (o -> __o_superclassname, superclass);
  o -> __o_superclass = o_superclass;
  o -> scope = scope;
#ifndef __sparc__
  o_value -> sig = 0xd3d3d3;
#else
  strcpy (o_value -> sig, "OBJECT");
#endif
  strcpy (o_value -> __o_name, "value");
  o_value -> attrs = OBJECT_IS_VALUE_VAR;
  strcpy (o_value -> __o_classname, 
	  ((is_class_object) ? name: class));
  o_value -> __o_class = __ctalkClassObject (o);
  if (superclass)
    strcpy (o_value -> __o_superclassname, superclass);
  o_value -> __o_superclass = o_superclass;
  o_value -> scope = scope;
  strcpy (o_value_ptr, NULLSTR);
  strcpy (o_value_value_ptr, NULLSTR);

  if (value)
    __ctalkSetObjectValueVar (o, value);

  return o;
}

OBJECT *__ctalkCreateObject (char *name, char *class, char *superclass,
			       int scope) {

  static OBJECT *o;
  static char *o_value_ptr;

  o = (OBJECT *)__xalloc (sizeof (struct _object));
  o_value_ptr = (char *)__xalloc (strlen (NULLSTR) + 1);

  o -> __o_value = o_value_ptr;

#ifndef __sparc__
  o -> sig = 0xd3d3d3;
#else
  strcpy (o -> sig, "OBJECT");
#endif
  strcpy (o -> __o_name, name);
  strcpy (o -> __o_classname, class);
  o -> __o_class = __ctalkClassObject (o);
  if (superclass) {
    strcpy (o -> __o_superclassname, superclass);
    o -> __o_superclass = __ctalkGetClass (superclass);
  }
  o -> scope = scope;
  strcpy (o -> __o_value, NULLSTR);

  return o;
  
}

/* From ctalklib - 
   The name of the new object is on the argument stack.
   The new object's class is the class of the receiver
   object. 
*/
extern int primitive_method_call_attrs; /* Declared in rt_methd.c */
OBJECT *__ctalk_new_object (ARG **args) {

  OBJECT *rcvr;

  rcvr = __ctalk_receiver_pop ();
  if (IS_OBJECT (rcvr)) {
    __ctalk_receiver_push (rcvr);
  } else {
    _warning ("__ctalk_new_object: Invalid receiver object.\n");
    return NULL;
  }

  if (!(primitive_method_call_attrs & METHOD_SUPER_ATTR)) {
    strcpy (ARG_CLASSNAME(args[0]), rcvr -> __o_name);
    strcpy (ARG_SUPERCLASSNAME(args[0]), rcvr -> __o_superclassname);
  }
  __ctalkInstanceVarsFromClassObject (ARG_OBJECT(args[0]));
  __objRefCntSet (OBJREF(args[0]->obj), 0);
  __ctalk_dictionary_add (ARG_OBJECT(args[0]));
  
  return ARG_OBJECT(args[0]);
}

/*
 *   Set an object's value, making sure that the value member's
 *   buffer is large enough for the value.
 */

int __ctalkSetObjectValue (OBJECT *__o, char *__value) {

  if (!__value) {
    free (__o -> __o_value);
    __o -> __o_value = strdup ("0x0");
    return SUCCESS;
  }

  if (strlen (__value) >= strlen (__o -> __o_value)) {
    free (__o -> __o_value);
    if ((__o -> __o_value = (char *)calloc (strlen (__value) + 1, sizeof (char)))
	== NULL) {
      _error ("__ctalkSetObjectValue (): %s.\n", strerror (errno));
      return ERROR;
    }
  }
  strcpy (__o -> __o_value, __value);

  return SUCCESS;
}

/*
 *  This should replace the function above.
 */

static char *new_value_ptr;
static char *new_value_value_ptr;

int __ctalkSetObjectValueVar (OBJECT *__o, char *__value) {

  OBJECT *__o_value_obj;

  if (!IS_OBJECT(__o))
    return ERROR;

  if (__o -> __o_value) {
    if (strlen (__value) > strlen (__o -> __o_value)) {
      new_value_ptr = __o->__o_value;
      new_value_ptr = realloc (new_value_ptr, strlen (__value) + 1);
      memset (new_value_ptr, '\0', strlen (__value) + 1);
      __o -> __o_value = new_value_ptr;
    }
    strcpy (__o -> __o_value, __value);
  } else {
    __o -> __o_value = strdup (__value);
  }

  if ((__o_value_obj = __ctalkGetInstanceVariable (__o, "value", FALSE)) != NULL) {
     if (strlen (__value) >= strlen (__o_value_obj -> __o_value)) {
       new_value_value_ptr = __o_value_obj->__o_value;
       new_value_value_ptr = realloc (new_value_value_ptr, 
				      strlen (__value) + 1);
       memset (new_value_value_ptr, '\0', strlen (__value) + 1);
       __o_value_obj -> __o_value = new_value_value_ptr;
     }
     strcpy (__o_value_obj -> __o_value, __value);
  }

  return SUCCESS;
}

/*
 *  Add an empty buffer to the object's value variable.
 */
int __ctalkSetObjectValueBuf (OBJECT *__o, char *__buf) {

  OBJECT *__o_value_obj;

  if ((__o_value_obj = __ctalkGetInstanceVariable (__o, "value", FALSE)) 
      != NULL) {
    free (__o_value_obj -> __o_value);
    __o_value_obj -> __o_value = __buf;
  } else {
    _warning 
      ("__ctalkSetObjectValueBuf: Instance variable, \"value,\" not found.\n");
    return ERROR;
  }

  return SUCCESS;
}

/*
 *  This function is only used by __ctalkDefineInstanceVariable and
 *  __ctalkDefineClassVariable, so there should be no need to print
 *  a warning, only return an ERROR.  Some classes don't have value
 *  variables, which the functions don't check for.
 */
int __ctalkSetObjectValueClass (OBJECT *__o, OBJECT *value_class_obj) {
  OBJECT *value_var;
  if ((value_var = __ctalkGetInstanceVariable (__o, "value", FALSE)) == NULL)
    return ERROR;
  strcpy (value_var -> __o_classname, value_class_obj -> __o_name);
  strcpy (value_var -> __o_superclassname, value_class_obj -> __o_superclassname);
  value_var -> __o_class = value_class_obj;
  value_var -> __o_superclass = value_class_obj -> __o_superclass;
  return SUCCESS;
}

int __ctalkSetObjectName (OBJECT *__o, char *__name) {

  if (!IS_OBJECT(__o))
    return ERROR;
  if (strlen(__name) > MAXLABEL)
    __name[MAXLABEL] = 0;
  strcpy (__o -> __o_name, __name);
  return SUCCESS;
}

/*
 *  If a constructor has made a copy of a class object, then 
 *  change to a local variable and set its reference count to
 *  0.
 */
static int __fixup_class_object_copy (OBJECT *o) {
  OBJECT *o_class;
  if (IS_CLASS_OBJECT(o)) {
    if ((o_class = __ctalkGetClass (o -> __o_name)) != NULL) {
      if (o_class != o) {
	__ctalkSetObjectScope (o, LOCAL_VAR);
	__objRefCntSet (OBJREF(o), 0);
      }
    }
  }
  return SUCCESS;
}

extern RT_INFO *__call_stack[MAXARGS+1];   /* Declared in rtinfo.c */
extern int __call_stack_ptr;

static void xlinked_object_warning (OBJECT *obj, OBJECT *obj_next) {
  int idx;
  if ((obj -> scope & METHOD_USER_OBJECT) || 
      (obj_next -> scope & METHOD_USER_OBJECT))
    return;
  fprintf (stderr, 
	   "Warning: __ctalkDeleteObjectInternal: Possibly cross linked object.\nWarning: Addr: %p, Name: \"%s,\" Class: \"%s.\"\n", 
	   obj_next, obj_next -> __o_name, 
	   obj_next -> __o_classname);
  
  for (idx = __call_stack_ptr + 1; idx <= MAXARGS; idx++) {
    if (__call_stack[idx] -> method) {
      fprintf (stderr, "\tFrom %s : %s\n", 
	       __call_stack[idx]->method->rcvr_class_obj->__o_name,
	       __call_stack[idx]->method->name);
    } else {
      fprintf (stderr, "\tFrom %s\n", "<cfunc>");
    }
  }
}

/*
 *  Delete an object after making sure that it is not in
 *  a dictionary or class library.  
 *
 *  NOTE - This is iffy, in case we want to delete a class 
 *  or a global object, in which case the application should
 *  call __ctalkDeleteObject, so it isn't checked here.
 *
 *  The function __ctalk_arg_cleanup checks that an object
 *  is not global before trying to delete it.
 *
 *  If the object is an instance variable, make sure that its
 *  reference count is the same as the dictionary object's 
 *  reference count.
 *
 *  Check for the recursion level during a cleanup, in
 *  case called randomly from within a signal handler.
 */

extern int __cleanup_deletion;
static int cleanup_lvl = 0;

static int is_class_variable (OBJECT *__o) {
  OBJECT *p;
  for (p = __o -> __o_p_obj; p && p -> __o_p_obj; p = p -> __o_p_obj)
    ;
  if (p && IS_CLASS_OBJECT(p))
    return TRUE;
  return FALSE;
}

extern VARENTRY *__ctalk_dictionary;
static int is_declared_outside_method (OBJECT *tgt) {
  VARENTRY *v, *v_t;
  int i;
  /*
   *  Start looking in the scope before the current method 
   *  == __call_stack_ptr + 2.
   */
  for (i = __call_stack_ptr + 2; i <= MAXARGS; i++) {
    if (__call_stack[i] -> method) {
      v = __call_stack[i] -> method -> local_objects[0].vars;
      for (v_t = v; v_t; v_t = v_t -> next)
	if (v_t -> var_object == tgt)
	  return TRUE;
    } else {
      v = __call_stack[i] -> _rt_fn -> local_objects.vars;
      for (v_t = v; v_t; v_t = v_t -> next)
	if (v_t -> var_object == tgt)
	  return TRUE;
    }
  }
  for (v_t = __ctalk_dictionary; v_t; v_t = v_t -> next)
    if (v_t -> var_object == tgt)
      return TRUE;
  return FALSE;
}

/*
*** WARNING: 
  Be careful here if we get a segfault -
  we may need to check the scope of the 
  reffed objects more carefully.
*/
void cleanup_reffed_object (OBJECT *__o, OBJECT *__r) {
  int has_super_scope;
  has_super_scope = is_declared_outside_method (__r);
  if (IS_OBJECT (__r) && !is_class_variable (__r) &&
      !(__r->scope & METHOD_USER_OBJECT)) {
    __ctalkSetObjectScope (__r, __r -> scope & ~VAR_REF_OBJECT);
    __fixup_class_object_copy (__r);
    if (!__cleanup_deletion) {
      if (has_super_scope) {
	if (__r -> nrefs > 1) {
	  __objRefCntDec(OBJREF(__r));
	}
      } else {
	__objRefCntDec(OBJREF(__r));
      }
      if (!parent_ref_is_circular (__o, __r)) {
	if (!has_super_scope)
	  _cleanup_temporary_objects (__r, NULL, NULL, 
				      rt_cleanup_obj_ref);
	/*
	 *  If the object still exists with a reference count of 0....
	 *  This case can occur with intermediate objects
	 *  created by Collection : atPut.
	 */
	if (IS_OBJECT(__r)) {
	  if (!IS_CLASS_OBJECT (__r)) {
	    if (__r -> nrefs == 0) {
	      if ((__r -> __o_p_obj) && 
		  (__r -> __o_p_obj -> nrefs > 0)) {
		/*  With a fixup kludge if necessary, should an instance
		 *  variable get dereffed independently of its parent 
		 *  object.
		 */
		__r -> nrefs = __r -> __o_p_obj -> nrefs;
	      } else {
		if (__o -> __o_p_obj) {
		  if (__o -> __o_p_obj -> nrefs == 0) {
		    __objRefCntZero (OBJREF(__r));
		    __ctalkDeleteObject (__r);
		  }
		} else {
		  if (__o -> nrefs == 0) {
		    __objRefCntZero (OBJREF(__r));
		    __ctalkDeleteObject (__r);
		  }
		}
	      }
	    }
	  }
	}
      } else {
	__o -> __o_value[0] = '\0';
      }
      /*
       *  NOTE: Kludge for references that have
       *  ARG_VAR and/or CREATED_PARAM, etc.
       */
    } else {
      if (!parent_ref_is_circular (__o, __r)) {
	if ((__r -> scope != GLOBAL_VAR) ||
	    ((__r -> scope == GLOBAL_VAR) &&
	     (!IS_CLASS_OBJECT(__r)))) {
	  if (__r -> scope & LOCAL_VAR) {
	    if (!__r -> __o_p_obj && !__r -> __o_varentry) {
	      __objRefCntZero (OBJREF(__r));
	      __call_delete_object_internal (__r);
	    }
	  } else {
	    __objRefCntZero (OBJREF(__r));
	    __call_delete_object_internal (__r);
	  }
	}
      }
      __o -> __o_value[0] = '\0';
    }
  }
}

void __ctalkDeleteObjectInternal (OBJECT *__o) {
  OBJECT *obj, *__r;

  if (__cleanup_deletion) {
    if (++cleanup_lvl > MAXARGS)
      exit (1);
  }
  if (!IS_OBJECT (__o)) {
#ifdef DEBUG_OBJECT_DESTRUCTION
    _warning ("__ctalkDeleteObjectInternal: Not an object.\n");
#endif
    return;
  }
  if (__o -> nrefs <= 0) {
    if (__o -> instancevars) {
      OBJECT *o_ptr, *o_ptr_prev;
      for (o_ptr = __o -> instancevars; 
	   IS_OBJECT(o_ptr) && o_ptr -> next; 
	   o_ptr = o_ptr -> next) {
	if (IS_OBJECT(o_ptr -> next) &&
	    (o_ptr -> scope != o_ptr -> next -> scope) &&
	    (o_ptr -> __o_p_obj != o_ptr -> next -> __o_p_obj)) {
	  xlinked_object_warning (o_ptr, o_ptr -> next);
	  goto xlinked_jmp;
	}
      }
    xlinked_jmp:
      if (o_ptr == __o -> instancevars) {
	if (o_ptr) __call_delete_object_internal (o_ptr);
	__o -> instancevars = NULL;
      } else {
	while (o_ptr && (o_ptr != __o -> instancevars)) {
	  o_ptr_prev = (o_ptr) ? o_ptr -> prev : NULL;
	  if (IS_OBJECT (o_ptr)) {
	      __call_delete_object_internal (o_ptr);
	  } else {
	    while (o_ptr_prev && 
		   !IS_OBJECT (o_ptr_prev) && 
		   (o_ptr_prev != __o -> instancevars)) {
	      if ((o_ptr_prev = o_ptr_prev -> prev) == NULL)
		break;
	    }
	  }
	  o_ptr = o_ptr_prev;
	  if (o_ptr) o_ptr -> next = NULL;
	}
	__call_delete_object_internal(__o -> instancevars);
	__o -> instancevars = NULL;
      }
    }
    if (!__cleanup_deletion) {
      if ((obj = 
	   __ctalk_get_object 
	   (__o -> __o_name, __o -> __o_classname)) != NULL) {
	/*
	 *  TODO - The call above depends too much on 
	 *  __ctalk_get_object's search for objects, and
	 *  the searches should be factored out.
	 */
	if (__o == obj)
	  __ctalk_remove_object (obj);
      }
    }
    if (IS_OBJECT (__o)) {
     if (__o -> __o_value)  {
       if ((__r = obj_ref_str (__o -> __o_value)) != NULL) {
	 cleanup_reffed_object (__o, __r);
       }
       if (IS_OBJECT(__o)) {
	 __xfree (__o -> __o_value);
	 __o ->__o_value = NULL;
       }
     }
     if (IS_OBJECT(__o)) {
#ifndef __sparc
       __o -> sig = 0;
#else
       __o -> sig[0] = '\0';
#endif
       __xfree (__o);
     }
     __o = NULL;
    }
  } else {
    __objRefCntDec (OBJREF (__o));
  }
}

/*
 *   Set an object's reference count to 0.  Also set the reference
 *   count of the object's instance variables to 0, unless they 
 *   are in the dictionary or the class dictionary.
 */

void __call_delete_object_internal (OBJECT *__o) {
  OBJECT *var;
  if (IS_OBJECT(__o)) {
    __objRefCntSet (OBJREF(__o), 0);
    for (var = __o -> instancevars; IS_OBJECT(var) && var -> next; var = var -> next)
      __objRefCntSet(OBJREF(var), 0);
    __ctalkDeleteObjectInternal (__o);
  }
}

/* NOTE: This function still cannot absolutely destroy an object on its own,
   without messing with the minimum reference counts for persistent 
   objects.  If you're calling this function when you know for certain 
   that you don't want to keep the object, zero the object's reference 
   count first with __objRefCntZero ().
*/
void __ctalkDeleteObject (OBJECT *__o) {
  if (__cleanup_deletion) cleanup_lvl = 0;
  __call_delete_object_internal (__o);
}

OBJECT *_dup_object (OBJECT *o) {

  static OBJECT *new_object, *var_object, *s, *t;

  new_object = __ctalkCreateObject (o -> __o_name, o -> __o_classname, 
				   o -> __o_superclassname, o -> scope);
  __ctalkSetObjectValue (new_object, o -> __o_value);

  if (o -> instancevars) {
    for (t = o -> instancevars; t; t = t -> next) {
      var_object = _dup_object (t);
      if (!new_object -> instancevars) {
	new_object -> instancevars = var_object;
      } else {
	for (s = new_object -> instancevars; s -> next; s = s -> next)
	  ;
	s -> next = var_object;
	var_object -> prev = s;
      }
    }
  }
    
  return new_object;
}

