/* $Id: wregex.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/* 
 *  Simple wildcard regex matching functions for strings.  Adapted
 *  from Kernighan & Pike, April, 1999, Dr. Dobb's Journal.  On return
 *  the offsets array in __ctalkMatchText contains the string indexes
 *  of the matches, terminated with a -1.  Returns the number of
 *  matches.
 *
 *  Remember that a '*' matches zero or more occurrences, so the 
 *  pattern, "l*o," matches both, "llo," and the second, "o," in,
 *  "Hello, world!"
 */

#include <string.h>

static int matchhere (char *, char *, int *);
static int matchstar (int, char *, char *, int *);

static int matchhere (char *pattern, char *text, int *matchlength) {
  int r;
  char *p_pat, *p_text;

  if (pattern[0] == '\0') {
    return 2;
  }

  if (pattern[0] == '$' && pattern[1] == '\0')
    return *text == '\0';

  if (pattern[1] == '*') {
    if (matchstar (pattern[0], pattern + 2, text, matchlength)) {
      p_pat = &pattern[2]; p_text = text;
      if (!*matchlength)
	*matchlength = strlen (&pattern[2]);
      while ((r = matchhere (++p_pat, ++p_text, matchlength)) == 1)
	(*matchlength)++;
      return *matchlength > 0;
    }
    return *matchlength > 0;
  }

  if (*text != '\0' && (pattern[0] == '.' || pattern[0] == *text)) {
    p_pat = pattern; p_text = text;
    do {
      (*matchlength)++;
    } while ((r = matchhere (++p_pat, ++p_text, matchlength)) == 1);
    return r > 0;
  }
  return 0;
}


static int matchstar (int c, char *pattern, char *text, int *matchlength) {
  *matchlength = 0;
  do {  /* A * matches zero or more instances. */
    if (matchhere (pattern, text, matchlength) == 1) {
      if (!(*matchlength))
	return strlen (pattern);
      else  /* We want the pattern to match as much as possible. */
	return *matchlength + strlen (pattern);
    }
    if (*text == c)
      (*matchlength)++;
  } while (*text != '\0' && (*text++ == c || c == '.'));
  if (*pattern)
    *matchlength = 0;
  return *matchlength;
}

/*
 *  Return value is the number of matches.
 */
int __ctalkMatchText (char *pattern, char *text, long long int *offsets) {
  int matchlength, s, s_offsets;

  if (pattern[0] == '^') {
    if (matchhere (pattern + 1, text, &matchlength)) {
      offsets[0] = 0;
      offsets[1] = -1;
      return 1;
    } else {
      offsets[0] = -1;
      return 0;
    }
  }
  s = s_offsets = matchlength = 0;
  do {
    if (matchhere (pattern, &text[s], &matchlength)) {
      offsets[s_offsets++] = s;
      s += matchlength + 1;
    } else {
      ++s;
    }
  } while (text[s] != '\0');
  offsets[s_offsets] = -1;
  return s_offsets;
}
