/* $Id: wiki.c,v 1.1.1.1 2014/05/01 19:46:37 rkiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005 - 2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

/*
 *  wiki.c - A (very) simple CGI Wiki written in Ctalk.
 *
 *  Installation: 
 * 
 *  Compile wiki.c with the CGIApp class and install 
 *  in the Apache server's cgi-bin directory.  The
 *  exact cgi-bin directory may be different for 
 *  other servers.
 *
 *    $ ctcc -I . wiki.c -o wiki
 *    $ su
 *    # cp wiki /usr/local/apache/cgi-bin
 *    # chmod +x /usr/local/apache/cgi-bin/wiki
 *
 *  Also make sure that the Wiki can write to the 
 *  htdocs directory. (It's a very simple Wiki, and
 *  it is easily modifiable to use another document
 *  directory.)
 *
 *  Usage: 
 *  
 *  To create a page named, "NewPage." 
 *
 *    http://<server>/cgi-bin/wiki?page=NewPage
 *
 *  To edit an existing page named, "HomePage."
 *
 *    http://<server>/cgi-bin/wiki?page=HomePage&action=edit
 *
 *  To view, "HomePage."
 *
 *    http://<server>/cgi-bin/wiki?page=HomePage&action=edit
 *
 *  If you don't specify any parameters, the Wiki uses 
 *  the default values, "HomePage," and, "view," for 
 *  the page and action CGI paramters.
 *
 *  In other words, the following two URLs are equivalent.
 *
 *    http://<server>/cgi-bin/wiki?page=HomePage&action=view
 *
 *    http://<server>/cgi-bin/wiki
 *  
 *  Things To Do:

 *  1. Handle document directories other than the main 
 *     Apache document directory. The Wiki needs 
 *     permissions to write documents in its document
 *     directory.  
 *  2. Recognize Wiki-compatible text markup.
 *  3. Add templates for page header forms, add footers, and 
 *     add time-and-date notation in the page forms.
 *  3. Implement a version control system for Wiki pages.
 *  4. Add log entries for pages and page searching.
 *  5. Add an on-line help system.
 */

CGIApp new wikiApp;

String new wikiPath;
String new documentPath;
String new serverName;
String new requestMethod;
Integer new error;

Exception instanceMethod errorToBrowser (void) {
  String new msg;
  wikiApp httpHeader;
  wikiApp htmlPageHeader "Ctalk Wiki | Error";
  wikiApp outputString "<h2>Ctalk Wiki Error:</h2>";
  msg = __ctalkPeekExceptionTrace ();
  wikiApp outputString "<pre>" + msg + "</pre><br>";
  msg = __ctalkPeekRunTimeException ();
  wikiApp outputString "<pre>" + msg + "</pre>";
  wikiApp outputString "Please contact your system adminstrator or " + 
    "the Ctalk authors, at rk3314042@gmail.com.";
  error = TRUE;
  return NULL;
}

CGIApp instanceMethod wikiPaths (void) {
  Exception new __pathError;
  __pathError installHandler "errorToBrowser";
  wikiPath = wikiApp serverEnvironment at "SCRIPT_NAME";
  if (wikiPath isNull)
    __pathError raiseException UNDEFINED_LABEL_X, "SCRIPT_NAME";
  documentPath = wikiApp serverEnvironment at "DOCUMENT_ROOT";
  if (documentPath isNull)
    __pathError raiseException UNDEFINED_LABEL_X, "DOCUMENT_ROOT";
  serverName = wikiApp serverEnvironment at "SERVER_NAME";
  if (serverName isNull)
    __pathError raiseException UNDEFINED_LABEL_X, "SERVER_NAME";
  requestMethod = wikiApp serverEnvironment at "REQUEST_METHOD";
  if (requestMethod isNull)
    __pathError raiseException UNDEFINED_LABEL_X, "REQUEST_METHOD";
  if (__pathError pending)
    __pathError handle;
  return NULL;
}

CGIApp instanceMethod getWikiPage (String pageName) {
  String new path;
  ReadFileStream new pageFile;
  String new text;
  Exception new wikiPageException;

  returnObjectClass String;

  wikiPageException installHandler "errorToBrowser";
  path = documentPath + ('/' asString) + pageName;

  if (pageFile exists path) {
    pageFile openOn path;
    if (wikiPageException pending) {
      wikiPageException handle;
      methodReturnString("");
    } else {
      text = pageFile readAll;
      pageFile closeStream;
      return text;
    }
  } else {
    methodReturnString ("");
  }
}

CGIApp instanceMethod constructURL (String __cgiParams) {
  String new url;
  returnObjectClass String;
  url = "http://" + serverName + wikiPath +  __cgiParams;
  return url;
}

CGIApp instanceMethod pageHeader (String pageName) {
  String new homePageString;
  String new homePageURL;
  String new editPageString;
  String new editPageURL;

  self outputString "<h2>" + pageName + "</h2>";
  homePageString = self constructURL "?page=HomePage&action=view";
  homePageURL = "<a href=\"" + homePageString + "\">HomePage</a>"; 
  self outputString homePageURL;
  self outputString "&nbsp;";
  editPageString = self constructURL "?page=" + pageName + "&action=edit";
  editPageURL = "<a href=\"" + editPageString + "\">Edit</a>"; 
  self outputString editPageURL;
  self outputString "<hr>";
  return NULL;
}

CGIApp instanceMethod editPage (String pageText, String pageName) {
  self outputString "<form action=\"?page=" + pageName + "\" method=\"post\">";
  self outputString "<table>";
  self outputString "<tr><td rowspan=\"2\">";
  self outputString "<textarea rows=\"25\" cols=\"80\" name=\"text\">";
  self outputString pageText;
  self outputString "</textarea>";
  self outputString "</td>";
  self outputString "<td>";
  self outputString "Edit the page text and click on, \"Save,\" below.";
  self outputString "</td></tr>";
  self outputString "<tr><td>";
  self outputString "<input type=\"submit\" value=\"Save\">";
  self outputString "</td></tr></table>";
  self outputString "</form>";
  return NULL;
}

CGIApp instanceMethod decodePostData (String postData) {
  String new decodedData;
  Integer new dataLength;
  Integer new i;

  returnObjectClass String;

  decodedData = postData;
  dataLength = decodedData length;
  for (i = 0; i < dataLength; i = i + 1) {
    if (decodedData at i == '+')
      decodedData atPut i, ' ';
  }
  decodedData = wikiApp httpUnescapeString decodedData;
  return decodedData;
}

CGIApp instanceMethod handlePostData (String pageName) {
  String new postInput;
  Array new postData;
  Integer new contentLength;
  String new contentLengthString;
  String new decodedPostInput;
  WriteFileStream new outputFile;
  String new path;
  Exception new postDataException;

  /*
   *  Initialize stdin stream.
   */
  ReadFileStream classInit;
  postDataException installHandler "errorToBrowser";

  contentLengthString = wikiApp serverEnvironment at "CONTENT_LENGTH";
  contentLength = contentLengthString asInteger;

  postInput = stdinStream readRec contentLength;
  postInput split '=', postData;
  decodedPostInput = wikiApp decodePostData (postData at 1);

  path = documentPath + ('/' asString) + pageName;
  outputFile openOn path;
  if (postDataException pending) {
    postDataException handle;
    exit (1);
  }
  outputFile writeStream decodedPostInput;
  if (postDataException pending) {
    postDataException handle;
    exit (1);
  }
  outputFile closeStream;
  return decodedPostInput;
}

int main (int argc, char **argv, char **envp) {
  String new page;
  String new action;
  String new pageText;

  wikiApp enableExceptionTrace;
  error = FALSE;

  wikiApp getServerEnv envp;
  wikiApp wikiPaths;
  if (error)
    exit (1);

  page = "HomePage";
  action = "view";
  
  wikiApp parseQueryString;

  if (wikiApp queryValues keyExists "page")
    page = wikiApp queryValues at "page";

  if (wikiApp queryValues keyExists "action")
    action = wikiApp queryValues at "action";

  if (requestMethod == "POST") {
    wikiApp handlePostData page;
    action = "view";
  }

  wikiApp httpHeader;
  wikiApp htmlPageHeader "Ctalk Wiki | " + page;

  wikiApp pageHeader page;

  pageText = wikiApp getWikiPage page;

  if (action == "view") {
    if (pageText length == 0) {
      wikiApp editPage ("Enter the text of " + page + " here."), page;
    } else {
      printf ("%s<br>", pageText);
    }
  }
  if (action == "edit")
    wikiApp editPage pageText, page;
  return 0;
}

