/* $Id: premethod.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright © 2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *  Read and write cached methods.
 *
 *  The fixups here work in synchronization, like the class library lookups,
 *  so there *could* yet be  _errors () if an individual cached method gets 
 *  deleted, if it was declared before the method being un-cached to the
 *  output.
 *
 *  If one cached method gets deleted, and it causes an error or exception,
 *  then the entire cache needs be cleared and rewritten.  Use the 
 *  --clearpreload command line option.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"


extern char output_file[];

HASHTAB selectors_output;
HASHTAB class_and_instance_vars_output;

extern NEWMETHOD *new_methods[MAXARGS+1];  /* Declared in lib/rtnwmthd.c. */
extern int new_method_ptr;

void init_pre_methods (void) {
  _new_hash (&selectors_output);
  _new_hash (&class_and_instance_vars_output);
}

int pre_method_is_output (char *selector) {
  if (_hash_get (selectors_output, selector))
    return TRUE;
  else
    return FALSE;
}

/*
 *  Save the information that generate_*_method_definition_call ()
 *  and generate_*_method_param_definition_call () need.
 *  We can queue this with the other global inits when
 *  reading the cached method.
 */
void save_pre_init_info (char *cache_fn, METHOD *n_method, 
			 OBJECT *class_object) {
  FILE *fp;
  int length, bytes_written, r, i;
  char path[FILENAME_MAX];
  char buf[MAXLABEL];
  OBJECT *var;
  NEWMETHOD *n;

  sprintf (path, "%s.%s", cache_fn, "init");

  if ((fp = fopen (path, FILE_WRITE_MODE)) == NULL) 
    _error ("save_pre_init_info (fopen): %s: %s.\n",
	    path, strerror (errno));

  n = new_methods[new_method_ptr+1];

  for (i = MAXARGS; i > n -> argblk_ptr; i--) {
    sprintf (buf, "a:%s:%s:%s:%s:%d:%d\n",
	     class_object -> __o_name,
	     n -> argblks[i] -> returnclass,
	     n -> argblks[i] -> selector,
	     n -> argblks[i] -> selector,
	     0, 0);

    length = strlen (buf);

    if ((bytes_written = fwrite ((void *)buf, sizeof (char), length, fp))
	!= length) 
      _error ("save_pre_init_info (fwrite (a:)): %s: %s.\n",
	      path, strerror (errno));

  }

  sprintf (buf, "m:%s:%s:%s:%s:%d:%d\n",
	   class_object -> __o_name,
	   n_method -> returnclass,
	   n_method -> name, 
	   n_method -> selector,
	   n_method -> n_params,
	   method_definition_attributes (n_method));
  length = strlen (buf);

  if ((bytes_written = fwrite ((void *)buf, sizeof (char), length, fp))
      != length) 
    _error ("save_pre_init_info (fwrite (m:)): %s: %s.\n",
	    path, strerror (errno));

  for (i = 0; i < n_method -> n_params; i++) {

    if (n_method -> params[i] -> attrs & PARAM_C_PARAM)
      continue;

    sprintf (buf, "p:%s:%s:%s:%s:%d:%d\n",
	     class_object -> __o_name,
	     n_method -> name, 
	     n_method -> params[i] -> name,
	     n_method -> params[i] -> class,
	     n_method -> params[i] -> is_ptr,
	     n_method -> params[i] -> is_ptrptr);

    length = strlen (buf);

    if ((bytes_written = fwrite ((void *)buf, sizeof (char), length, fp))
	!= length) 
      _error ("save_pre_init_info (fwrite (p:)): %s: %s.\n",
	      path, strerror (errno));

  }

  if (class_object -> instancevars) {
    for (var = class_object -> instancevars; var; var = var -> next) {

      /* value gets created automatically at run time. */
      if (!strcmp (var -> __o_name, "value"))
	continue;

      sprintf (buf, "i:%s:%s:%s:%s\n",
	       class_object -> __o_name,
	       var -> __o_name,
	       ((var -> instancevars) ? 
		var -> instancevars -> __o_classname : 
		var -> __o_classname),
	       var -> __o_value);

    length = strlen (buf);

    if ((bytes_written = fwrite ((void *)buf, sizeof (char), length, fp))
	!= length) 
      _error ("save_pre_init_info (fwrite (i:)): %s: %s.\n",
	      path, strerror (errno));

    }

  }

  if (class_object -> classvars) {
    for (var = class_object -> classvars; var; var = var -> next) {

      /* again, value gets created automatically at run time. */
      if (!strcmp (var -> __o_name, "value"))
	continue;

      sprintf (buf, "c:%s:%s:%s:%s\n",
	       class_object -> __o_name,
	       var -> __o_name,
	       ((var -> instancevars) ? 
		var -> instancevars -> __o_classname : 
		var -> __o_classname),
	       var -> __o_value);

    length = strlen (buf);

    if ((bytes_written = fwrite ((void *)buf, sizeof (char), length, fp))
	!= length) 
      _error ("save_pre_init_info (fwrite (c:)): %s: %s.\n",
	      path, strerror (errno));

    }

  }

  if ((r = fclose (fp)) != 0)
    _error ("save_pre_init_info (fclose): %s: %s.\n",
	    path, strerror (errno));

}

static struct _var_init_info {
  char classname[MAXLABEL],
    varname[MAXLABEL],
    varclass[MAXLABEL],
    init[MAXMSG];
} var_info;

struct _param_init_info {
  char classname[MAXLABEL],
    methodname[MAXLABEL],
    paramname[MAXLABEL],
    paramclass[MAXLABEL];
  int is_ptr,
    is_ptrptr;
};

struct _argblk_init_info {
  char classname[MAXLABEL],
    returnclassname[MAXLABEL],
    methodname[MAXLABEL],
    selectorname[MAXLABEL];
};

static struct _init_info {
  char path[FILENAME_MAX],
    classname[MAXLABEL],
    returnclassname[MAXLABEL],
    methodname[MAXLABEL],
    selectorname[MAXLABEL];
  int required_args,
    method_attrs;
  struct _param_init_info params[MAXARGS];
  struct _argblk_init_info argblks[MAXARGS];
} init_info;

static int init_param_ptr = 0;
static int init_argblk_ptr = 0;

#define I_OR_C_CLASS "class"
#define I_OR_C_INSTANCE "instance"

void set_class_and_instance_var_init_output (char *i_or_c) {
  char buf[MAXMSG];
  sprintf (buf, "%s_%s_%s", 
	   var_info.classname,
	   i_or_c,
	   var_info.varname);
  _hash_put (class_and_instance_vars_output, buf, buf);
}

void set_class_and_instance_var_init_output_2 (char *classname, char *i_or_c,
					       char *varname) {
  char buf[MAXMSG];
  sprintf (buf, "%s_%s_%s", 
	   classname,
	   i_or_c,
	   varname);
  _hash_put (class_and_instance_vars_output, buf, buf);
}

int class_or_instance_var_init_is_output (char *i_or_c) {
  char buf[MAXMSG];
  sprintf (buf, "%s_%s_%s", 
	   var_info.classname,
	   i_or_c,
	   var_info.varname);
  if (_hash_get (class_and_instance_vars_output, buf))
    return TRUE;
  else
    return FALSE;
}

int class_or_instance_var_init_is_output_2 (char *classname,
					    char *i_or_c, 
					    char *varname) {
  char buf[MAXMSG];
  sprintf (buf, "%s_%s_%s", 
	   classname,
	   i_or_c,
	   varname);
  if (_hash_get (class_and_instance_vars_output, buf))
    return TRUE;
  else
    return FALSE;
}

static void fill_in_init_info (char *selector, int gen_var_init) {
  char path[FILENAME_MAX], *buf;
  char ibuf[0xff];
  int fsize;
  char *lptr, *rptr, *rptr2, *eptr;

  sprintf (path, "%s/%s.init", home_libcache_path (), selector);

  fsize = file_size (path);

  if ((buf = (char *)malloc (fsize + 1)) == NULL)
    _error ("fill_in_init_info (malloc): %s.\n", strerror (errno));

  memset (buf, 0, fsize + 1);

  (void)read_file (buf, path);

  memset ((void *)&init_info, 0, sizeof (struct _init_info));
  
  init_param_ptr = 0;
  init_argblk_ptr = 0;

  lptr = buf;
  eptr = strchr (buf, '\0');
  rptr2 = NULL;

  do {

    switch (*lptr)
      {
      case 'm':
	if ((rptr = strchr (lptr, ':')) != NULL) {
	  ++rptr;
	  if ((rptr2 = strchr (rptr, ':')) != NULL) {
	    strncpy (init_info.classname, rptr, rptr2 - rptr);
	    rptr = rptr2 + 1;
	  }
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.returnclassname, rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.methodname, rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.selectorname, rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  memset ((void *)ibuf, 0, 0xff);
	  strncpy (ibuf, rptr, rptr2 - rptr);
	  init_info.required_args = atoi (ibuf);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, '\n')) != NULL) {
	  memset ((void *)ibuf, 0, 0xff);
	  strncpy (ibuf, rptr, rptr2 - rptr);
	  init_info.method_attrs = atoi (ibuf);
	  rptr = rptr2 + 1;
	}

	lptr = rptr2 + 1;

	break;

      case 'a':
	if ((rptr = strchr (lptr, ':')) != NULL) {
	  ++rptr;
	  if ((rptr2 = strchr (rptr, ':')) != NULL) {
	    strncpy (init_info.argblks[init_argblk_ptr].classname, 
		     rptr, rptr2 - rptr);
	    rptr = rptr2 + 1;
	  }
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.argblks[init_argblk_ptr].returnclassname, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.argblks[init_argblk_ptr].methodname, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.argblks[init_argblk_ptr].selectorname, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	/* Skip past the required args for arg blocks - they're always 0. */
	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  /* memset ((void *)ibuf, 0, 0xff);c */
	  /* strncpy (ibuf, rptr, rptr2 - rptr); */
	  /* init_info.required_args = atoi (ibuf); */
	  rptr = rptr2 + 1;
	}

	/* The same with the attributes - they're always 0 for arg blocks. */
	if ((rptr2 = strchr (rptr, '\n')) != NULL) {
	  /* memset ((void *)ibuf, 0, 0xff); */
	  /* strncpy (ibuf, rptr, rptr2 - rptr); */
	  /* init_info.method_attrs = atoi (ibuf); */
	  rptr = rptr2 + 1;
	}

	lptr = rptr2 + 1;

	++init_argblk_ptr;

	break;

      case 'p':
	if ((rptr = strchr (lptr, ':')) != NULL) {
	  ++rptr;
	  if ((rptr2 = strchr (rptr, ':')) != NULL) {
	    strncpy (init_info.params[init_param_ptr].classname, 
		     rptr, rptr2 - rptr);
	    rptr = rptr2 + 1;
	  }
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.params[init_param_ptr].methodname, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}


	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.params[init_param_ptr].paramname, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (init_info.params[init_param_ptr].paramclass, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  memset ((void *)ibuf, 0, 0xff);
	  strncpy (ibuf, rptr, rptr2 - rptr);
	  init_info.params[init_param_ptr].is_ptr = atoi (ibuf);
	  rptr = rptr2 + 1;
	}


	if ((rptr2 = strchr (rptr, '\n')) != NULL) {
	  memset ((void *)ibuf, 0, 0xff);
	  strncpy (ibuf, rptr, rptr2 - rptr);
	  init_info.params[init_param_ptr].is_ptrptr = atoi (ibuf);
	  rptr = rptr2 + 1;
	}

	lptr = rptr2 + 1;

	++init_param_ptr;

	break;
      case 'i':

	memset ((void *)&var_info, 0, sizeof (var_info));

	if ((rptr = strchr (lptr, ':')) != NULL) {
	  ++rptr;
	  if ((rptr2 = strchr (rptr, ':')) != NULL) {
	    strncpy (var_info.classname, 
		     rptr, rptr2 - rptr);
	    rptr = rptr2 + 1;
	  }
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (var_info.varname, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	
	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (var_info.varclass, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, '\n')) != NULL) {
	  strncpy (var_info.init, 
		     rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	lptr = rptr2 + 1;

	if (gen_var_init)
	  generate_define_instance_variable_call 
	    (var_info.classname, var_info.varname, 
	     var_info.varclass, var_info.init);

	break;

      case 'c':

	memset ((void *)&var_info, 0, sizeof (var_info));

	if ((rptr = strchr (lptr, ':')) != NULL) {
	  ++rptr;
	  if ((rptr2 = strchr (rptr, ':')) != NULL) {
	    strncpy (var_info.classname, 
		     rptr, rptr2 - rptr);
	    rptr = rptr2 + 1;
	  }
	}

	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (var_info.varname, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	
	if ((rptr2 = strchr (rptr, ':')) != NULL) {
	  strncpy (var_info.varclass, 
		   rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	if ((rptr2 = strchr (rptr, '\n')) != NULL) {
	  strncpy (var_info.init, 
		     rptr, rptr2 - rptr);
	  rptr = rptr2 + 1;
	}

	lptr = rptr2 + 1;

	if (gen_var_init)
	  generate_define_class_variable_call 
	    (var_info.classname, var_info.varname, 
	     var_info.varclass, var_info.init);

	break;
      }

    if (!rptr2)
      break;
    if (*lptr == 0)
      break;

  }  while (lptr < eptr);

  free (buf);

}

char *get_init_return_class (char *selector) {

  static char *returnclass;

  fill_in_init_info (selector, FALSE);

  returnclass = init_info.returnclassname;
  return returnclass;

}

void register_init (char *selector) {
  int param_idx, argblk_idx;

  fill_in_init_info (selector, TRUE);

  for (argblk_idx = 0; argblk_idx < init_argblk_ptr; argblk_idx++) {
    generate_instance_method_definition_call
      (init_info.argblks[argblk_idx].classname,
       init_info.argblks[argblk_idx].selectorname,
       init_info.argblks[argblk_idx].selectorname,
       0, 0);
  }

  if (strstr (init_info.selectorname, "_instance_")) {
    generate_instance_method_definition_call
      (init_info.classname,
       init_info.methodname,
       init_info.selectorname,
       init_info.required_args,
       init_info.method_attrs);
    generate_instance_method_return_class_call
      (init_info.classname,
       init_info.methodname,
       init_info.returnclassname);
  } else if (strstr (init_info.selectorname, "_class_")) {
    generate_class_method_definition_call
      (init_info.classname,
       init_info.methodname,
       init_info.selectorname,
       init_info.required_args);
    generate_class_method_return_class_call
      (init_info.classname,
       init_info.methodname,
       init_info.returnclassname);
  }

  for (param_idx = 0; param_idx < init_param_ptr; param_idx++) {

    if (strstr (init_info.selectorname, "instance")) {
      generate_instance_method_param_definition_call
	(init_info.params[param_idx].classname,
	 init_info.params[param_idx].methodname,
	 /***/
	 init_info.selectorname,
	 init_info.params[param_idx].paramname,
	 init_info.params[param_idx].paramclass,
	 init_info.params[param_idx].is_ptr,
	 init_info.params[param_idx].is_ptrptr);
    } else if (strstr (init_info.selectorname, "class")) {
      _warning ("Parameter definitions for class methods aren't "
		"implemented yet!");
    }

  }

}

static int cache_method_src_to_output (char *selector) {

  char cache_fn[FILENAME_MAX], *srcbuf;
  int src_size, chars_read;

  sprintf (cache_fn, "%s/%s", home_libcache_path (), selector);

  /* 
     TO DO! If a cached method for a fixup doesn't exist, 
     generate an exception and reparse the entire class, including 
     that method. This could happen if an individual cached method 
     gets deleted.

     Also, if a method's dependencies get updated, the entire
     class should need to be re-parsed.  The functions in 
     libdeps.c and used in primitives.c don't handle methods
     included in fixups, so this should also generate an 
     exception and cause a re-parse.

  */

  src_size = file_size (cache_fn);

  if ((srcbuf = (char *) malloc (src_size + 1 * sizeof (char))) == NULL)
    _error ("output_pre_method (malloc): %s.\n",
	    strerror (errno));

  memset (srcbuf, 0, src_size);

  if ((chars_read = read_file (srcbuf, cache_fn)) != src_size) {
    _error ("output_pre_method (read_file): %s.",
	    strerror (errno));
    return -1;
  }

  output_pre_method (selector, srcbuf);

  free (srcbuf);
  
  return 0;
}

static inline void elide_fixup_marker (char *marker_start) {
  char *s;

  s = marker_start;
  do {
    *s = ' ';
    ++s;
  } while (*s != '\n');

}

void output_pre_method (char *selector, char *src) {

  char *o, *p, *q, *r, buf[MAXLABEL + 1];

  if (_hash_get (selectors_output, selector))
    return;

  if (!strstr (src, "__ctalk_method")) {

    _hash_put (selectors_output, selector, selector);

    __fileout_cache (src);
    __fileout_cache ("\n\n");
    register_init (selector);

  } else {

    o = src;

    /* Fixup the classes first. */
    while ((p = strstr (o, CONSTRUCTOR_ARG_FN)) != NULL) {

      q = strchr (p, '"'); ++q;
      p = strchr (q, '"');

      memset (buf, 0, MAXLABEL);
      strncpy (buf, q, p - q);

      if (is_pending_class (buf) ||
	  is_cached_class (buf)) {
	o = q;
	continue;
      }

      load_cached_class (buf);
      o = q;

    }

    o = src;

    /* Then any method calls. */
    while ((p = strstr (o, "__ctalk_method")) != NULL) {

      q = strchr (p, ',');

      ++q;
      while (isspace ((int)*q))
	++q;

      p = strchr (q, ')');

      memset (buf, 0, MAXLABEL);

      strncpy (buf, q, p - q);

      if (_hash_get (selectors_output, buf)) {
	o = q;
	continue;
      }

      cache_method_src_to_output (buf);

      o = q;

    }

    o = src;

    /* 
       Also take care of fixup markers. They're only written
       in eval_arg () right now for method messages that occur in 
       argument expressions.
    */
    while ((p = strstr (o, "/@ctalkfixup/")) != NULL) {

      r = p;     /* Find start of fixup marker. */
      while (*(--r) != '#')
	;

      ++p; q = strchr (p, '/'); /* Find end of selector. */
      ++q; p = q;
      q = strchr (p, '"');

      memset ((void *)buf, 0, MAXLABEL+1);
      strncpy (buf, p, q - p);

      if (_hash_get (selectors_output, buf)) {
	elide_fixup_marker (r);
	o = q;
	continue;
      }

      cache_method_src_to_output (buf);

      elide_fixup_marker (r);

      o = q;

    }

    if (!_hash_get (selectors_output, selector)) {

      _hash_put (selectors_output, selector, selector);
      __fileout_cache (src);
      __fileout_cache ("\n\n");
      register_init (selector);

    }

  }
}
