/* $Id: resolve.c,v 1.6 2014/05/08 15:04:08 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "lex.h"
#include "classlib.h"
#include "symbol.h"

/* 
 *   The function resolve() looks up object and method references.
 */

extern I_PASS interpreter_pass;         /* Declared in lib/rtinfo.c.  */
extern int method_from_proto;           /* Declared in mthdref.c.     */
extern NEWMETHOD *new_methods[MAXARGS+1];  /* Declared in lib/rtnwmthd.c.*/
extern int new_method_ptr;
extern int verbose_opt;                 /* Declared in main.c.        */
extern CTRLBLK *ctrlblks[MAXARGS + 1];  /* Declared in control.c.     */
extern int ctrlblk_ptr;
extern int ctrlblk_pred, ctrlblk_blk;
extern int for_init, for_term, for_inc;
extern int argblk;                      /* Declared in argblk.c.      */
extern int have_complex_arg_block;     /* Set and reset in complexmethod.c, */
				       /* also reset in method_call ().    */
extern int complex_arg_block_start;
extern int complex_arg_block_rcvr_ptr; /* Set by method_call ()            */
extern OBJECT *rcvr_class_obj;         /* declared in lib/rtnwmthd.c */

/*
 *  Mark a prefix method syntactically _unless_
 *  it is a ! operator and it is the first token
 *  of a contrl block predicate, in which case we 
 *  let the compiler handle the negation in context....
 */
#define PREFIX_PREV_OP_TOKEN_NOEVAL(tok) (IS_C_OP_TOKEN_NOEVAL(tok) || \
			       (tok==OPENPAREN) || \
			     (tok==CLOSEPAREN))
static void prefix_method_attr (MESSAGE_STACK messages, int message_ptr) {
  int prev_tok_idx;
  MESSAGE *m_prev_tok;
  MESSAGE *m = messages[message_ptr];
  prev_tok_idx = prevlangmsg (message_stack (), message_ptr);
  m_prev_tok = messages[prev_tok_idx];
  if (M_TOK(m) == EXCLAM) {
    if (ctrlblk_pred) {
      /*
       *  If a contrl block predicate begins with a sequence 
       *  like these, then use the C operators.  We can send
       *  them to the output verbatim.
       *
       *     <negation_prefix_tok> : !
       *     <negation_prefix_tok_sequence> :  
       *       <negation_prefix_tok> |
       *       ++ | -- | ~ | & | * <negation_prefix_tok> |
       *       <negation_prefix_tok_sequence> <negation_prefix_tok> |
       *
       *    <if|while> (<prefix_tok_sequence>!<expr>) { ...
       *  
       *  When compiling, though, only sequences of ! and ~ do not generate
       *  errors or warnings.
       *
       *  Sequences of multiple ++|-- operators are not legal C, so we
       *  don't support them in C expressions both here and elsewhere.
       */
      if (!is_leading_ctrlblk_prefix_op_tok
	  (message_stack (), 
	   message_ptr,
	   stack_start (message_stack ()))) {
	if (IS_C_UNARY_MATH_OP(M_TOK(m))) {
	  if (PREFIX_PREV_OP_TOKEN_NOEVAL(M_TOK(m_prev_tok))) {
	    m -> attrs |= TOK_IS_PREFIX_OPERATOR;
	  }
	}
      }
    } else { /* if (ctrlblk_pred) */
      if (IS_C_UNARY_MATH_OP(M_TOK(m))) {
	if (PREFIX_PREV_OP_TOKEN_NOEVAL(M_TOK(m_prev_tok))) {
	  m -> attrs |= TOK_IS_PREFIX_OPERATOR;
	}
      }
    }
  } else { /* if (M_TOK(m) == EXCLAM) { */
    if (IS_C_UNARY_MATH_OP(M_TOK(m))) {
      if (PREFIX_PREV_OP_TOKEN_NOEVAL(M_TOK(m_prev_tok))) {
	m -> attrs |= TOK_IS_PREFIX_OPERATOR;
      }
    }
  }
}

/* 
 *  Remove parens in cases like 
 * 
 *    (struct -> mbr) <method>
 *  or
 *    ((struct -> mbr)) <method>
 *
 *   when "struct -> mbr" is registered as the CVAR
 *   receiver.
 *
 *  Note that Ctalk does not yet see an expression like
 *
 *   (struct) -> mbr
 *  
 *  as a struct expression, so an expression like this
 *
 *   ((struct) -> mbr) <method>
 *
 *  doesn't get parsed here.
 */

static void remove_outer_receiver_parens_a (MESSAGE_STACK messages, 
					    int method_message_idx) {
  int stack_begin;
  int close_paren_ptr, open_paren_ptr, last_tok_idx;

  stack_begin = stack_start (messages);

  last_tok_idx = method_message_idx;
  while (1) {

    if ((close_paren_ptr = prevlangmsg (messages, last_tok_idx)) != ERROR) {

      if (M_TOK(messages[close_paren_ptr]) == CLOSEPAREN) {

	if ((open_paren_ptr = match_paren_rev (messages, 
					       close_paren_ptr,
					       stack_begin)) != ERROR) {

	  ++messages[open_paren_ptr] -> evaled;
	  ++messages[open_paren_ptr] -> output;
	  ++messages[close_paren_ptr] -> evaled;
	  ++messages[close_paren_ptr] -> output;

	}
      } else {
	
	return;

      }

      last_tok_idx = close_paren_ptr;
    }

  }

}

/*
 *  Translates a CVAR of the previous token into an object of a basic class
 *  if possible.  Also fills in <rcvr_msg>->obj, <method_msg>->receiver_msg, 
 *  and <method_msg>->receiver_obj.
 */
static int cvar_is_method_receiver (MESSAGE_STACK messages, int method_idx) {
  int prev_tok_idx;
  int prev_tok_idx_2;
  int arrayopen_idx, j, stack_start_idx, n_blocks;
  int struct_start_tok_idx;
  int have_parens;
  MESSAGE *m_prev_tok;
  MESSAGE *m_prev_tok_2;
  CVAR *c, *c_mbr, *c_defn;
  char classname[MAXLABEL];
  OBJECT *class_object;
  METHOD *m;

  stack_start_idx = stack_start (messages);

  have_parens = FALSE;

  if ((prev_tok_idx = prevlangmsg (messages, method_idx)) != ERROR) {
    m_prev_tok = messages[prev_tok_idx];

    /*
     *  If we see a series of <label> <label> <label>, 
     *  check for a C data type.  This is instead of going
     *  through is_c_var_declaration_msg () again.
     *  
     */
    if (m_prev_tok -> attrs & TOK_IS_DECLARED_C_VAR) {
      if ((prev_tok_idx_2 = prevlangmsg (messages, prev_tok_idx)) != ERROR) {

	m_prev_tok_2 = messages[prev_tok_idx_2];

	if (m_prev_tok_2 -> attrs & TOK_IS_DECLARED_C_VAR) {

	  if (is_c_data_type (M_NAME(m_prev_tok_2)))
	    return FALSE;

	}
      }
    }

    /*
     *  Handle a cases where a CVAR is enclosed in parentheses.
     */
    if (M_TOK(m_prev_tok) == CLOSEPAREN)
      have_parens = TRUE;

    while (M_TOK(m_prev_tok) == CLOSEPAREN) {
      if ((prev_tok_idx = prevlangmsg (messages, prev_tok_idx)) != ERROR) {
	m_prev_tok = messages[prev_tok_idx];
      }
    }

    if (M_TOK(m_prev_tok) == ARRAYCLOSE) {
      for (n_blocks = 0, j = prev_tok_idx, arrayopen_idx = -1; 
	   (j <= stack_start_idx) && (arrayopen_idx == -1); j++) {
	if (M_TOK(messages[j]) == ARRAYCLOSE)
	  --n_blocks;
	if (M_TOK(messages[j]) == ARRAYOPEN)
	  ++n_blocks;
	if (!n_blocks) {
	  arrayopen_idx = j;
	}
      }
      prev_tok_idx = prevlangmsg (messages, arrayopen_idx);
      m_prev_tok = messages[prev_tok_idx];
      if (((c = get_local_var (M_NAME(m_prev_tok))) != NULL) ||
	  (( c = get_global_var (M_NAME(m_prev_tok))) != NULL)) {
	strcpy (classname, 
		basic_class_from_cvar (messages[method_idx], c, 1));
	if ((class_object = get_class_object (classname)) != NULL) {
	  if (((m = get_instance_method 
		(m_prev_tok, class_object, M_NAME(messages[method_idx]), 
		 ERROR, FALSE)) 
	       != NULL) ||
	      ((m = get_class_method
		(m_prev_tok, class_object, M_NAME(messages[method_idx]), 
		 ERROR, FALSE)) 
	       != NULL)) {
	    m_prev_tok -> obj = create_object_init 
	      (class_object -> __o_name,
	       class_object -> __o_superclassname,
	       c -> name,
	       NULLSTR);
	    messages[method_idx] -> tokentype = METHODMSGLABEL;
	    messages[method_idx] -> receiver_msg = m_prev_tok;
	    messages[method_idx] -> receiver_obj = m_prev_tok -> obj;
	    messages[method_idx] -> attrs |= RCVR_OBJ_IS_C_ARRAY_ELEMENT;
	    return TRUE;
	  }
	}
      }
    } else { /* if (M_TOK(m_prev_tok) == ARRAYCLOSE) */
      if (m_prev_tok -> attrs & TOK_IS_POSTFIX_OPERATOR) {
	/*
	 *  receiver_msg is set by set_cvar_rcvr_postfix_attrs_a ().
	 */
	if (((c = get_local_var (M_NAME(m_prev_tok->receiver_msg)))!=NULL)||
	    (( c = get_global_var (M_NAME(m_prev_tok->receiver_msg)))!=NULL)){
	  if (c -> attrs & CVAR_ATTR_STRUCT_DECL) {
	    CVAR *__c_mbr, *__c_defn;
	    char *__s;
	    int __s_term, __s_start;
	    MESSAGE *m_start;
	    stack_start_idx = stack_start (messages);
	    for (__s_start = prev_tok_idx; __s_start <= stack_start_idx; 
		 __s_start++) {
	      if (messages[__s_start] == m_prev_tok -> receiver_msg)
		break;
	    }
	    m_start = messages[__s_start];
	    if ((__s_term = is_struct_expr_b (messages, 
					      __s_start,
					      get_stack_top (messages)))
		> 0) {
	      if (((__c_defn = get_local_struct_defn (M_NAME(m_start)))
		   !=NULL) ||
		  ((__c_defn = get_global_var (M_NAME(m_start)))!=NULL)) {
		__s = collect_tokens (messages, __s_start, __s_term);
		if ((__c_mbr=struct_member_from_expr(__s, __c_defn))!=NULL) {
		  free (__s);
		  strcpy (classname, basic_class_from_cvar 
			  (messages[method_idx], __c_mbr, 0));
		  if ((class_object = get_class_object (classname)) != NULL) {
		    if (((m = get_instance_method 
			  (m_prev_tok, class_object, 
			   M_NAME(messages[method_idx]), 
			   ERROR, FALSE)) 
			 != NULL) ||
			((m = get_class_method
			  (m_prev_tok, class_object, 
			   M_NAME(messages[method_idx]), 
			   ERROR, FALSE)) 
			 != NULL)) {
		      m_prev_tok -> obj = create_object_init 
			(class_object -> __o_name,
			 class_object -> __o_superclassname,
			 c -> name,
			 NULLSTR);
		      messages[method_idx] -> tokentype = METHODMSGLABEL;
		      messages[method_idx] -> attrs |= 
			RCVR_OBJ_IS_C_STRUCT_ELEMENT;
		      messages[method_idx] -> receiver_msg = 
			m_prev_tok -> receiver_msg;
		      messages[method_idx] -> receiver_obj = m_prev_tok -> obj;
		      return TRUE;
		    } else { /* if (((m = get_instance_method  ...  */
		      return FALSE;
		    } /* if (((m = get_instance_method ...  */
		  } else { /* if ((class_object = get_class_object... */
		    return FALSE;
		  } /* if ((class_object = get_class_object (classname))...*/
		}
	      }
	    }
	  } else {/* if (c -> attrs & CVAR_ATTR_STRUCT_DECL) */
	    char *__s;
	    int __struct_expr_end;
	    /* 
	     *  Case where the tag is a not the first tag:
	     *  struct _s {
	     *   ...
	     *  } <tag1>, <tag2>;
	     *
	     */
	    if ((c -> type_attrs & CVAR_ATTR_STRUCT_TAG) ||
		(c -> type_attrs & CVAR_ATTR_STRUCT_PTR_TAG)) {
	      for (j = prev_tok_idx; j <= stack_start_idx; j++)
		if (messages[j] == m_prev_tok -> receiver_msg)
		  break;
	      struct_start_tok_idx = j;
	      if ((__struct_expr_end = is_struct_expr_b
		   (messages, struct_start_tok_idx,
		    get_stack_top (messages))) <= 0)
		return FALSE;
	      __s = collect_tokens (messages, struct_start_tok_idx,
				    __struct_expr_end);
	      if (((c_defn = get_local_struct_defn (c -> type)) 
		   != NULL) ||
		  ((c_defn = get_global_var (c -> type)) != NULL)) {
		c = c_defn;
	      }
	      if ((c_mbr = struct_member_from_expr (__s, c)) != NULL) {
		strcpy (classname, basic_class_from_cvar 
			(messages[method_idx], c_mbr, 0));
		if ((class_object = get_class_object (classname)) != NULL) {
		  if (((m = get_instance_method 
			(m_prev_tok, class_object, 
			 M_NAME(messages[method_idx]), 
			 ERROR, FALSE)) 
		       != NULL) ||
		      ((m = get_class_method
			(m_prev_tok, class_object, 
			 M_NAME(messages[method_idx]), 
			 ERROR, FALSE)) 
		       != NULL)) {
		    m_prev_tok -> obj = create_object_init 
		      (class_object -> __o_name,
		       class_object -> __o_superclassname,
		       c -> name,
		       NULLSTR);
		    messages[method_idx] -> tokentype = METHODMSGLABEL;
		    messages[method_idx] -> attrs 
		      |= RCVR_OBJ_IS_C_STRUCT_ELEMENT;
		    messages[method_idx] -> receiver_msg = 
		      m_prev_tok -> receiver_msg;
		    messages[method_idx] -> receiver_obj = m_prev_tok -> obj;
		    return TRUE;
		  } else { /* if (((m = get_instance_method  ...  */
		    return FALSE;
		  } /* if (((m = get_instance_method ...  */
		} else { /* if ((class_object = get_class_object (classname))*/
		  return FALSE;
		} /* if ((class_object = get_class_object (classname)) */
	      }
	      free (__s);
	    } else {/* if ((c -> type_attrs & CVAR_ATTR_STRUCT_TAG)... */
	      if ((c -> attrs & CVAR_ATTR_STRUCT_TAG) ||
		  (c -> attrs & CVAR_ATTR_STRUCT_PTR_TAG) ||
		  ((c -> attrs & CVAR_ATTR_STRUCT_PTR) &&
		   !c -> members)) {
		stack_start_idx = stack_start (messages);
		for (j = prev_tok_idx; j <= stack_start_idx; j++)
		  if (messages[j] == m_prev_tok -> receiver_msg)
		    break;
		struct_start_tok_idx = j;
		if ((__struct_expr_end = is_struct_expr_b
		     (messages, struct_start_tok_idx,
		      get_stack_top (messages))) <= 0)
		  return FALSE;
		__s = collect_tokens (messages, struct_start_tok_idx,
				      __struct_expr_end);
		if (((c_defn = get_local_struct_defn (c -> type)) 
		     != NULL) ||
		    ((c_defn = get_global_var (c -> type)) != NULL)) {
		  c = c_defn;
		}
		if ((c_mbr = struct_member_from_expr (__s, c)) != NULL) {
		  strcpy (classname, basic_class_from_cvar 
			  (messages[method_idx], c_mbr, 0));
		  if ((class_object = get_class_object (classname)) != NULL) {
		    if (((m = get_instance_method 
			  (m_prev_tok, class_object, 
			   M_NAME(messages[method_idx]), 
			   ERROR, FALSE)) 
			 != NULL) ||
			((m = get_class_method
			  (m_prev_tok, class_object, 
			   M_NAME(messages[method_idx]), 
			   ERROR, FALSE)) 
			 != NULL)) {
		      m_prev_tok -> obj = create_object_init 
			(class_object -> __o_name,
			 class_object -> __o_superclassname,
			 c -> name,
			 NULLSTR);
		      messages[method_idx] -> tokentype = METHODMSGLABEL;
		      messages[method_idx] -> attrs 
			|= RCVR_OBJ_IS_C_STRUCT_ELEMENT;
		      messages[method_idx] -> receiver_msg = 
			m_prev_tok -> receiver_msg;
		      messages[method_idx] -> receiver_obj = m_prev_tok -> obj;
		      return TRUE;
		    } else { /* if (((m = get_instance_method  ...  */
		      return FALSE;
		    } /* if (((m = get_instance_method ...  */
		  } else { /* if ((class_object = get_class_object ... */
		    return FALSE;
		  } /* if ((class_object = get_class_object (classname)) */
		}
		free (__s);
	      } else { /* if ((c -> attrs & CVAR_ATTR_STRUCT_TAG)... */
		strcpy (classname, basic_class_from_cvar 
			(messages[method_idx], c, 0));
	      } /* if ((c -> attrs & CVAR_ATTR_STRUCT_TAG)... */
	    }/* if ((c -> type_attrs & CVAR_ATTR_STRUCT_TAG)... */
	  }/* if (c -> attrs & CVAR_ATTR_STRUCT_DECL) */
	  if ((class_object = get_class_object (classname)) != NULL) {
	    int __s_close_idx;
	    if (((m = get_instance_method 
		  (m_prev_tok, class_object, M_NAME(messages[method_idx]), 
		   ERROR, FALSE)) 
		 != NULL) ||
		((m = get_class_method
		  (m_prev_tok, class_object, M_NAME(messages[method_idx]), 
		   ERROR, FALSE)) 
		 != NULL)) {
	      m_prev_tok -> obj = create_object_init 
		(class_object -> __o_name,
		 class_object -> __o_superclassname,
		 c -> name,
		 NULLSTR);
	      messages[method_idx] -> tokentype = METHODMSGLABEL;
	      messages[method_idx] -> receiver_msg = m_prev_tok;
	      messages[method_idx] -> receiver_obj = m_prev_tok -> obj;
	      if ((__s_close_idx = prevlangmsg (message_stack (),
						prev_tok_idx)) != ERROR) {
		if (M_TOK(messages[__s_close_idx]) == ARRAYCLOSE) {
		  stack_start_idx = stack_start (messages);
		  for (n_blocks = 0, j = prev_tok_idx, arrayopen_idx = -1; 
		       (j <= stack_start_idx) && (arrayopen_idx == -1); j++) {
		    if (M_TOK(messages[j]) == ARRAYCLOSE)
		      --n_blocks;
		    if (M_TOK(messages[j]) == ARRAYOPEN)
		      ++n_blocks;
		    if (!n_blocks) {
		      arrayopen_idx = j;
		    }
		  }
		  messages[method_idx] -> attrs |= 
		    RCVR_OBJ_IS_C_ARRAY_ELEMENT;
		} /* if (M_TOK(messages[__s_close_idx]) == ARRAYCLOSE) */
	      }
	      return TRUE;
	    }
	  }
	}
      } else { /* if (m_prev_tok -> attrs & TOK_IS_POSTFIX_OPERATOR) */
	if (m_prev_tok -> attrs & RCVR_TOK_IS_C_STRUCT_EXPR) {
	  char *__s;
	  for (j = prev_tok_idx; j <= stack_start_idx; j++)
	    if (messages[j] == m_prev_tok -> receiver_msg)
	      break;
	  struct_start_tok_idx = j;
	  if (((c = get_local_var (M_NAME(messages[struct_start_tok_idx]))) 
		!= NULL) ||
	      (( c = get_global_var (M_NAME(messages[struct_start_tok_idx]))) 
	       != NULL)) {
	    __s = collect_tokens (messages, struct_start_tok_idx,
				  prev_tok_idx);
	    if (c -> type_attrs & CVAR_TYPE_STRUCT) {
	      /*
	       *  Case where struct declaration also has tags,
	       *  the struct will be the first tag, e.g.:
	       *  
	       *  struct _s {
	       *   ...
	       *  } s_tag, *s_ptr_tag;
	       *
	       *  and the CVAR contains the struct members.
	       */
	      if (!c -> members) {
		/*
		 *  Case where a struct declaration and tags are 
		 *  separate; e.g.:
		 *  
		 *  struct _s {
		 *   ...
		 *  };
		 *  struct _s s_tag, *s_ptr_tag;
		 *
		 *  Then we need to look up the struct definition.
		 */
		if (((c_defn = get_local_struct_defn (c -> type)) 
		     != NULL) ||
		    ((c_defn = get_global_var (c -> type)) != NULL)) {
		  c = c_defn;
		}
	      }
	    } else {
	      /* 
	       *  Case where the tag is a not the first tag:
	       *  struct _s {
	       *   ...
	       *  } <tag1>, <tag2>;
	       *
	       *  Here also we need to look up the struct definition.
	       *
	       */
	      if ((c -> attrs & CVAR_ATTR_STRUCT_TAG) ||
		  (c -> attrs & CVAR_ATTR_STRUCT_PTR_TAG)) {
		if (((c_defn = get_local_struct_defn (c -> type)) 
		     != NULL) ||
		    ((c_defn = get_global_var (c -> type)) != NULL)) {
		  c = c_defn;
		}
	      }
	    }
	    if ((c_mbr = struct_member_from_expr (__s, c)) != NULL) {
	      strcpy (classname, basic_class_from_cvar 
		      (messages[method_idx], c_mbr, 0));
	      if ((class_object = get_class_object (classname)) != NULL) {
		if (((m = get_instance_method 
		      (m_prev_tok, class_object, M_NAME(messages[method_idx]), 
		       ERROR, FALSE)) 
		     != NULL) ||
		    ((m = get_class_method
		      (m_prev_tok, class_object, M_NAME(messages[method_idx]), 
		       ERROR, FALSE)) 
		     != NULL)) {
		  m_prev_tok -> obj = create_object_init 
		    (class_object -> __o_name,
		     class_object -> __o_superclassname,
		     __s,
		     NULLSTR);
		  /***/
		  /*
		   * This way we don't have to collect the tokens
		   * for the expression several times later on.
		   */
		  free (__s);
		  messages[method_idx] -> tokentype = METHODMSGLABEL;
		  messages[method_idx] -> attrs |= RCVR_OBJ_IS_C_STRUCT_ELEMENT;
		  if (struct_start_tok_idx != prev_tok_idx)
		    messages[method_idx] -> attrs |= 
		      RCVR_TOK_IS_C_STRUCT_EXPR;
		  messages[method_idx] -> receiver_msg = 
		    m_prev_tok -> receiver_msg;
		  messages[method_idx] -> receiver_obj = m_prev_tok -> obj;

		  if (have_parens)
		    remove_outer_receiver_parens_a 
		      (messages, method_idx);

		  return TRUE;
		} else { /* if (((m = get_instance_method  ...  */
		  return FALSE;
		} /* if (((m = get_instance_method ...  */
	      } else { /* if ((class_object = get_class_object (classname))*/
		return FALSE;
	      } /* if ((class_object = get_class_object (classname)) */
	    }  else { /* if ((c_mbr = struct_member_from_expr (__s, c))... */
	      free (__s);
	      return FALSE;
	    }  /* if ((c_mbr = struct_member_from_expr (__s, c))... */
	  } else {/*if(((c=get_local_var (M_NAME(messages[struct_start_tok_idx]...*/
	    return FALSE;
	  }/*if(((c=get_local_var (M_NAME(messages[struct_start_tok_idx]...*/
	} else { /* if (m_prev_tok -> attrs & RCVR_TOK_IS_C_STRUCT_EXPR) */
	  if (((c = get_local_var (M_NAME(m_prev_tok))) != NULL) ||
	      (( c = get_global_var (M_NAME(m_prev_tok))) != NULL)) {
	    strcpy (classname, basic_class_from_cvar 
		    (messages[method_idx], c, 0));
	    if ((class_object = get_class_object (classname)) != NULL) {
	      if (((m = get_instance_method 
		    (m_prev_tok, class_object, M_NAME(messages[method_idx]), 
		     ERROR, FALSE)) 
		   != NULL) ||
		  ((m = get_class_method
		    (m_prev_tok, class_object, M_NAME(messages[method_idx]), 
		     ERROR, FALSE)) 
		   != NULL)) {
		m_prev_tok -> obj = create_object_init 
		  (class_object -> __o_name,
		   class_object -> __o_superclassname,
		   c -> name,
		   NULLSTR);
		messages[method_idx] -> tokentype = METHODMSGLABEL;
		messages[method_idx] -> receiver_msg = m_prev_tok;
		messages[method_idx] -> receiver_obj = m_prev_tok -> obj;
		return TRUE;
	      }
	    }
	  } /* if (((c = get_local_var (M_NAME(m_prev_tok))) != NULL) */
	}  /* if (m_prev_tok -> attrs & RCVR_TOK_IS_C_STRUCT_EXPR) */
      } /* if (m_prev_tok -> attrs & TOK_IS_POSTFIX_OPERATOR) */
    } /* if (M_TOK(m_prev_tok) == ARRAYCLOSE) */
  }
  return FALSE;
}


static void undefined_receiver_exception (int message_ptr, 
					  int actual_method_idx) {
  OBJECT_CONTEXT error_context;
  MESSAGE *m, *m_actual_method;
  char e[MAXLABEL];
  m = message_stack_at (message_ptr);
  m_actual_method = message_stack_at (actual_method_idx);
  error_context = object_context (message_stack (), 
				  actual_method_idx);
  switch (error_context)
    {
    case receiver_context:
      __ctalkExceptionInternal 
	(m, undefined_receiver_x, M_NAME(m_actual_method),0);
      break;
    case c_argument_context:
      if (is_fn_style_label (message_stack (), actual_method_idx,
			     get_stack_top (message_stack ()))) {
	sprintf 
	  (e, 
   "\"%s,\" is either not defined as a method, or prototyped as a function", 
	   M_NAME(m_actual_method));
	__ctalkExceptionInternal 
	  (m, ambiguous_operand_x, e,0);
      } else {
	sprintf 
	  (e, 
	   "Undefined receiver or variable argument, \"%s\"",
	   M_NAME(m_actual_method));
	__ctalkExceptionInternal 
	  (m, invalid_operand_x, e,0);
      }
      break;
    default:
      if (!is_method_name (M_NAME(m_actual_method))) {
	__ctalkExceptionInternal 
	  (m, undefined_label_x, m_actual_method -> name,0);
      }
      break;
    }
}

#define CVAR_AGGREGATE_DECL(__c) (((__c) -> attrs & CVAR_ATTR_ARRAY_DECL)||\
				  ((__c) -> attrs & CVAR_ATTR_STRUCT_DECL))

OBJECT *resolve (int message_ptr) {

  OBJECT *result_object = NULL;
  OBJECT *class_object; 
  OBJECT *instancevar_object,
    *classvar_object;
  MESSAGE *m;      
  int msg_frame_top = 0;       /* Avoid a warning.                     */
  int prev_label_ptr;          /* Stack pointers to messages with      */
  int prev_tok_ptr;
  int sender_idx;
  int method_attrs;
  int expr_close_paren,
    expr_open_paren;
  CVAR *cvar;
  MESSAGE *m_sender;
  MESSAGE *m_prev_label=NULL;  /* Messages that refer to previously    */
  int next_label_ptr;          /* Forward message references to check  */
  MESSAGE *m_next_label;       /*  method arguments, if necessary.     */
  MESSAGE *m_prev_tok;
  METHOD *method;
  OBJECT_CONTEXT param_context;
  int stack_top;
  int stack_start_idx;
  char cvar_block_alias[MAXLABEL];

  stack_top = get_stack_top (message_stack ());
  stack_start_idx = stack_start (message_stack ());

  m = message_stack_at (message_ptr);

  if (!m || !IS_MESSAGE(m))
    _error ("Invalid message in resolve.");

  if (m -> evaled)
    return NULL;

   if (!strcmp (M_NAME(m), "return")) {
     eval_return_expr (message_stack (), message_ptr);
     return NULL;
   }

   if (((cvar = get_local_var (M_NAME(m))) != NULL) ||
       ((cvar = get_global_var (M_NAME(m))) != NULL)) {
     m -> attrs |= TOK_IS_DECLARED_C_VAR;
     if (handle_cvar_arg_before_terminal_method_a (message_stack (), message_ptr)
	 == ERROR) {
       if ((next_label_ptr = nextlangmsg (message_stack (), message_ptr))
	   != ERROR) {
	 m_next_label = message_stack_at (next_label_ptr);
	 if (M_TOK(m_next_label) == LABEL) {
	   if (!is_c_keyword (M_NAME(m)) &&
	       !is_c_keyword (cvar -> type) &&
	       !function_not_shadowed_by_method (message_stack (), message_ptr) &&
	       !typedef_is_declared (M_NAME (m))) {
	     m -> obj = create_object 
	       (basic_class_from_cvar 
		(message_stack_at (message_ptr), cvar, 0), M_NAME(m));
	     if (((method = get_instance_method (m_next_label, m->obj,
						 m_next_label->name,
						 ERROR, FALSE)) 
		  != NULL) ||
		 ((method = get_class_method (m_next_label, m->obj,
					      m_next_label->name,
					      ERROR, FALSE)) 
		  != NULL) ||
		 is_method_proto (m->obj->__o_class->__o_name,
				  m -> name)) {
	       m_next_label -> receiver_msg = m;
	       m_next_label -> receiver_obj = m -> obj;
	       m_next_label -> tokentype = METHODMSGLABEL;
	       cvar -> scope |= LVAL_OBJECT_ALIAS;
	     }
	   }
	 } else { /* if (M_TOK(m_next_label) == LABEL) */
	   if ((M_TOK(m_next_label) == EQ) ||
	       (M_TOK(m_next_label) == ASR_ASSIGN) ||
	       (M_TOK(m_next_label) == ASL_ASSIGN) ||
	       (M_TOK(m_next_label) == PLUS_ASSIGN) ||
	       (M_TOK(m_next_label) == MINUS_ASSIGN) ||
	       (M_TOK(m_next_label) == MULT_ASSIGN) ||
	       (M_TOK(m_next_label) == DIV_ASSIGN) || 
	       (M_TOK(m_next_label) == BIT_AND_ASSIGN) ||
	       (M_TOK(m_next_label) == BIT_OR_ASSIGN) ||
	       (M_TOK(m_next_label) == BIT_XOR_ASSIGN)) {
	     cvar -> scope |= LVAL_OBJECT_ALIAS;
	   } else {
	     if (M_TOK(m_next_label) == ARRAYOPEN) {
	       int __a, __i, __n;
	       if ((__a = 
		    is_subscript_expr (message_stack (),
				       message_ptr, stack_top ))  > 0) {
		 for (__i = next_label_ptr; __i >= __a; __i--) {
		   message_stack_at (__i) -> receiver_msg = m;
		 }
		 if ((__n = nextlangmsg (message_stack (), __a)) != ERROR) {
		   MESSAGE *__m_n = message_stack_at (__n);
		   if ((M_TOK(__m_n) == INCREMENT) ||
		       (M_TOK(__m_n) == DECREMENT)) {
		     __m_n -> attrs |= TOK_IS_POSTFIX_OPERATOR;
		     __m_n -> receiver_msg = 
		       message_stack_at (__a) -> receiver_msg;
		   }
		   if (M_TOK(__m_n) == ARRAYOPEN) {
 		     if (get_object (M_NAME(m), NULL)) {
		       warning (__m_n, 
	      "Object, \"%s,\" has more than one subscript.", M_NAME(m));
		       warning (__m_n, 
   "The standard class library does not support multi-dimension arrays.");
 		     }
		   }
		 }
	       }
	     } else { /* if (M_TOK(m_next_label) == ARRAYOPEN) */
	       if ((M_TOK(m_next_label) == INCREMENT) ||
		   (M_TOK(m_next_label) == DECREMENT)) {
		 m_next_label -> attrs |= TOK_IS_POSTFIX_OPERATOR;
		 m_next_label -> receiver_msg = m;
	       } else { /* if ((M_TOK(m_next_label) == INCREMENT)... */
 		 if ((M_TOK(m_next_label) == PERIOD) ||
 		     (M_TOK(m_next_label) == DEREF)) {
		   int __t, __i, __postfix_idx;
		   MESSAGE *__m_postfix;
 		   if ((__t = is_struct_expr_b (message_stack (),
						message_ptr, stack_top)) > 0) {
		     m -> attrs |= RCVR_TOK_IS_C_STRUCT_EXPR;
		     for (__i = next_label_ptr; __i >= __t; __i--) {
		       message_stack_at (__i) -> receiver_msg = m;
		       message_stack_at (__i) -> attrs |= 
			 RCVR_TOK_IS_C_STRUCT_EXPR;
		     }
		     if ((__postfix_idx = is_trailing_postfix_op 
			  (message_stack (), __t, stack_top)) > 0) {
		       __m_postfix = message_stack_at (__postfix_idx);
		       if ((M_TOK(__m_postfix) == INCREMENT) ||
			   (M_TOK(__m_postfix) == DECREMENT)) {
			 __m_postfix -> attrs |= TOK_IS_POSTFIX_OPERATOR;
			 __m_postfix -> receiver_msg = m;
		       }
		     }
		   }/* if ((__t = is_struct_expr_b (message_stack ()...*/
		 } /* if ((M_TOK(m_next_label) == PERIOD) ... */
	       } /* if ((M_TOK(m_next_label) == INCREMENT) ... */
	     } /* if (M_TOK(m_next_label) == ARRAYOPEN) */
	   }
	 } /* if (M_TOK(m_next_label) == LABEL) */
       }
     }
   }

  if (is_c_keyword (M_NAME(m)) ||
      cvar ||
      global_var_is_declared_not_duplicated (message_stack (), message_ptr,
					     M_NAME(m)) ||
      function_not_shadowed_by_method (message_stack (), message_ptr) ||
      typedef_is_declared (M_NAME (m))) {
    if ((prev_tok_ptr = prevlangmsg (message_stack (), message_ptr))
	!= -1) {
      if ((m_prev_tok = message_stack_at (prev_tok_ptr)) != NULL) {
	if ((M_TOK(m_prev_tok) == LABEL) && (m_prev_tok -> obj != NULL))
	  warning (m, 
   "Object, \"%s,\" (Class %s) does not understand message, \"%s.\"",
		   m_prev_tok -> obj -> __o_name,
		   m_prev_tok -> obj -> __o_classname,
		   M_NAME(m));
      }
    }

    if (argblk && cvar) {

	if ((cvar = get_local_var (M_NAME(m))) != NULL) {

	  if (interpreter_pass == method_pass) {

	    sprintf (cvar_block_alias, "*%s", 
		     method_cvar_alias_basename
		     (new_methods[new_method_ptr+1] -> method,
		      cvar));
	  } else {
	    if (interpreter_pass == parsing_pass) {
	      sprintf (cvar_block_alias, "*%s", 
		       function_cvar_alias_basename( cvar));
	    }
	  }

	  free (m -> name);
	  m -> name = strdup (cvar_block_alias);

	}
    }

    return NULL;
  }
      
  /* Look for a class object. */

  if ((result_object = class_object_search (m -> name, TRUE)) != NULL) {
    m -> obj = result_object;
    default_method (message_stack (), message_ptr);
    return result_object;
  }

  if (is_expr_obj (m)) return M_VALUE_OBJ (m);

  /*
   *   Check for a method parameter reference and create an
   *   object if necessary.
   */
  if (interpreter_pass == method_pass) {
    int i, n;
    METHOD *method;
    /* method = new_methods[new_method_ptr + 1]; */
    method = new_methods[new_method_ptr + 1] -> method;

    if (method && IS_METHOD (method)) {
      for (i = 0; i < method -> n_params; i++) {

	if (!strcmp (m -> name, method -> params[i] -> name) &&
	    !is_instance_variable_message (message_stack (), message_ptr)) {

	  param_context = object_context (message_stack (), message_ptr);

  	  if ((param_context == c_context) ||
	      (param_context == c_argument_context)) {
	    if (!is_in_rcvr_subexpr_obj_check 
		(message_stack (), message_ptr, stack_top)) {

	      /* 
	       *  Replace the parameter name with the argument accessor
	       *  function.  In an argument context, the arguments are on 
	       *  the run-time stack in reverse order. 
	       */
	      for (n = 0; n < method -> n_params; n++) {
		if (!strcmp (m -> name, method -> params[n] -> name) &&
		    !(method -> params[n] -> attrs & PARAM_C_PARAM)) {
		  int __fn_idx, __n_th_arg;
		  int __end_idx;
		  if (need_rt_eval (message_stack (), message_ptr)) {
		    if ((__n_th_arg = 
			 obj_expr_is_fn_arg (message_stack (),
					     message_ptr,
					     stack_start_idx,
					     stack_top,
					     &__fn_idx)) != ERROR) {
		      (void)rt_obj_arg (message_stack (), 
					message_ptr, &__end_idx);
		    } else {
		      MESSAGE *__m;
		      __m = message_stack_at (message_ptr);
		      __m -> attrs |= TOK_IS_METHOD_ARG;
		      ctrlblk_pred_rt_expr (message_stack (), message_ptr);
		      __m -> attrs &= !TOK_IS_METHOD_ARG;
		    }
		  } else {
		    if (is_single_token_method_param (message_stack (),
						      message_ptr,
						      method)) {
		      if ((__n_th_arg = 
			   obj_expr_is_fn_arg (message_stack (),
					       message_ptr,
					       stack_start_idx,
					       stack_top,
					       &__fn_idx)) != ERROR) {
			// ***
			// TODO - could use an abbreviated fn here.
			//
			(void)rt_obj_arg 
			  (message_stack (), message_ptr, &__end_idx);
		      } else {
			(void)simple_param_accessor (message_stack (), 
						     message_ptr,
						     method);
		      }
		    } else {
		      c_param_expr_arg (message_stack (), message_ptr);
		    }
		  }
		}
	      }
	    } /* if (!is_in_rcvr_subexpr_obj_check (messages, msg_ptr... */
  	  } else { /* if ((param_context == c_context) ... */

	    /*   If the reference occurs in an argument, replace the
	     *   parameter with the rt arg call, and make sure that
	     *   there is an object associated with it so we can
	     *   perform further rt substitutions if necessary.
	     */
	    /*   Take 2 - There's no way, at the moment, of knowing
	     *   if an argument expression has a valid method.  Unless
	     *   the interpreter adds a second pass, or does something
	     *   really clever, we'll have to leave if as a run time 
	     *   expression.
	     */

	    if (param_context == argument_context) {
	      /***/
	      /* method_arg_rt_expr (message_stack (), message_ptr); */
	      method_param_arg_rt_expr (message_stack (), message_ptr,
					method);
	    } else {
	      if (param_context == receiver_context) {
		m -> obj = create_object_init
		  (((strlen (method -> params[i] -> class) != 0) ?
		    method -> params[i] -> class : "Object"),
		   "Object",
		   method -> params[i] -> name,
		   NULLSTR);
		save_method_object (m -> obj);
		if (is_argblk_expr (message_stack(), message_ptr)) {
		  return m -> obj;
		} else {
		  int __t;
		  rt_expr (message_stack (), message_ptr, &__t);
		}
	      }
	    }
	    if (!IS_OBJECT (m -> obj)) {
	      if (((class_object = 
		    class_object_search (method -> params[i] -> class, FALSE)) 
		  == NULL) &&
		  !is_pending_class(method -> params[i] -> class))
 		error (m, "Resolve: Undefined class %s.",
 		       method -> params[i] -> class);

 	      result_object =
 		create_object_init 
		(((strlen (method -> params[i] -> class) != 0) ?
		  method -> params[i] -> class : "Object"),
		 "Object",
		 method -> params[i] -> name,
		 NULLSTR);
	      save_method_object (result_object);
	      m -> obj = result_object;
	      return result_object;
	    }
  	  }
	}
      } /* for (i = 0; i < method -> n_params; i++) */
    } /* if (method && IS_METHOD (method)) */

    /* End of the method parameter evaluation stuff (hopefully). */

  } /* if (interpreter_pass == method_pass) */

  /* 
   *  Look for a global or local object.  
   */
  prev_label_ptr = prevlangmsg (message_stack (), message_ptr);
  if (prev_label_ptr != ERROR)
    m_prev_label = message_stack_at (prev_label_ptr);


  /*
   *  Handle simple <object> <var> type expressions. And in
   *  more complex expressions, if we can go on later without 
   *  backtracking to the first receiver label, it simplifies 
   *  the later processing a lot.
   *
   *  Note that finding a previous label depends on the parser
   *  framing if the label is in a receiver context.  Especially
   *  true of class variables.
   *
   *  Case for <classname> <classvarname> expression.
   */
  if (M_TOK(m) == LABEL) {
    if ((prev_label_ptr != ERROR) && 
	((result_object = 
	  get_class_variable (M_NAME(m), M_NAME(m_prev_label), FALSE))
	 != NULL)) {
      class_variable_expression (message_stack (), message_ptr);
      return result_object;
    } else {
      /*
       *  Case for <classvar> without receiver class.
       */
      if ((result_object = find_class_variable_2 
	   (message_stack (), message_ptr)) != NULL) {
	m -> obj = result_object;
	m -> attrs |= OBJ_IS_CLASS_VAR;
	default_method (message_stack (), message_ptr);
	return result_object;
      } else {
	/*
	 *  Case for <instance object> <classvar>.  We can
	 *  eliminate the instance object.
	 *
	 *  The <classvar> should be treated as stateless here,
	 *  not pre-evaled; i.e., 
	 *  the next time through resolve, the interpreter
	 *  starts to interpret the following methods and arguments.
	 */
	if (((next_label_ptr = nextlangmsg (message_stack (), message_ptr))
	     != ERROR) && 
	    get_object (M_NAME(m), NULL) && 
	    ((result_object = 
	      find_class_variable_2 (message_stack (),
				     next_label_ptr))
	     != NULL)) {
	  ++m -> evaled;
	  ++m -> output;
	  return result_object;
	} else {
	  /*
	   *  When retrieving an object, also check for instance
	   *  variables with the same name.  The function handles
	   *  expressions with instance variables later on.
	   */
	  if ((result_object = get_object (m -> name, NULL)) != NULL) {
	    if (!is_instance_variable_message (message_stack (), 
					       message_ptr)) {
	      /*
	       *  This is sort of a duplication of find_class_variable
	       *  above.  But get_object looks for class variables, too,
	       *  and we'll need to refactor a lot to get it to catch
	       *  this.  We also can't use a closure here.
	       */
	      if (!is_class_variable_message (message_stack (),
					      message_ptr)) {
		if ((prev_label_ptr != -1) && 
		    (m_prev_label -> attrs & TOK_IS_PREFIX_OPERATOR)) {
		  m -> obj = result_object;
		  prefix_method_expr_a (message_stack (), prev_label_ptr,
					message_ptr);
		  return result_object;
		} else {
		  if (m_prev_label && 
		      m_prev_label -> obj &&
		      !strcmp (m_prev_label -> obj -> __o_classname,
			       "Class")) {
		    /* Do something later. It's probably a 
		       class method with the same name as an 
		       object. */
		  } else {
		    /* Needed for default_method () */
 		    m -> obj = result_object;  
		    /*
		     *  These functions need to be grouped together.
		     *  If we need to adjust the evaluation of terminal
		     *  objects, then add to these functions.
		     */
		    if (use_new_c_rval_semantics 
			(message_stack (), message_ptr)) {
		      if (terminal_rexpr (message_stack (), message_ptr)
			  == ERROR)  {
			terminal_printf_arg (message_stack (), message_ptr);
		      }
		    } else {
		      default_method (message_stack (), message_ptr);
		    }
		    return result_object;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  } /* if (M_TOK(m) == LABEL) */

  /* 
   *  Check for, "self," "require," "return," "noMethodInit," 
   *  "eval," and "returnObjectClass." If they don't translate 
   *  into run time code, they get elided here also.
   */
   if (!strcmp (m -> name, "require")) {
     require_class (message_stack (), message_ptr);
     return NULL;
   }

   if (!strcmp (M_NAME(m), "eval")) {
     eval_keyword_expr (message_stack (), message_ptr);
     return NULL;
   }

   if (!strcmp (m -> name, "self"))  {
     /*
      *  Check for "self" within a constructor and issue
      *  a warning if necessary.
      */
       if ((interpreter_pass == method_pass) &&
	   IS_CONSTRUCTOR(new_methods[new_method_ptr+1]->method) &&
	   verbose_opt)
	 warning (m, "\"self\" used within a constructor.");
       if (need_rt_eval (message_stack (), message_ptr)) {
	 if (ctrlblk_pred) {
	   int __fn_idx, __n_th_arg;
	   if ((__n_th_arg = 
		obj_expr_is_fn_arg (message_stack (),
				    message_ptr,
				    stack_start_idx,
				    stack_top,
				    &__fn_idx)) != ERROR) {
	     int __end_idx;
	     rt_obj_arg (message_stack (), message_ptr, &__end_idx);
	   } else {
	     ctrlblk_pred_rt_expr (message_stack (), message_ptr);
	   }
	 } else { /* if (ctrlblk_pred) */
	   /***/
	   if (is_argblk_expr (message_stack(), message_ptr)) {
	     m -> obj = self_object (message_stack (), message_ptr);
	     return m -> obj;
	   } else {
	     int _expr_end;
	     if ((interpreter_pass != expr_check) && 
		 !have_complex_arg_block) {
	       (void)rt_self_expr (message_stack (), 
				   message_ptr, &_expr_end);
	     } else {
	       if (!IS_OBJECT (m -> obj)) {
		 return instantiate_self_object ();
	       }
	     }
	   }
	 }
       } else { /* if (need_rt_eval (message_stack (), message_ptr)) */
	 if ((result_object =
	      self_object (message_stack (), message_ptr)) != NULL) {
	   m -> obj = result_object;
	   param_context = object_context (message_stack (), message_ptr);
	   if (param_context == argument_context) {
	     if (!is_in_rcvr_subexpr_obj_check 
		 (message_stack (), message_ptr, stack_top)) {
	       /***/
	       /* method_arg_rt_expr (message_stack (), message_ptr); */
	       method_self_arg_rt_expr (message_stack (), message_ptr);
	       /*
		*  self_object () or some other function sets the
		*  output member. Clear it so the expression can
		*  be output.
		*/
	       m -> output = 0;
	       return result_object;
	     }
	   } else {
	     if (!IS_OBJECT (m->obj)) 
	       m -> obj = result_object = 
		 instantiate_self_object ();
	     default_method (message_stack (), message_ptr);
	     return result_object;
	   }
	 } /* if ((result_object = ... */
       } /* if (need_rt_eval (message_stack (), message_ptr)) */
   } /* if (!strcmp (m -> name, "self")  */

   /* 
    * If super is a receiver within an argument code block, inside
    * a method, replace with the method's self object.
    *
    * Otherwise, for now, wait until we evaluate whatever method
    * follows the super message.
    */
   if (str_eq (m -> name, "super")) {

     if (!argblk_super_expr (message_stack (), message_ptr))
       return NULL;
     /***/
#if 0
     if (interpreter_pass == method_pass) {
       OBJECT_CONTEXT super_context;
       super_context = object_context (message_stack (), message_ptr);
       switch (super_context)
	 {
	 case receiver_context:
	   if (argblk_stmt (message_ptr)) {
	     super_argblk_expr 
	       (message_stack (), message_ptr,
		new_methods[new_method_ptr+1]->method->rcvr_class_obj);
	     return NULL;
	   }
	   break;
	 default:
	   break;
	 }
     } else { /* if (interpreter_pass == method_pass) */
       OBJECT_CONTEXT super_context;
       OBJECT *blk_rcvr_object;
       super_context = object_context (message_stack (), message_ptr);
       switch (super_context)
	 {
	 case receiver_context:
	   if (argblk_stmt (message_ptr)) {
	     if ((blk_rcvr_object = 
		  argblk_rcvr_object (message_stack (), message_ptr))
		 != NULL) {
	       super_argblk_expr 
		 (message_stack (), message_ptr, blk_rcvr_object);
	       return NULL;
	     }
	   }
	   break;
	 default:
	   break;
	 }
     } /* if (interpreter_pass == method_pass) */
#endif
   } /* if (str_eq (m -> name, "super")) */

  if (fn_return (message_stack (), message_ptr) == 0)
    return NULL;

  /* 
   *  Now do the keyword stuff specifically for methods.
   */

  if (interpreter_pass == method_pass) {
    int i, next_label_ptr;
    if (!strcmp (m -> name, "returnObjectClass")) {
      if ((next_label_ptr = scanforward (message_stack (), message_ptr - 1,
			 message_frame_top_n (parser_frame_ptr () - 1), 
					 LABEL)) == ERROR) {
	warning (m, "Syntax error");
      }
      strcpy (new_methods[new_method_ptr + 1] -> method -> returnclass,
	      message_stack_at(next_label_ptr) -> name);
      if ((next_label_ptr = scanforward (message_stack (), message_ptr + 1,
			 message_frame_top_n (parser_frame_ptr () - 1), 
					 SEMICOLON)) == ERROR) {
	warning (m, "Syntax error");
      }
      for (i = message_ptr; i >= next_label_ptr; i--) {
	++(message_stack_at (i)) -> evaled;
	++(message_stack_at (i)) -> output;
      }
      return NULL;
    }
    if (!strcmp (M_NAME(m), "noMethodInit")) {
      new_methods[new_method_ptr + 1] -> method -> no_init = TRUE;
      ++(message_stack_at (message_ptr)) -> evaled;
      ++(message_stack_at (message_ptr)) -> output;
      if ((i = nextlangmsg (message_stack (), message_ptr)) == ERROR)
	warning (m, "Syntax error");
      if (message_stack_at (i) -> tokentype != SEMICOLON)
	warning (m, "Syntax error");
      ++(message_stack_at (i)) -> evaled;
      ++(message_stack_at (i)) -> output;
      return NULL;
    }
  }

  /* Attempt to resolve methods and arguments. */

  /*
   *  Check for modifiers or predicates to the method.  At the 
   *  moment, the only modifier is, "super."
   *
   *  sender_idx, m_sender, and m_actual_method are sort of misnomers,
   *  because they can point to anything that sends a message
   *  to a receiver.
   */
  
  sender_idx = 
    compound_method (message_stack (), message_ptr,
		     &method_attrs);
  m_sender = message_stack_at (sender_idx);

  /*
   *  We do, "for," loops separately because they all occur within 
   *  the same frame, so we use the CTRLBLK structure indexes 
   *  instead of the frame top.  
   */
  if (ctrlblk_pred) {
    if (C_CTRL_BLK -> stmt_type == stmt_for) {
      if (for_init)
	msg_frame_top = C_CTRL_BLK -> for_init_start;
      if (for_term)
	msg_frame_top = C_CTRL_BLK -> for_term_start;
      if (for_inc)
	msg_frame_top = C_CTRL_BLK -> for_inc_start;
    } else {
      /*
       *  NOTE - It may be necessary to add conditions for other
       *  special loop cases here.
       */
      msg_frame_top = message_frame_top_n (parser_frame_ptr ());
    }
  } else {
    if (ctrlblk_blk) {
      /*
       *  If we have a code block without braces, then set 
       *  msg_frame_top to blk_start_ptr.  If the control block
       *  is delimited by braces, we can simply use the frame 
       *  top.
       */
      if (C_CTRL_BLK -> braces == 0) {
	msg_frame_top = C_CTRL_BLK -> blk_start_ptr;
      } else {
	msg_frame_top = message_frame_top_n (parser_frame_ptr ());
      }
    } else {
      msg_frame_top = message_frame_top_n (parser_frame_ptr ());
    }
  }

  prev_label_ptr = scanback (message_stack (), sender_idx + 1,
  			     msg_frame_top, LABEL);
  prev_tok_ptr = prevlangmsg (message_stack (), sender_idx);

  /*  If the label isn't resolved as an object above, it could be
   *  interpreted as an argument or method of an object in the
   *  predicate.
   */

  /* 
   *  This might not be strictly necessary, but we can quit 
   *  now when we handle control structure blocks somewhere
   *  else.
   */

  if (ctrlblk_blk) {
    if ((sender_idx <= C_CTRL_BLK -> blk_start_ptr) &&
	(sender_idx >= C_CTRL_BLK -> blk_end_ptr) &&
	(prev_label_ptr <= C_CTRL_BLK -> pred_start_ptr) &&
	(prev_label_ptr >= C_CTRL_BLK -> pred_end_ptr))
      return NULL;
  }

  if (prev_label_ptr == ERROR) {
    if ((prev_label_ptr = scanback (message_stack (), sender_idx + 1,
 				    msg_frame_top, C_KEYWORD)) != -1) {
      return NULL;
    } else {
      OBJECT_CONTEXT c;
      /* 
       *  Issue a warning only for label tokens for which a receiver object
       *  has not been defined, or if we can determine that the 
       *  undefined label token appears in an object context.
       */
      if ((m_sender -> tokentype == LABEL) && 
	  !IS_DEFINED_LABEL (m_sender -> name) && 
	  !result_object) {
	(void)undefined_label_check (message_stack (), sender_idx);
	return NULL;
      }

      c = object_context (message_stack (), sender_idx);
      switch (c)
	{
	case argument_context:
	case receiver_context:
	  if ((M_TOK(m_sender) == LABEL) || 
	      (M_TOK(m_sender) == METHODMSGLABEL)) {
	    if (!IS_DEFINED_LABEL (m_sender -> name))
	      __ctalkExceptionInternal (m, undefined_receiver_x, m_sender -> name,0);
	    return NULL;
	  }
	  break;
	case c_context:
	  if (prev_tok_ptr != -1) {
	    MESSAGE *__m_prev_tok;
	    __m_prev_tok = message_stack_at (prev_tok_ptr);
	    if (IS_NUMERIC_CONSTANT_TOK (M_TOK(__m_prev_tok))) {
	      if (!IS_DEFINED_LABEL (m_sender -> name)) {
		char __tmpbuf[MAXLABEL];
		sprintf 
		  (__tmpbuf, 
   "Undefined label, \"%s,\" follows a numeric constant receiver", 
		   M_NAME(m_sender));
		__ctalkExceptionInternal (m, undefined_receiver_x, 
					  __tmpbuf,0);
		return NULL;
	      }
	    }
	  }
	  break;
	default:
	  warning (m, "resolve: context not handled in switch.");
	  break;
	}
    } /* else */
  } else { /* if (prev_label_ptr == ERROR) */
    m_prev_label = message_stack_at (prev_label_ptr);
  } /* if (prev_label_ptr == ERROR) */

  /*   
   *   This is here mainly as a syntax check for odd variable 
   *   cases, but it might save a lot of work further on, too.
   */


  if (m_prev_label && !m_prev_label -> value_obj) {
    if (is_c_data_type (M_NAME(m_prev_label)) ||
	get_typedef (M_NAME(m_prev_label)) ||
	is_incomplete_type (M_NAME(m_prev_label)))
      return NULL;
  }

  /* 
   *   Check if the the most recent reference is an object, or a 
   *   near reference is a method.
   */

  if (prev_label_ptr != ERROR) {
    if (IS_OBJECT (m_prev_label -> obj)) {

      /* Now do the method stuff. */

      /* The previous reference is a class object. */
      if (IS_CLASS_OBJECT(m_prev_label -> obj)) {
	/*
	 *  This might need a check for the start of 
	 *  the parser frame.
	 */
	if (interpreter_pass != expr_check) {
	  if (((method = get_instance_method (m_prev_label, m_prev_label->obj,
					      m_sender->name,
					      ERROR, FALSE)) 
	       != NULL) ||
	      ((method = get_class_method (m_prev_label, m_prev_label->obj,
					   m_sender->name,
					   ERROR, FALSE)) 
	       != NULL) ||
	      is_method_proto (m_prev_label->obj->__o_class->__o_name,
			       m_sender -> name)) {
	    m_sender -> receiver_msg = m_prev_label;
	    m_sender -> receiver_obj = m_prev_label -> obj;
	    m_sender -> tokentype = METHODMSGLABEL;
	  } else {
	    if (m_prev_label -> obj &&
		IS_CLASS_OBJECT (m_prev_label -> obj)) {
	      char __s[MAXLABEL];
	      sprintf (__s, "Undefined label, \"%s,\" follows a class object without a constructor.", m_sender -> name);
	      warning (m_sender, __s);
	      __ctalkExceptionInternal (m_sender, undefined_method_x,
					m_sender -> name,0);
	    } else {
	      __ctalkExceptionInternal (m_sender, undefined_method_x,
					m_sender -> name,0);
	    }
	    return NULL;
	  }
	}
      } else {
	
	/* 
	 *   The previous label refers to an object instance. 
	 *   Determine if this label refers to a method or 
	 *   an instance variable.  Give precedence to instance
	 *   variables.
	 */

	if ((class_object = 
	     get_class_object (m_prev_label -> obj -> __o_classname))
	    == NULL)
 	  error (m_prev_label, "Unknown class \"%s.\"", 
		 m_prev_label ->obj -> __o_classname);
	/*
	 *   We still have to make sure that the method (at sender_idx)
	 *   immediately follows the receiver (at prev_label_ptr).
	 *   If not, it's a complex receiver expression, probably in
	 *   parens, which gets handled below.
	 */
	if (prev_tok_ptr == prev_label_ptr) {
	  m_sender -> receiver_msg = m_prev_label;
 	  m_sender -> receiver_obj = m_prev_label -> obj;
	  instancevar_object = NULL;
 	  if (((instancevar_object = 
 		__ctalkGetInstanceVariable (m_prev_label -> obj, m_sender -> name,
 					     FALSE)) != NULL) ||
 	      ((method = get_instance_method (m_sender,
 					      class_object, 
				      m_sender -> name, 
					      ERROR, FALSE)) != NULL) ||
	      ((method = get_class_method (m_sender,
					   class_object,
					   m_sender -> name, 
					   ERROR, FALSE)) != NULL) ||
	      is_method_proto (m_prev_label->obj->__o_class->__o_name,
			       m_sender -> name)) {
	    /* 
	     *  This needs to be lazy right now because of
	     *  possible object mutations -- we simply defer
	     *  everything until run-time evaluation when
	     *  trying to generate code in method_call.
	     */
	    m_sender -> tokentype = METHODMSGLABEL;
	    m_sender -> obj = instancevar_object;
	    if (instancevar_object)
	      m_sender -> attrs |= OBJ_IS_INSTANCE_VAR;
	  } else {
	    /*
	     *  Class variable alone, possibly (almost certainly if it's
	     *  a complex object) with value class different than var 
	     *  class.
	     */
	    if ((m_prev_label->attrs & OBJ_IS_CLASS_VAR) &&
		m_prev_label->obj->instancevars && 
		((method = 
		  get_instance_method (m_sender,
       get_class_object (m_prev_label->obj->instancevars->__o_classname), 
				       M_NAME(m_sender), 
				       ERROR, FALSE)) != NULL)) {
	      m_sender -> tokentype = METHODMSGLABEL;
	    } else {
	      if (is_method_proto (CLASSNAME(class_object), 
				   M_NAME(m_sender)) &&
		  !method_proto_is_output (CLASSNAME(class_object),
					   M_NAME(m_sender)) &&
		  !method_from_proto) {
		method_from_prototype (M_NAME(m_sender));
	      } else {
		MESSAGE *m_prev_tok;

		m_prev_tok = message_stack_at (prev_tok_ptr);

		/*
		 * First look for a case where we have a 
		 * <rcvr> <undefined-label>, and if its
		 * prototype is defined in a superclass that is 
		 * still being evaluated, issue a warning and 
		 * defer the evaluation until run time.
		 * 
		 * We can't simply evaluate the method proto
		 * without getting into circular, recursive 
		 * evaluations, which causes a lot of stress on the
		 * parser and a lot of exceptions to the validity
		 * checking of methods and parameters, so we only 
		 * issue a warning.  
		 *
		 * If the warning appears in one of the basic classes,
		 * it's probably better to add that method to the 
		 * basic class anyway, even if the method is already in 
		 * the superclass.
		 */
		if ((M_TOK (m_sender) == LABEL) && 
		    (M_TOK (m_prev_tok) == LABEL) &&
		    (m_prev_tok -> obj != NULL)) {
		  if (is_superclass_method_proto 
		      (m_prev_tok -> obj -> __o_classname, 
		       M_NAME(m_sender))) {
		    warning (m_sender, "Undefined Label.\n"
			     "\tLabel, \"%s,\" is not (yet) defined. "
			     "Deferring evaluation until run time.",
			     M_NAME (m_sender));
		    deferred_method_eval_a (message_stack (),
					    prev_tok_ptr,
					    &sender_idx);
		  } else {
		    __ctalkExceptionInternal (m_sender, undefined_method_x,
					      M_NAME(m_sender),0);
		  }
		} else {		    
		  __ctalkExceptionInternal (m_sender, undefined_method_x,
					    M_NAME(m_sender),0);
		}
	      }
	      return NULL;
	    }
	  }
	}
      }
    } else { /*     if (IS_OBJECT (m_prev_label -> obj)) */
      if (!strcmp (M_NAME(m_prev_label), "self")) {
 	  if ((instancevar_object = 
 		__ctalkGetInstanceVariable (rcvr_class_obj, 
					    M_NAME(m),
					    FALSE)) != NULL) {
	    /* TO DO - 
	     * factor this out here, and in self_object (),
	     * in object.c, and try to use either 
	     * instance_variables_from_class_definition (),
	     * or instance_vars_from_class_object (), 
	     * everywhere, not both.
	     *
	     * This can also come earlier in the evaluation.
	     * Annd... create a SELF_OBJECT attribute, so
	     * we don't need to strcmp so much.
	     */
	    if (!IS_OBJECT (m_prev_label -> obj))
	      m_prev_label -> obj
		= instantiate_self_object ();
	    m -> obj = instancevar_object;
	    m -> receiver_obj = rcvr_class_obj;
	    m -> attrs |= OBJ_IS_INSTANCE_VAR;
	  } else {
	    if ((classvar_object = 
		 __ctalkGetClassVariable (rcvr_class_obj, 
					     M_NAME(m),
					     FALSE)) != NULL) {
	    if (!IS_OBJECT (m_prev_label -> obj))
	      m_prev_label -> obj
		= instantiate_self_object ();
	      m -> obj = classvar_object;
	      m -> receiver_obj = rcvr_class_obj;
	      m -> attrs |= OBJ_IS_CLASS_VAR;
	    }
	  }
      } /* if (!strcmp (M_NAME(m_prev_label), "self")) */
    } /*     if (IS_OBJECT (m_prev_label -> obj)) */

    if ((next_label_ptr = nextlangmsg (message_stack (), sender_idx))
	!= ERROR) {
      m_next_label = message_stack_at (next_label_ptr);

      /*
       *  The m_next_label token is set in is_argblk_expr ().
       *  Fixup here now that we know the actual receiver object
       *  and receiver message.
       */
      if ((M_TOK(m_next_label) == METHODMSGLABEL) && argblk) {
	m_next_label->receiver_obj = m_sender->receiver_obj;
	m_next_label->receiver_msg = m_sender->receiver_msg;
      }

      if (m_sender -> tokentype == METHODMSGLABEL) {
  	if ((expr_close_paren = 
	     is_in_rcvr_subexpr_obj_check 
	     (message_stack (), sender_idx, stack_top)) == 0) {
	  method_call (sender_idx); 
  	} else {
	  int i_2;
	  expr_open_paren = match_paren_rev (message_stack (),
					     expr_close_paren,
					     stack_start_idx);
	  /*
	   *  If there are previous objects, they get resolved later,
	   *  so clear them here.  But check for a C variable as 
	   *  the argument and register it if necessary.
	   */
	  for (i_2 = expr_open_paren; i_2 >= expr_close_paren; i_2--) {
	    MESSAGE *m_2;
	    m_2 = message_stack_at (i_2);
	    if ((M_TOK(m_2) == LABEL) && m_2 -> obj)
	      m_2 -> obj = NULL;
	  }
	  register_cvar_arg_expr_a (message_stack (), sender_idx);
	}
      } else {
	if (m->receiver_obj && (M_TOK(m_sender) == LABEL)) {
	  undefined_label_check (message_stack (), sender_idx);
	} else {
	  if (interpreter_pass == parsing_pass) {
	      /*
	       *  Here is where we handle complex receiver
	       *  expressions that occur in C functions.
	       *  Expressions in methods get handled below,
	       *  because we also need evaluate method 
	       *  parameters.
	       */
	    if ((m_sender -> tokentype == LABEL) && 
		!m_sender -> obj) {
   	      if (!IS_DEFINED_LABEL (m_sender -> name)) {
 		if (is_possible_receiver_subexpr_postfix (message_stack (),
 						  sender_idx)) {
		  handle_subexpr_rcvr (sender_idx);
		  return NULL;
 		} else { /* if (is_possible_receiver_subexpr_postfix ... */
		  if (cvar_is_method_receiver (message_stack (), 
						sender_idx)) {
		    method_call (sender_idx);
		    return NULL;
		  } else {

		    /* 
		     *  If there's a series <label> <label> <label>
		     *  that hasn't been resolved yet, look for 
		     *  a C type.  This is instead of going through
		     *  is_c_var_declaration_msg () again.
		     */

		    if (m_prev_label -> attrs & TOK_IS_DECLARED_C_VAR) {
		      int prev_label_ptr_2;
		      MESSAGE *m_prev_label_2;

		      if ((prev_label_ptr_2 = prevlangmsg (message_stack (),
							   prev_label_ptr))
			  != ERROR) {
			m_prev_label_2 = message_stack_at (prev_label_ptr_2);

			if (m_prev_label_2 -> attrs & TOK_IS_DECLARED_C_VAR) {
			  if (!is_c_data_type (M_NAME(m_prev_label_2))) {

			    undefined_receiver_exception (message_ptr,
							  sender_idx);
			    return NULL;

			  } else {

			    ++m -> evaled;
			    return NULL;

			  }

			} else {

			  undefined_receiver_exception (message_ptr,
							sender_idx);
			  return NULL;

			}
		      }

		    } else {

		      undefined_receiver_exception (message_ptr,
						  sender_idx);
		      return NULL;
		    }
		  }
 		}
 	      } else {
		/* 
		 *  If we got this far, first check for a postfix
		 *  method that shadows a C name before giving up. 
		 *  First check for a C syntax.  We've already
		 *  checked for a control structure predicate, so
		 *  check for the C statements, by syntax, 
		 *  which is quicker.  Handle only the simple cases
		 *  right now - there will probably be more complicated
		 *  occurrences of shadowing later.
		 */
		CVAR *c_1;
		int i_1;
		if (prev_tok_ptr != -1) {

		  m_prev_tok = message_stack_at (prev_tok_ptr);

		  if ((c_1 = get_global_var(M_NAME(m_sender))) != NULL) {
		    if (IS_C_OP_TOKEN_NOEVAL (M_TOK(m_prev_tok)))
		      return M_VALUE_OBJ (m_prev_label);
		  }
		  if ((c_1 = get_local_var(M_NAME(m_sender))) != NULL) {
		    if (IS_C_OP_TOKEN_NOEVAL (M_TOK(m_prev_tok)))
		      return M_VALUE_OBJ (m_prev_label);
		  }
		  if ((i_1 = is_struct_member(M_NAME(m_sender))) != 0) {
		    if ((M_TOK(m_prev_tok) == PERIOD) || 
			(M_TOK(m_prev_tok) == DEREF))
		      return M_VALUE_OBJ (m_prev_label);
		  }
		}
		/* 
		 *  Okay, it's not a C expression (or it's not going to
		 *  parse anyway).
		 */
		if (is_possible_receiver_subexpr_postfix (message_stack (),
							  sender_idx)) {
		  handle_subexpr_rcvr (sender_idx);
		  return NULL;
		}
	      }
	    } else {/* if ((m_sender -> tokentype == LABEL) &&  */
	      if (!m_sender -> obj && !m_prev_label -> obj) {
		if (prev_tok_ptr != prev_label_ptr) {
		  m_prev_tok = message_stack_at (prev_tok_ptr);
 		  if (IS_CONSTANT_TOK(M_TOK(m_prev_tok))) {
		    /* If not a case statement... case <const><:> */
		    /* Should be more specific if we ever decide to 
		       overload a ':' operator. */
		    if (M_TOK(m_sender) != COLON) {
		      register_cvar_arg_expr_a 
			(message_stack (), sender_idx);
		      if (method_call_constant_tok_expr_a (message_stack (),
						   sender_idx)) {
			return NULL;
		      }
  		    }
		  } else {
		    /*
		     *  An <array>[<idx>] <receiver> expression.
		     */
		    if (is_aggregate_term_a (message_stack (), prev_tok_ptr)) {
		      register_cvar_rcvr_expr_b 
			(message_stack (), sender_idx);
		      if (method_call_subexpr_postfix_a (message_stack (),
							 sender_idx)) {
			return NULL;
		      }
		    } else {
		      /*
		       *  A <close paren> <math_op> expression
		       */
		      if (method_call_subexpr_postfix_a (message_stack (),
							 sender_idx)) {
			return NULL;
		      } else {
			/*
			 *  Mark a prefix method syntactically _unless_
			 *  it is a ! operator and it is the first token
			 *  of a contrl block predicate. We can pass the 
			 *  ! token verbatim to the output.
			 */
			prefix_method_attr (message_stack (), message_ptr);
		      }
		    }
		  }
		} else {
		  /*
		   *   If the previous label is a CVAR, and we're
		   *   in a receiver expression.
		   */
		  CVAR *prev_label_cvar;
		  if (((prev_label_cvar=get_local_var (M_NAME(m_prev_label)))
		       != NULL) ||
		      ((prev_label_cvar=get_global_var (M_NAME(m_prev_label)))
		       != NULL)) {
		    if (CVAR_AGGREGATE_DECL(prev_label_cvar)) {
		      if (method_call_subexpr_postfix_b 
			  (message_stack (),
			   sender_idx) == -2) {
			constrcvr_unhandled_case_warning (message_stack (),
							  sender_idx);
		      }
		      return NULL;
		    } else {
		      if (register_cvar_rcvr_expr_a (message_stack (),
						     sender_idx,
						     prev_label_cvar,
						     M_NAME(m_prev_label))) {
			return NULL;
		      } else { /* if (register_cvar_rcvr_expr_a .. */
			if (set_cvar_rcvr_postfix_attrs_a 
			    (message_stack (),
			     sender_idx,
			     prev_label_ptr,
			     prev_label_cvar,
			     M_NAME(m_prev_label))) {
			  return NULL;
			}
		      } /* if (register_cvar_rcvr_expr_a .. */
		    }
		  }
		}
	      }
	    }
	  } else {
	    /*
	     *  Resolve handles methods separately than functions
	     *  because it needs to check method parameters also.
	     */
	    if (interpreter_pass == method_pass) {
	      if ((m_sender -> tokentype == LABEL) && 
		  !m_sender -> obj) {
		if (!IS_DEFINED_LABEL (m_sender -> name) &&
		    !is_method_parameter 
		    (message_stack (), sender_idx)) {
		  if (is_possible_receiver_subexpr_postfix (message_stack (),
 						  sender_idx)) {
		    handle_subexpr_rcvr (sender_idx);
		    return NULL;
		  } else { /* if (is_possible_receiver_subexpr_postfix ... */
		    if (cvar_is_method_receiver (message_stack (),
						 sender_idx)) {
		      method_call (sender_idx);
		      return NULL;
		    } else {
		      undefined_receiver_exception (message_ptr,
						    sender_idx);
		      return NULL;
		    }
		  } /* if (is_possible_receiver_subexpr_postfix ... */
		}  else { /* if (!IS_DEFINED_LABEL (m_sender -> name) &&...*/

		  /* 
		   *  Look for a shadowed C variable again, if we've gotten
		   *  this far.  All of these near-duplicate clauses could be
		   *  squirreled away in some file and catalogued, if we
		   *  can factor out the context any further.... 
		   *  But this runs minimally faster, anyway.  And the fn
		   *  really isn't that big, either.
		   *
		   *  Again, for now, the clause looks only for the syntax
		   *  of common C statements - parsing the syntax a little
		   *  is faster.
		   */
		  CVAR *c_1;
		  int i_1;
		  if (prev_tok_ptr != -1) {

		    m_prev_tok = message_stack_at (prev_tok_ptr);

		    if ((c_1 = get_global_var(M_NAME(m_sender))) != NULL) {
		      if (IS_C_OP_TOKEN_NOEVAL (M_TOK(m_prev_tok)))
			return M_VALUE_OBJ (m_prev_label);
		    }
		    if ((c_1 = get_local_var(M_NAME(m_sender))) != NULL) {
		      if (IS_C_OP_TOKEN_NOEVAL (M_TOK(m_prev_tok)))
			return M_VALUE_OBJ (m_prev_label);
		    }
		    if ((i_1 = is_struct_member(M_NAME(m_sender))) != 0) {
		      if ((M_TOK(m_prev_tok) == PERIOD) || 
			  (M_TOK(m_prev_tok) == DEREF))
			return M_VALUE_OBJ (m_prev_label);
		    }
		  }

		  /* Okay, again, it's a ctalk expression (or it 
		   * wouldn't compile anyway).
		   */

		  if (is_possible_receiver_subexpr_postfix (message_stack (),
							    sender_idx)) {
		    handle_subexpr_rcvr (sender_idx);
		    return NULL;
		  }

		}
	      } else {/* if ((m_sender -> tokentype == LABEL) ... */
		if (!m_sender -> obj && !m_prev_label -> obj) {
		  if (prev_tok_ptr != prev_label_ptr) {
		    m_prev_tok = message_stack_at (prev_tok_ptr);
		    if (IS_CONSTANT_TOK(M_TOK(m_prev_tok))) {
		    /* If not a case statement... case <char * const><:> */
		      if (M_TOK(m_sender) != COLON) {
			register_cvar_arg_expr_a 
			  (message_stack (), sender_idx);
			if (method_call_constant_tok_expr_a (message_stack (),
							     sender_idx)) {
			  return NULL;
			}
		      }
		    } else { /* if (CONSTANT_TOK(m_prev_tok))  */

		      /* An <array>[<idx>] <tok> expression */
		      if (is_aggregate_term_a (message_stack (), prev_tok_ptr)) {
			register_cvar_rcvr_expr_b 
			  (message_stack (), sender_idx);
			if (method_call_subexpr_postfix_a (message_stack (),
							   sender_idx)) {
			  return NULL;
			} else {
			  /*
			   *  A <close paren> <math_op> expression
			   */
			  if (method_call_subexpr_postfix_a (message_stack (),
							     sender_idx)) {
			    return NULL;
			  } else {
			    /*
			     *  Mark a prefix method syntactically _unless_
			     *  it is a ! operator and it is the first token
			     *  of a contrl block predicate.  These we
			     *  can send verbatim to the output.
			     */
			    prefix_method_attr (message_stack (), message_ptr);
			  }
			} /* if (method_call_subexpr_postfix_a ... */
		      }  else { /* if (is_aggregate_term_a ... */
			if (method_call_subexpr_postfix_a (message_stack (),
							   sender_idx)) {
			  return NULL;
			} else {
			  /*
			   *  A <close paren> <math_op> expression
			   */
			  if (method_call_subexpr_postfix_a (message_stack (),
							     sender_idx)) {
			    return NULL;
			  } else {
			    /*
			     *  Again, mark a prefix method syntactically 
			     *  _unless_ it is a ! operator and it is the 
			     *  first token of a contrl block predicate.
			     */
			    prefix_method_attr (message_stack (), message_ptr);
			  }
			} /* if (method_call_subexpr_postfix_a ... */
		      } /* if (is_aggregate_term_a ... */
		    }  /* if (CONSTANT_TOK(m_prev_tok))  */
		  } else { /* if (prev_tok_ptr != prev_label_ptr) */
		    /*
		     *   If the previous label is a CVAR, and we're
		     *   in a receiver expression.
		     */
		    CVAR *prev_label_cvar;
		    if (((prev_label_cvar=get_local_var 
			  (M_NAME(m_prev_label)))
			 != NULL) ||
			((prev_label_cvar=get_global_var 
			  (M_NAME(m_prev_label)))
			 != NULL)) {
		      if (CVAR_AGGREGATE_DECL(prev_label_cvar)) {
			if (!method_call_subexpr_postfix_b 
			    (message_stack (), sender_idx) == -2) {
			  constrcvr_unhandled_case_warning (message_stack (),
							    sender_idx);
			}
			return NULL;
		      } else {
			if (register_cvar_rcvr_expr_a (message_stack (),
						       sender_idx,
						       prev_label_cvar,
						       M_NAME(m_prev_label))){
			  return NULL;
			} else { /* if (register_cvar_rcvr_expr_a */
			  if (set_cvar_rcvr_postfix_attrs_a 
			      (message_stack (),
			       sender_idx,
			       prev_label_ptr,
			       prev_label_cvar,
			       M_NAME(m_prev_label))){
			    return NULL;
			  }
			} /* if (register_cvar_rcvr_expr_a */
		      }
		    }
		  } /* if (prev_tok_ptr != prev_label_ptr) */
		} /* if (!m_sender -> obj && !m_prev_label -> obj)*/ 
	      } /* if ((m_sender -> tokentype == LABEL) ... */
	    } /* if (interpreter_pass == method_pass) */
	  }
	}
      }
      return M_VALUE_OBJ (m_prev_label);
    }

    return m_prev_label -> obj;
  }

  m_sender -> obj = result_object;
  if (m_prev_label && IS_MESSAGE(m_prev_label)) 
    m_sender -> receiver_obj = m_prev_label -> obj;

  return result_object;

}
